//===============================================================================
// Microsoft Caching Application Block for .NET
// http://msdn.microsoft.com/library/default.asp?url=/library/en-us/dnpag/html/CachingBlock.asp
//
// CacheManager.cs
// This class manages all the activities of the Cache Application Block like
// adding, updating, getting, flushing and removing the data pertaining to the
// Cache Data Store.
//
//===============================================================================
// Copyright (C) 2003 Microsoft Corporation
// All rights reserved.
// THIS CODE AND INFORMATION IS PROVIDED "AS IS" WITHOUT WARRANTY
// OF ANY KIND, EITHER EXPRESSED OR IMPLIED, INCLUDING BUT NOT
// LIMITED TO THE IMPLIED WARRANTIES OF MERCHANTABILITY AND/OR
// FITNESS FOR A PARTICULAR PURPOSE.
//===============================================================================


using System;
using System.Configuration;
using System.Reflection;
using System.Runtime.Remoting;

// Used to handle exceptions
using Microsoft.ApplicationBlocks.ExceptionManagement;

namespace Microsoft.ApplicationBlocks.Cache
{
	/// <summary>
	///	This class manages all the activities of the Cache Application Block 
	///	like adding, updating, getting, flushing and removing the data 
	///	pertaining to the Cache Data Store.
	/// </summary>
	public sealed class CacheManager
	{
		#region Private members

		private CacheService manageCacheService;
		private static ICacheStorage manageCacheStorage;
		
		#endregion

		#region Property

		/// <summary>
		/// 
		/// </summary>
		public static ICacheStorage CacheStorage
		{
			get{return manageCacheStorage;}
		}

		#endregion

		// Creating a static instance
		private static CacheManager cachManager = new CacheManager();

		#region Constructor
		
		/// <summary>
		///	Class constructor to read the storage information from the 
		///	App.config file.
		/// </summary>
		private CacheManager()
		{
			try
			{
				// Read storage mode from config file
				if(! Object.Equals(
					CacheConfiguration.Config.StorageInformation, null) )
				{
					string mode = 
						CacheConfiguration.Config.StorageInformation[ 
						CacheResources.ResourceManager [ "RES_Mode" ] ];
					if( Object.Equals(mode.ToLower(), 
						CacheResources.ResourceManager[
							"RES_OutProc" ] ) )
					{
						string remotingUrl = 
							CacheConfiguration.Config.StorageInformation[
							CacheResources.ResourceManager[ 
								"RES_RemotingUrl" ] ];

						if( Object.Equals(remotingUrl, null) || 
							Object.Equals(remotingUrl.Length, 0) )
						{
							throw new ConfigurationException( 
								CacheResources.ResourceManager[ 
								"RES_ExceptionInvalidConfigurationStorageProviderURL" 
								] );
						}

						// Configure the Remoting framework
						RemotingConfiguration.RegisterWellKnownClientType(
							typeof(
							CacheService ),
							remotingUrl );
					}
				
					// Initialize the CacheStorage
					manageCacheStorage = 
						CacheConfiguration.Config.StorageInformation.
						CreateInstance();
				}
				// Initialize the default CacheStorage
				else 
				{
					manageCacheStorage = new Storages.SingletonCacheStorage(); 
				}

				// Initialize the CacheService 
				manageCacheService = CacheService.GetCacheService();
				manageCacheService.InitializeLifetimeService();

			}
			catch( Exception genException )
			{
				ExceptionManager.Publish( genException );
				throw;
			}
		}
		#endregion

		#region Indexer

		/// <summary>
		///	Gets the value for the key.
		/// </summary>
		public object this[ string key ] 
		{
			get{return this.GetData(key);} 
		}
			
		#endregion

		#region Public Methods

		/// <summary>
		///	This method is used to return the singleton instance of the 
		///	CacheManager object.
		/// </summary>
		/// <remarks>
		///	CacheManager manager = CacheManager.GetCacheManager();
		/// </remarks>
		/// <returns>
		///	Returns the singleton instance of the CacheManager object
		/// </returns>
		public static CacheManager GetCacheManager()
		{
			return cachManager;
		}

		#region Add method overloads

		/// <summary>
		///	Adds an element with the specified key and value into the cache.
		/// </summary>
		/// <remarks>
		///	Add(key, keyData);
		/// </remarks>
		/// <param name="key">
		///	The unique key used to identify the value stored
		/// </param>
		/// <param name="keyData">
		///	The value to be stored
		/// </param>
		public void Add ( string key, object keyData)
		{
			try
			{
				#region Validation for input parameters

				if( Object.Equals(key, null) )
				{
					throw new ArgumentNullException( "key",
						CacheResources.ResourceManager[ 
						"RES_ExceptionNullKey" ] );
				}
				if( key.Length == 0)
				{
					throw new ArgumentOutOfRangeException( "key",
						CacheResources.ResourceManager[ 
						"RES_ExceptionEmptyKey" ] );
				}
				if( Object.Equals(keyData, null) )
				{
					throw new ArgumentNullException( "keyData",
						CacheResources.ResourceManager[ 
						"RES_ExceptionNullValue" ] );
				}
				#endregion
				
				// Synchronize for threads
				lock(this)
				{
					// Add key and value to the cache storage.
					manageCacheStorage.Add (key, keyData);

					// Add key and default priority to CacheService for 
					// scavenging reasons
					AddMetadataToCacheService( key,
						null,
						CacheItemPriority.Normal,
						null );
				}
			}
			catch(Exception genException)
			{
				ExceptionManager.Publish(genException);
				throw;
			}
		}

		/// <summary>
		///	Adds an element with the specified key, value and metadata into 
		///	the storage.
		/// </summary>
		/// <remarks>
		///	Add(key, keyData, expirations, priority, onRemoveCallback);
		/// </remarks>
		/// <param name="key">
		///	The unique key used to identify the value stored
		/// </param>
		/// <param name="keyData">
		///	The value to be stored
		/// </param>
		/// <param name="expirations">
		///	Specifies the expirations for the item
		/// </param>
		/// <param name="priority">
		///	Specifies the item priority levels
		/// </param>
		/// <param name="onRemoveCallback">
		///	To be called when the item is removed
		/// </param>
		public void Add ( string key,
			object keyData,
			ICacheItemExpiration[] expirations,
			CacheItemPriority priority,
			CacheItemRemovedCallback onRemoveCallback )
		{
			try
			{
				#region Validation for input parameters
			
				if( Object.Equals(key, null) )
				{
					throw new ArgumentNullException( "key",
						CacheResources.ResourceManager[ 
						"RES_ExceptionNullKey" ] );
				}
				if( key.Length == 0)
				{
					throw new ArgumentOutOfRangeException( "key",
						CacheResources.ResourceManager[ 
						"RES_ExceptionEmptyKey" ] );
				}
				if( Object.Equals(keyData, null) )
				{
					throw new ArgumentNullException( "keyData",
						CacheResources.ResourceManager[ 
						"RES_ExceptionNullValue" ] );
				}

				#endregion

				// Synchronize for threads
				lock(this)
				{
					// Add the value to the store
					manageCacheStorage.Add ( key, keyData );

					// Add the metadata to the CacheService
					manageCacheService.BeginAdd( key,
						expirations,
						priority,
						onRemoveCallback, 
						null,
						null );				
				}
			}
			catch( Exception genException )
			{
				ExceptionManager.Publish( genException );
				throw;
			}
		}

		#endregion		

		#region GetData method 

		/// <summary>
		///	Gets the element with the specified key. 	
		/// </summary>
		/// <remarks>
		///	object result = GetData( key );
		/// </remarks>
		/// <param name="key">
		///	The unique key used to identify the value stored
		/// </param>
		/// <returns>
		///	The cache item value
		/// </returns>
		public object GetData ( string key )
		{
			try
			{
				#region Validation for key
				if( Object.Equals(key, null) )
				{
					throw new ArgumentNullException( "key",
						CacheResources.ResourceManager[ 
						"RES_ExceptionNullKey" ] );
				}
				if( key.Length == 0 )
				{
					throw new ArgumentOutOfRangeException( "key",
						CacheResources.ResourceManager[ 
						"RES_ExceptionEmptyKey" ] );
				}
				#endregion

				// Asynch Notify CacheService
				lock(this)
				{
					manageCacheService.BeginNotify( key,
						null,
						null );
					Object result = manageCacheStorage.GetData(key);
					return result;
				}
			}
			catch( Exception genException )
			{
				ExceptionManager.Publish( genException );
				throw;
			}			
		}	
		
		#endregion

		#region GetItem method 

		/// <summary>
		///	Gets the cache item with the specified key.
		/// </summary>
		/// <remarks>
		///	CacheItem itemValue = GetItem(key);
		/// </remarks>
		/// <param name="key">
		///	The unique key used to identify the value stored
		/// </param>
		/// <returns>
		///	The cache item value and its metadata
		/// </returns>
		public CacheItem GetItem ( string key )
		{
			try
			{
				#region Validation for key

				if( Object.Equals(key, null) )
				{
					throw new ArgumentNullException( "key",
						CacheResources.ResourceManager[ 
						"RES_ExceptionNullKey" ] );
				}
				if( key.Length == 0 )
				{
					throw new ArgumentOutOfRangeException( "key",
						CacheResources.ResourceManager[ 
						"RES_ExceptionEmptyKey" ] );
				}
		
				#endregion

				CacheItem cacheItemMetadata = null;
				
				lock(this)
				{
					// Get the item's metadata from the CacheService
					CacheItem item = manageCacheService.GetData( key );

					if(!Object.Equals(item, null))
					{
						// Get the item's data from the CacheStorage
						item.Value = manageCacheStorage.GetData( key );

						// Add the key
						item.Key = key;
						cacheItemMetadata = item;

						// Notify CacheService
						manageCacheService.BeginNotify( key,
							null,
							null );	

						return cacheItemMetadata;	
					}
					else
					{
						return null;
					}
				}
			}
			catch( Exception genException )
			{
				ExceptionManager.Publish( genException );
				throw;
			}			
		}	
		
		#endregion

		#region Remove method

		/// <summary>
		///	Removes the element with the specified key.
		/// </summary>
		/// <remarks >
		///	Remove(key);
		/// </remarks>
		/// <param name="key">
		///	The unique key used to identify the value stored
		/// </param>
		public void Remove ( string key )
		{
			try
			{
				#region Validation for key
				
				if( Object.Equals(key, null) )
				{
					throw new ArgumentNullException( "key",
						CacheResources.ResourceManager[ 
						"RES_ExceptionNullKey" ] );
				}
				if( key.Length == 0 )
				{
					throw new ArgumentOutOfRangeException( "key",
						CacheResources.ResourceManager[ 
						"RES_ExceptionEmptyKey" ] );
				}
				#endregion
				
				lock(this)
				{
					// Remove the item's data from the storage
					manageCacheStorage.Remove( key );
		
					// Call the CacheService's Asynch Remove method
					// to remove the item's expirations, dependencies
					// and its priority
					manageCacheService.BeginRemove( key,
						null,
						null );
				}
				
			}
			catch( Exception genException )
			{
				ExceptionManager.Publish( genException );
				throw;
			}
		}		
		#endregion

		#region Flush method
		/// <summary>
		///	Clears all elements from the cache.
		/// </summary>
		/// <remarks>
		///	Flush();
		/// </remarks>
		public void Flush()
		{
			try
			{
				lock(this)
				{
					// Remove all the cache items from storage
					manageCacheStorage.Flush();

					// Call the CacheService's Asynch Remove method
					// to flush the item's expirations, dependencies
					// and its priority for the cache items 
					manageCacheService.BeginFlush( 
						null,
						null );
				}
			}
			catch( Exception genException )
			{
				ExceptionManager.Publish( genException );
				throw;
			}
		}	
	
		#endregion
 
		#endregion

		#region Private Methods

		#region AddMetadataToCacheService

		/// <summary>
		///	Add the metadata to the CacheService.
		/// </summary>
		/// <remarks>
		///	AddMetadataToCacheService( key, expirations, priority, 
		///	onRemoveCallback );
		/// </remarks>
		/// <param name="key">
		///	The unique key used to identify the value stored
		/// </param>
		/// <param name="expirations">
		///	Specifies the expirations for the item
		/// </param>
		/// <param name="priority">
		///	Specifies the item priority levels
		/// </param>
		/// <param name="onRemoveCallback">
		///	To be called when the item is removed
		/// </param>
		private void AddMetadataToCacheService ( string key,
			ICacheItemExpiration[] expirations,
			CacheItemPriority priority, 
			CacheItemRemovedCallback onRemoveCallback )
		{
			try
			{
				#region Validation for the input parameters
				if( Object.Equals(key, null) )
				{
					throw new ArgumentNullException( "key",
						CacheResources.ResourceManager[ 
						"RES_ExceptionNullKey" ] );
				}
				if( key.Length == 0 )
				{
					throw new ArgumentOutOfRangeException( "key",
						CacheResources.ResourceManager[ 
						"RES_ExceptionEmptyKey" ] );
				}				
			
				#endregion

				// Call the cacheService's Asynch Add method 
				// to add the metadata
				manageCacheService.BeginAdd( key,
					expirations,
					priority,
					onRemoveCallback,
					null,
					null );
			}
			catch(Exception genException)
			{
				ExceptionManager.Publish(genException);
				throw;
			}
		}
		#endregion
		
		#region AddMetadataToCacheServiceCallback

		/// <summary>
		///	The callback method to which AsyncResult is passed when specified 
		///	as part of the begin operation of adding the metadata to the 
		///	CacheService.
		/// </summary>
		/// <remarks>
		///	AddMetadataToCacheServiceCallback( result );
		/// </remarks>
		/// <param name="result">
		///	Represents the status of an asynchronous operation
		/// </param>
		private void AddMetadataToCacheServiceCallback ( IAsyncResult result )
		{
			try
			{
				#region Validation for AsyncResult

				if( Object.Equals(result, null) )
				{
					throw new ArgumentNullException( "result",
						CacheResources.ResourceManager[ 
						"RES_ExceptionNullAsyncResult" ] );
				}
				#endregion

				// Ends the pending asynchronous add request
				manageCacheService.EndAdd(result);
			}
			catch( Exception genException )
			{
				ExceptionManager.Publish( genException );
				throw;
			}
		}

		#endregion

		#region GetCacheItemDataCallback
		
		/// <summary>
		///	The callback method to which AsyncResult is passed when specified 
		///	as part of the begin operation of gettting the data from the 
		///	CacheService.
		/// </summary>
		/// <remarks>
		///	GetCacheItemDataCallback( result );
		/// </remarks>
		/// <param name="result">
		///	Represents the status of an asynchronous operation
		/// </param>
		private void GetCacheItemDataCallback ( IAsyncResult result )
		{
			try
			{
				#region Validation for AsyncResult

				if( Object.Equals(result, null) )
				{
					throw new ArgumentNullException( "result",
						CacheResources.ResourceManager[ 
						"RES_ExceptionNullAsyncResult" ] );
				}
				#endregion

				// Ends a pending asynchronous notify request
				manageCacheService.EndNotify(result);
			}
			catch( Exception genException )
			{
				ExceptionManager.Publish( genException );
				throw;
			}
		}	
		
		#endregion

		#region GetCacheItemCallback
		
		/// <summary>
		///	The callback method to which AsyncResult is passed when specified 
		///	as part of the begin operation of gettting the metadata from the 
		///	CacheService.
		/// </summary>
		/// <remarks>
		///	GetCacheItemCallback( result );
		/// </remarks>
		/// <param name="result">
		///	Represents the status of an asynchronous operation
		/// </param>
		private void GetCacheItemCallback ( IAsyncResult result )
		{
			try
			{
				#region Validation for AsyncResult

				if( Object.Equals(result, null) )
				{
					throw new ArgumentNullException( "result",
						CacheResources.ResourceManager[ 
						"RES_ExceptionNullAsyncResult" ] );
				}
				#endregion

				// Ends a pending asynchronous notify request
				manageCacheService.EndNotify(result);
			}
			catch( Exception genException )
			{
				ExceptionManager.Publish( genException );
				throw;
			}
		}	
		
		#endregion

		#region RemoveCacheItemMetadataCallback
		
		/// <summary>
		///	The callback method to which AsyncResult is passed when specified 
		///	as part of the begin operation of removing the metadata from the 
		///	CacheService.
		/// </summary>
		/// <remarks>
		///	RemoveCacheItemMetadataCallback( result );
		/// </remarks>
		/// <param name="result">
		///	Represents the status of an asynchronous operation
		/// </param>
		private void RemoveCacheItemMetadataCallback ( IAsyncResult result )
		{
			try
			{
				#region Validation for AsyncResult

				if( Object.Equals(result, null) )
				{
					throw new ArgumentNullException( "result",
						CacheResources.ResourceManager[ 
						"RES_ExceptionNullAsyncResult" ] );
				}

				#endregion

				// Ends a pending asynchronous remove request
				manageCacheService.EndRemove(result);
			}
			catch( Exception genException )
			{
				ExceptionManager.Publish( genException );
				throw;
			}
		}
		
		#endregion

		#region FlushCacheCallback
		
		/// <summary>
		///	The callback method to which AsyncResult is passed when specified 
		///	as part of the begin operation of flushing the metadata from the 
		///	CacheService.
		/// </summary>
		/// <remarks>
		///	FlushCacheCallback( result );
		/// </remarks>
		/// <param name="result">
		///	Represents the status of an asynchronous operation
		/// </param>
		private void FlushCacheCallback ( IAsyncResult result )
		{
			try
			{
				#region Validation for AsyncResult

				if( Object.Equals(result, null) )
				{
						throw new ArgumentNullException( "result",
						CacheResources.ResourceManager[ 
						"RES_ExceptionNullAsyncResult" ] );
				}

				#endregion

				// Ends a pending asynchronous flush request
				manageCacheService.EndFlush(result);
			}
			catch( Exception genException )
			{
				ExceptionManager.Publish( genException );
				throw;
			}
		}
		
		#endregion

		#endregion
	}
}