//===============================================================================
// Microsoft Caching Application Block for .NET
// http://msdn.microsoft.com/library/default.asp?url=/library/en-us/dnpag/html/CachingBlock.asp
//
// CacheService.cs
// This class controls the cache items lifetime for an instance.
//
//===============================================================================
// Copyright (C) 2003 Microsoft Corporation
// All rights reserved.
// THIS CODE AND INFORMATION IS PROVIDED "AS IS" WITHOUT WARRANTY
// OF ANY KIND, EITHER EXPRESSED OR IMPLIED, INCLUDING BUT NOT
// LIMITED TO THE IMPLIED WARRANTIES OF MERCHANTABILITY AND/OR
// FITNESS FOR A PARTICULAR PURPOSE.
//===============================================================================


using System;
using System.Collections;
using System.Configuration;
using System.Threading;
using System.Collections.Specialized;
using System.Runtime.Remoting;
using System.Runtime.Remoting.Contexts;
using System.Runtime.Remoting.Messaging;

// Used for handling Exceptions
using Microsoft.ApplicationBlocks.ExceptionManagement;
using Microsoft.ApplicationBlocks.Cache.Expirations;

namespace Microsoft.ApplicationBlocks.Cache
{
	/// <summary>
	///	This class controls the cache items lifetime.
	/// </summary>
	public sealed class CacheService : MarshalByRefObject
	{
		#region Private members

		ICacheStorage cacheStorage;
		ICacheMetadata cacheMetadata;		
		IMmfReference mmfReference;
		IScavengingAlgorithm storageScavengingImplementation;

		private HybridDictionary itemsExpiration;
		internal HybridDictionary cacheItemsPriority;
		private HybridDictionary itemsOnRemoveCallbacks;
		private HybridDictionary itemsMmfHandles;

		// Creating a static instance
		private static CacheService cachService = new CacheService();

		#endregion

		#region Private constants
		private static readonly int CONVERT_TO_MILLISECONDS_VALUE = 1000;			
		#endregion

		#region InitializeLifetimeService
		/// <summary>
		///	Obtains a lifetime service object to control the lifetime policy 
		///	for this instance.
		/// </summary>
		public override object InitializeLifetimeService() 
		{ 
			return null; 
		}

		#endregion

		#region Constructor
		/// <summary>
		///	Constructor to initialize the scavenging and expiration process.
		/// </summary>
		private CacheService()
		{
			string mode;
			// Init the metadata local stores
			itemsExpiration = new HybridDictionary(false);
			cacheItemsPriority = new HybridDictionary(false);
			itemsOnRemoveCallbacks = new HybridDictionary(false);
			itemsMmfHandles = new HybridDictionary(false);

			try
			{
				if(! Object.Equals(
					CacheConfiguration.Config.StorageInformation, null) )
				{
					// Create a pointer to the required storage
					cacheStorage = 
						CacheConfiguration.Config.StorageInformation.
						CreateInstance();

					// Get the mode from the config file
					mode = CacheConfiguration.Config.StorageInformation[ 
						CacheResources.ResourceManager[ "RES_Mode" ] ];

					if(! (Object.Equals(mode, null) || (mode.Length == 0)) )
					{
						if(Object.Equals(mode.ToLower(), CacheResources.
							ResourceManager["RES_InProc"]))
						{
							cacheStorage = CacheManager.CacheStorage;
						}
						else if( !object.Equals(mode.ToLower(), CacheResources.
							ResourceManager["RES_OutProc"]))
						{
							throw(new ConfigurationException(CacheResources.
								ResourceManager["RES_ExceptionInvalidMode"]));
						}
					}
					else
					{
						throw(new ConfigurationException(CacheResources.
							ResourceManager["RES_ExceptionNullOrEmptyMode"]));
					}
				}
				else 
				{
					throw new ConfigurationException(CacheResources.
						ResourceManager["RES_ConfigurationInfoUnavailable"]);
				}
            
				if(! Object.Equals(
					CacheConfiguration.Config.ScavengingInformation, null) )
				{
					// Initialize scavenging process
					storageScavengingImplementation = 
						CacheConfiguration.Config.ScavengingInformation.
						CreateInstance(
						this,
						cacheStorage, 
						cacheMetadata ); 
				}
				else
				{
					// Initialize default scavenging process
					storageScavengingImplementation = 
						new Scavenging.LruScavenging(); 
					storageScavengingImplementation.Init( this, 
						cacheStorage,
						null,
						null );
				}

				// Load the meta data from the persistent store
				LoadMetadataFromStorage();

				// Get the Mmf reference if cache storage 
				// implements IMmfReference
				mmfReference = cacheStorage as IMmfReference;
			
				// Start a background thread and start monitoring
				// for expirations
				ThreadStart expirationThreadStart = new ThreadStart(
					this.MonitorForExpirations);
				Thread expirationThread = new Thread(expirationThreadStart);
				expirationThread.IsBackground = true;
				expirationThread.Start();
			}
			catch(Exception genException)
			{
				ExceptionManager.Publish(genException);
				throw;
			}
		}

		/// <summary>
		///	This method is used to return the singleton instance of the 
		///	CacheService object.
		/// </summary>
		/// <remarks>
		///	CacheService service = CacheService.GetCacheService();
		/// </remarks>
		/// <returns>
		///	Returns the singleton instance of the CacheService object
		/// </returns>
		public static CacheService GetCacheService()
		{
			return cachService;
		}

		#endregion

		#region Internal Methods

		#region Add Functions

		#region Add

		/// <summary>
		///	Adds a new item with the specified metadata to the cache service.  
		///	If a CacheItemRemovedCallback is specified, then it will be called 
		///	when the item is removed.
		/// </summary>
		/// <remarks>
		///	Add( key, expirations, priority, onRemoveCallback );
		///	</remarks>
		///	<param name="key">
		///	The unique key used to identify the value stored
		/// </param>
		/// <param name="expirations">
		///	Specifies the expirations for the item
		/// </param>
		/// <param name="priority">
		///	Specifies the item priority levels
		/// </param>
		/// <param name="onRemoveCallback">
		///	To be called when the item is removed
		/// </param>
		public void Add ( string keyVal,
			ICacheItemExpiration[] expirations,
			CacheItemPriority priority,
			CacheItemRemovedCallback onRemoveCallback )
		{
			try
			{
				#region Validation for the input parameters
			
				if( Object.Equals(keyVal, null) )
				{
					throw new ArgumentNullException( "keyVal",
						CacheResources.ResourceManager[ 
						"RES_ExceptionNullKey" ] );
				}
				if( Object.Equals(keyVal, String.Empty) )
				{
					throw new ArgumentOutOfRangeException( "keyVal",
						CacheResources.ResourceManager[ 
						"RES_ExceptionEmptyKey" ] );
				}
				if(! Object.Equals(expirations, null) )
				{
					for( int counter = 0;
						counter < expirations.Length;
						counter ++ )
					{
						if( Object.Equals(expirations[counter],	null) )
						{
							throw new ArgumentNullException( 
								"expirations array item",
								CacheResources.ResourceManager[ 
								"RES_ExceptionExpirationsItem" ] );
						}
					}
				}
				
				#endregion

				#region Remove existing metadata

				// Check if the item's metadata exists, if it does - remove
				// the previous item's metadata
				lock( itemsExpiration )
				{
					itemsExpiration.Remove(keyVal);
				}
				lock( cacheItemsPriority )
				{
					cacheItemsPriority.Remove(keyVal);
				}
				lock( itemsOnRemoveCallbacks )
				{
					itemsOnRemoveCallbacks.Remove(keyVal);
				}
				lock( itemsMmfHandles)
				{
					itemsMmfHandles.Remove(keyVal);
				}

				#endregion

				#region Add new metadata

				// If the storage implements the IMmfReference
				// interface - Add the MMFS object in the hash table
				if(mmfReference != null)
				{
					lock(itemsMmfHandles)
					{
						itemsMmfHandles.Add(keyVal, 
							mmfReference.AddReference(keyVal));
					}
				}

				// Insert the new metadata
				if (! Object.Equals(expirations, null) )
				{
					// Insert the key to each expiration class provided
					for ( int counter = 0; 
						counter < expirations.Length;
						counter++ )
					{
						expirations[counter].Key(keyVal);

						// Register to the OnChange event
						expirations[counter].Change += 
							new ItemDependencyChangeEventHandler( 
							this.ItemDependencyChanged );
					}
					// Insert the array to the expirations collection
					lock( itemsExpiration )
					{
						itemsExpiration.Add(keyVal, expirations);
					}
				}

				if (! Object.Equals(onRemoveCallback, null) )
				{
					lock( itemsOnRemoveCallbacks )
					{
						itemsOnRemoveCallbacks[keyVal] = onRemoveCallback;
					}
				}

				lock( cacheItemsPriority )
				{
					cacheItemsPriority.Add(keyVal, priority);
				}

				// If the storage implements the ICacheMetadata 
				// interface - Add the Metadata to it as well
				if (! Object.Equals(cacheMetadata, null) )
				{
					if (cacheMetadata is ICacheMetadataWithCallback) {
						(((ICacheMetadataWithCallback)cacheMetadata)).Add(keyVal, expirations, priority, onRemoveCallback);
					} else {
						cacheMetadata.Add(keyVal, expirations, priority);
					}					
				}

				// Add the new key to the scavenging class
				storageScavengingImplementation.Add(keyVal, priority);

				// Notifying the addition of key to Cacheservice 
				// so that the Cacheservice will take the time the
				// data is added as the last time used for calculating
				// the sliding expirations
				Notify(keyVal);

				storageScavengingImplementation.Execute();					
				
				#endregion
			}
			catch( Exception genException )
			{
				ExceptionManager.Publish( genException );
			}
		}

		#endregion
	
		#region BeginAdd
		
		/// <summary>
		///	Begins an asynchronous request for adding a new item.
		/// </summary>
		/// <remarks>
		///	IAsyncResult result = BeginAdd( key, expirations, priority, 
		///	onRemoveCallback, callback, asyncState );
		/// </remarks>
		/// <param name="key">
		///	The unique key used to identify the value stored
		/// </param>
		/// <param name="expirations">
		///	Specifies the expirations for the item
		/// </param>
		/// <param name="priority">
		///	Specifies the item priority levels
		/// </param>
		/// <param name="onRemoveCallback">
		///	To be called when the item is removed
		/// </param>
		/// <param name="callback">
		///	Method to be called when Async operation is completed
		/// </param>
		/// <param name="asyncState">
		///	Property to get the last parameter of a BeginInvoke method call
		/// </param>
		/// <returns>
		///	The status of an asynchronous operation
		/// </returns>
		public void BeginAdd ( string key,
			ICacheItemExpiration[] expirations,
			CacheItemPriority priority,
			CacheItemRemovedCallback onRemoveCallback,
			AsyncCallback callback, 
			object asyncState )
		{
			try
			{
				#region Validation for the input parameters
			
				if( Object.Equals(key, null) )
				{
					throw new ArgumentNullException( "key",
						CacheResources.ResourceManager[ 
						"RES_ExceptionNullKey" ] );
				}
				if( Object.Equals(key, String.Empty) )
				{
					throw new ArgumentOutOfRangeException( "key",
						CacheResources.ResourceManager[ 
						"RES_ExceptionEmptyKey" ] );
				}

				#endregion

				AddDataHandler handleAdd = new AddDataHandler( this.Add );
				
				IAsyncResult asyncResult = handleAdd.BeginInvoke( key,
					expirations,
					priority, 
					onRemoveCallback , 
					new AsyncCallback(EndAdd), 
					asyncState );

				WaitHandle handle = asyncResult.AsyncWaitHandle;
				handle.WaitOne();
			}
			catch( Exception genException )
			{
				ExceptionManager.Publish( genException );
			}		
		}

		#endregion

		#region EndAdd
		
		/// <summary>
		///	Ends a pending asynchronous add request.
		/// </summary>
		/// <remarks>
		///	EndAdd( asyncResult );
		/// </remarks>
		/// <param name="asyncResult">
		///	Represents the status of an asynchronous operation
		/// </param>
		internal void EndAdd ( IAsyncResult asyncResult )
		{
			try
			{
				#region Validation for AsyncResult
			
				if( Object.Equals(asyncResult, null) )
				{
					throw new ArgumentNullException( "asyncResult",
						CacheResources.ResourceManager[ 
						"RES_ExceptionNullAsyncResult" ] );
				}
				
				#endregion

				AsyncResult result = (AsyncResult)asyncResult;
				AddDataHandler handleAdd = 
					(AddDataHandler)result.AsyncDelegate;
				handleAdd.EndInvoke( asyncResult );
				
			}
			catch( Exception genException )
			{
				ExceptionManager.Publish( genException );
			}
		}

		#endregion

		#endregion

		#region Get Functions
		
		/// <summary>
		///	Gets a item with the specified key to the cache service.
		/// </summary>
		/// <remarks>
		///	CacheItem Get(string key);
		/// </remarks>
		/// <param name="key">
		///	The unique key used to identify the value and metadata stored  
		/// </param>
		/// <returns>
		///	The cache item metadata
		/// </returns>
		public CacheItem GetData ( string key )
		{
			try
			{
				#region Check for valid key
			
				if( Object.Equals(key, null) )
				{
					throw new ArgumentNullException( "key",
						CacheResources.ResourceManager[ 
						"RES_ExceptionNullKey" ] );
				}
				if( Object.Equals(key, String.Empty) )
				{
					throw new ArgumentOutOfRangeException( "key",
						CacheResources.ResourceManager[ 
						"RES_ExceptionEmptyKey" ] );
				}
				
				#endregion

				// Create a new CacheItem class
				CacheItem cacheItemMetadata = new CacheItem();

				// Insert expirations, dependencies and priority
				cacheItemMetadata.Expirations = 
					(ICacheItemExpiration[])itemsExpiration[ key ];
				
				if(cacheItemsPriority.Contains(key))
				{
					cacheItemMetadata.Priority = 
						(CacheItemPriority)cacheItemsPriority[ key ];
				}

				return cacheItemMetadata;
			}
			catch( Exception genException )
			{
				ExceptionManager.Publish( genException );
			}	
			return null;
		}

		#endregion

		#region Flush Functions

		#region Flush
		
		/// <summary>
		///	Clears all items from the cache service.
		/// </summary>
		/// <remarks>
		///	Flush();
		/// </remarks>
		public void Flush ()
		{
			try
			{
				// Flush everything, If the storage implements ICacheMetadata
				// flush it as well
				lock( itemsExpiration )
				{
					itemsExpiration.Clear();
				}
				storageScavengingImplementation.Flush();

				if (! Object.Equals(cacheMetadata, null) )
				{
					cacheMetadata.Flush();
				}
				lock( itemsOnRemoveCallbacks )
				{
					itemsOnRemoveCallbacks.Clear();
				}
				// If the cache storage implements IMmfReference,
				// during flush close all the memory mapped objects
				// and clear the HybridDictionary
				if(mmfReference != null)
				{
					lock(itemsMmfHandles)
					{
						foreach (DictionaryEntry entry in itemsMmfHandles)
						{
							if(entry.Value != null)
							{
								mmfReference.RemoveReference(entry.Value);
							}
						}
						itemsMmfHandles.Clear();
					}
				}
			}
			catch( Exception genException )
			{
				ExceptionManager.Publish( genException );
			}
		}

		#endregion
	
		#region BeginFlush
		
		/// <summary>
		///	Begins an asynchronous request for clear all items from the cache 
		///	service.
		/// </summary>
		/// <remarks>
		///	IAsyncResult result = BeginFlush( callback, asyncState );
		/// </remarks>
		/// <param name="callback">
		///	Method to be called when Async operation is completed
		/// </param>
		/// <param name="asyncState">
		///	Property to get the last parameter of a BeginInvoke method call
		/// </param>
		/// <returns>
		///	The status of an asynchronous operation
		/// </returns>
		public void BeginFlush ( 
			AsyncCallback callback,
			object asyncState )
		{
			try
			{
				FlushHandler handleFlush = new FlushHandler( this.Flush );
				handleFlush.BeginInvoke( 
					new AsyncCallback(EndFlush), 
					asyncState );
			}
			catch( Exception genException )
			{
				ExceptionManager.Publish( genException );
			}			
		}

		#endregion

		#region EndFlush
		
		/// <summary>
		///	Ends a pending asynchronous flush request.
		/// </summary>
		/// <remarks>
		///	EndFlush( asyncResult );
		/// </remarks>
		/// <param name="asyncResult">
		///	Represents the status of an asynchronous operation
		/// </param>
		internal void EndFlush ( IAsyncResult asyncResult  )
		{
			try
			{
				#region Validation for AsyncResult
			
				if( Object.Equals(asyncResult, null) )
				{
					throw new ArgumentNullException( "asyncResult",
						CacheResources.ResourceManager[ 
						"RES_ExceptionNullAsyncResult" ] );
				}
				#endregion

				AsyncResult result = (AsyncResult)asyncResult;
				FlushHandler handleFlush = 
					(FlushHandler)result.AsyncDelegate;
				handleFlush.EndInvoke( asyncResult );				
			}
			catch( Exception genException )
			{
				ExceptionManager.Publish( genException );
			}
		}
		#endregion

		#endregion

		#region Notify Functions

		#region Notify

		/// <summary>
		///	Notifies that the item with the specified key was recently used.
		/// </summary>
		/// <remarks>
		///	Notify( key );
		/// </remarks>
		/// <param name="key">
		///	The unique key used to identify the value and the metadata stored
		/// </param>
		public void Notify ( string key )
		{
			try
			{
				#region Check for valid key

				if( Object.Equals(key, null) )
				{
					throw new ArgumentNullException( "key",
						CacheResources.ResourceManager[ 
						"RES_ExceptionNullKey" ] );
				}
				if( Object.Equals(key, String.Empty) )
				{
					throw new ArgumentOutOfRangeException( "key",
						CacheResources.ResourceManager[ 
						"RES_ExceptionEmptyKey" ] );
				}
				
				#endregion

				// Notify the item's expiration classes and
				// the scavenging class
				storageScavengingImplementation.Notify( key );

				if ( itemsExpiration.Contains(key) )
				{			
					ICacheItemExpiration[] itemExpirations = null; 
					itemExpirations = 
						(ICacheItemExpiration[])itemsExpiration[ key ];

					for( int counter = 0; 
						counter < itemExpirations.Length; 
						counter++ )
					{
						itemExpirations[counter].Notify();
					}
				}
			}
			catch( Exception genException )
			{
				ExceptionManager.Publish( genException );
			}
		}
		#endregion
	
		#region BeginNotify

		/// <summary>
		///	Begins an asynchronous request for notify that a item was recently 
		///	used.
		/// </summary>
		/// <remarks>
		///	IAsyncResult result = BeginNotify( key, callback, asyncState );
		/// </remarks>
		/// <param name="key">
		///	The unique key used to identify the value and the metadata stored 
		/// </param>
		/// <param name="callback">
		///	Method to be called when Async operation is completed
		/// </param>
		/// <param name="asyncState">
		///	Property to get the last parameter of a BeginInvoke method call
		/// </param>
		/// <returns>
		///	The status of an asynchronous operation
		/// </returns>
		public void BeginNotify ( string key, 
			System.AsyncCallback callback, 
			object asyncState )
		{
			try
			{
				#region Validation for input parameters
			
				if( Object.Equals(key, null) )
				{
					throw new ArgumentNullException( "key",
						CacheResources.ResourceManager[ 
						"RES_ExceptionNullKey" ] );
				}
				if( key.Length == 0 )
				{
					throw new ArgumentOutOfRangeException( "key",
						CacheResources.ResourceManager[ 
						"RES_ExceptionEmptyKey" ] );
				}

				#endregion

				NotifyHandler handleNotify = new NotifyHandler( this.Notify );
				handleNotify.BeginInvoke( key, 
					new AsyncCallback(EndNotify), 
					asyncState );
			}
			catch( Exception genException )
			{
				ExceptionManager.Publish( genException );
			}		
		}
		#endregion

		#region EndNotify

		/// <summary>
		///	Ends a pending asynchronous notify request.
		/// </summary>
		/// <remarks>
		///	EndNotify( asyncResult );
		/// </remarks>
		/// <param name="asyncResult">
		///	Represents the status of an asynchronous operation
		/// </param>
		internal void EndNotify ( IAsyncResult asyncResult )
		{
			try
			{
				#region Validation for AsyncResult
			
				if( Object.Equals(asyncResult, null) )
				{
					throw new ArgumentNullException( "asyncResult",
						CacheResources.ResourceManager[ 
						"RES_ExceptionNullAsyncResult" ] );
				}
				#endregion

				AsyncResult result = (AsyncResult)asyncResult;
				NotifyHandler handleNotify = 
					(NotifyHandler)result.AsyncDelegate;
				handleNotify.EndInvoke( asyncResult );
			}
			catch( Exception genException )
			{
				ExceptionManager.Publish( genException );
			}
		}
		#endregion

		#endregion

		#region Remove Functions

		#region Remove

		/// <summary>
		///	Removes the item with the specified key from the cache service.
		/// </summary>
		/// <remarks>
		///	Remove( key );
		/// </remarks>
		/// <param name="key">
		///	The unique key used to identify the value and the metadata stored 
		/// </param>
		public void Remove ( string key )
		{
			try
			{
				#region Check for valid key

				if( Object.Equals(key, null) )
				{
					throw new ArgumentNullException( "key",
						CacheResources.ResourceManager[ 
						"RES_ExceptionNullKey" ] );
				}
				if( Object.Equals(key, String.Empty) )
				{
					throw new ArgumentOutOfRangeException( "key",
						CacheResources.ResourceManager[ 
						"RES_ExceptionEmptyKey" ] );
				}
				#endregion

				// Remove all metadata from the CacheService, if the storage
				// implements ICacheMetadata remove item's metadata from there
				// as well
				RemoveItem( key, CacheItemRemoveCause.Removed );
			}
			catch( Exception genException )
			{
				ExceptionManager.Publish( genException );
			}
		}
		#endregion
	
		#region BeginRemove

		/// <summary>
		///	Begins an asynchronous request for remove a item.
		/// </summary>
		/// <remarks>
		///	IAsyncResult result = BeginRemove( key, callback, asyncState );
		/// </remarks>
		/// <param name="key">
		///	The unique key used to identify the value and the metadata stored 
		/// </param>
		/// <param name="callback">
		///	Method to be called when Async operation is completed
		/// </param>
		/// <param name="asyncState">
		///	Property to get the last parameter of a BeginInvoke method call
		/// </param>
		/// <returns>
		///	The status of an asynchronous operation
		/// </returns>
		public void BeginRemove( string key, 
			System.AsyncCallback callback, 
			object asyncState )
		{
			try
			{
				#region Validation for input parameters
			
				if( Object.Equals(key, null) )
				{
					throw new ArgumentNullException( "key",
						CacheResources.ResourceManager[ 
						"RES_ExceptionNullKey" ] );
				}
				if( key.Length == 0 )
				{
					throw new ArgumentOutOfRangeException( "key",
						CacheResources.ResourceManager[ 
						"RES_ExceptionEmptyKey" ] );
				}

				#endregion

				RemoveDataHandler handleRemove = 
					new RemoveDataHandler( this.Remove );
				handleRemove.BeginInvoke( 
					key, 
					new AsyncCallback(EndRemove),
					asyncState );
			}
			catch( Exception genException )
			{
				ExceptionManager.Publish( genException );
			}			
		}
		#endregion

		#region EndRemove

		/// <summary>
		///	Ends a pending asynchronous remove request.
		/// </summary>
		/// <remarks>
		///	EndRemove( asyncResult );
		/// </remarks>
		/// <param name="asyncResult">
		///	Represents the status of an asynchronous operation
		/// </param>
		internal void EndRemove ( IAsyncResult asyncResult )
		{
			try
			{
				#region Validation for AsyncResult
			
				if( Object.Equals(asyncResult, null) )
				{
					throw new ArgumentNullException( "asyncResult",
						CacheResources.ResourceManager[ 
						"RES_ExceptionNullAsyncResult" ] );
				}
				
				#endregion

				AsyncResult result = (AsyncResult)asyncResult;
				RemoveDataHandler handleRemove = 
					(RemoveDataHandler)result.AsyncDelegate;
				handleRemove.EndInvoke( asyncResult );
			}
			catch( Exception genException )
			{
				ExceptionManager.Publish( genException );
			}
		}
		#endregion		

		#region RemoveItem

		/// <summary>
		///	Remove the metadata from the local store.
		/// </summary>
		/// <remarks>
		///	RemoveItem( key, cause );
		/// </remarks>
		/// <param name="key">
		///	The unique key to identify the value and the metadata of a cache 
		///	item
		/// </param>
		/// <param name="cause">
		///	Specifies the causes by which an item was removed from the cache
		/// </param>
		internal void RemoveItem ( string key, CacheItemRemoveCause cause )
		{
			try
			{
				#region Check for valid key

				if( Object.Equals(key, null) )
				{
					throw new ArgumentNullException( "key",
						CacheResources.ResourceManager[ 
						"RES_ExceptionNullKey" ] );
				}
				if( Object.Equals(key, String.Empty) )
				{
					throw new ArgumentOutOfRangeException( "key",
						CacheResources.ResourceManager[ 
						"RES_ExceptionEmptyKey" ] );
				}
			
				#endregion

				// Remove the expirations
				if( itemsExpiration.Contains(key) )
				{
					lock( itemsExpiration )
					{
						itemsExpiration.Remove(key);
					}
				}
				// Remove the metadata
				if (! Object.Equals(cacheMetadata, null) )
				{ 
					cacheMetadata.Remove( key );
				}

				// Remove the reference to the mmf object
				if(mmfReference != null)
				{
					if (itemsMmfHandles.Contains(key))
					{
						lock(itemsMmfHandles)
						{
							mmfReference.RemoveReference(itemsMmfHandles[key]);
							itemsMmfHandles.Remove(key);
						}
					}
				}

				// Remove the Priority
				lock( cacheItemsPriority )
				{
					cacheItemsPriority.Remove( key );
				}
				// Remove the scavenging implementations
				storageScavengingImplementation.Remove( key );

				// Remove the callback methods 
				if ( itemsOnRemoveCallbacks.Contains( key ) )
				{
					CacheItemRemovedCallback callBack = 
						(CacheItemRemovedCallback)itemsOnRemoveCallbacks[ key ];

					// Call the callback only for inproc
					
					callBack(key, cause);
					
					lock( itemsOnRemoveCallbacks )
					{
						itemsOnRemoveCallbacks.Remove( key );
					}
				}
			}
			catch( Exception genException )
			{
				ExceptionManager.Publish( genException );
			}
		}
		#endregion

		#endregion

		#endregion

		#region Private functions

		#region MonitorForExpirations

		/// <summary>
		///	Monitoring the expirations of the cache items.
		/// </summary>
		/// <remarks>
		///	MonitorForExpirations();
		/// </remarks>
		private void MonitorForExpirations()
		{
			try
			{
				int checkIntervalInSeconds;
				int checkIntervalInMilliseconds;
				int counter;
				bool isExpired;	
				string key;
				ArrayList expiredItems = new ArrayList();

				if(! Object.Equals(
					CacheConfiguration.Config.ExpirationInformation, null) )
				{
					// Get expiration check interval from config file
					checkIntervalInSeconds = 
						CacheConfiguration.Config.ExpirationInformation.
						Interval;
					checkIntervalInMilliseconds = 
						checkIntervalInSeconds * CONVERT_TO_MILLISECONDS_VALUE;
				}
				else
				{
					checkIntervalInSeconds = Convert.ToInt32(
						CacheResources.ResourceManager[ 
						"RES_DefaultExpirationInterval" ] );
					checkIntervalInMilliseconds = 
						checkIntervalInSeconds * CONVERT_TO_MILLISECONDS_VALUE;
				}
		
				while ( true )
				{
					// The use of enumerations is not a thread safe operation
					lock( itemsExpiration )
					{
						// Iterate over the expirations list and
						// check for expired items

						foreach ( DictionaryEntry dictionary in 
							itemsExpiration )
						{
							ICacheItemExpiration[] exp = 
								(ICacheItemExpiration[])dictionary.Value;													
							
							counter = 0;
							isExpired = false;
							while ( !isExpired && counter < exp.Length )
							{
								if( exp[counter].HasExpired() )
								{
									isExpired = true;
									key = dictionary.Key.ToString();
									expiredItems.Add( key );
								}
								counter++;
							}						
						}
					}					
					for ( counter = 0; 
						counter < expiredItems.Count;
						counter ++ )
					{
						cacheStorage.Remove(expiredItems[counter].ToString());
						RemoveItem(expiredItems[counter].ToString(), 
							CacheItemRemoveCause.Expired);
					}
					// Clear the content of expired items array list
					expiredItems.Clear();
					Thread.Sleep(checkIntervalInMilliseconds);
				}
			}
			catch( Exception genException )
			{
				ExceptionManager.Publish( genException );
				throw;
			}
		}
		#endregion
		
		#region ItemDependencyChanged

		/// <summary>
		///	Raising dependency change event invokes the event handler through 
		///	the ItemDependencyChange delegate. 
		/// </summary>
		/// <remarks>
		///	ItemDependencyChanged ( key );
		/// </remarks>
		/// <param name="key">
		///	The unique key to identify the value and the metadata stored
		/// </param>
		private void ItemDependencyChanged ( object sender, 
			CacheEventArgs e  )
		{
			try
			{
				#region Check for valid key

				string key = e.KeyValue;

				if( Object.Equals(key, null) )
				{
					throw new ArgumentNullException( "key",
						CacheResources.ResourceManager[ 
						"RES_ExceptionNullKey" ] );
				}
				if( Object.Equals(key, String.Empty) )
				{
					throw new ArgumentOutOfRangeException( "key",
						CacheResources.ResourceManager[ 
						"RES_ExceptionEmptyKey" ] );
				}
				#endregion

				cacheStorage.Remove(key);

				RemoveItem( key, CacheItemRemoveCause.Expired );
			}
			catch( Exception genException )
			{
				ExceptionManager.Publish( genException );
			}
		}
		#endregion
		
		#region LoadMetadataFromStorage

		/// <summary>
		///	Get the metadata of cache items from local store.
		/// </summary>
		/// <remarks>
		///	LoadMetadataFromStorage();
		/// </remarks>
		private void LoadMetadataFromStorage()
		{
			try
			{
				cacheMetadata = cacheStorage as ICacheMetadata;
				if (! Object.Equals(cacheMetadata, null) )
				{
					Hashtable metadata = cacheMetadata.GetMetadata();

					foreach ( DictionaryEntry entry in metadata )
					{
						// Check types and load expirations, priorities
						// and scavenging classes
						
						DictionaryEntry metaEntry = 
							(DictionaryEntry)entry.Value;

						if (metaEntry.Value == null) {
							continue;
						}

						if ( object.Equals(metaEntry.Value.GetType(), 
							typeof(ICacheItemExpiration[])) )
						{
							if ( !(itemsExpiration.Contains(metaEntry.Key)) )
							{
								itemsExpiration.Add( metaEntry.Key.ToString(),
									(ICacheItemExpiration[])metaEntry.Value );
							}
						}
						if ( Object.Equals(metaEntry.Value.GetType(), 
							typeof(CacheItemPriority)) )
						{
							cacheItemsPriority.Add( metaEntry.Key,
								(CacheItemPriority)metaEntry.Value );

							storageScavengingImplementation.Add(
								metaEntry.Key.ToString(),
								(CacheItemPriority)metaEntry.Value );
						}
						if (cacheMetadata is ICacheMetadataWithCallback) {
							if ( Object.Equals(metaEntry.Value.GetType(), 
								typeof(CacheItemRemovedCallback)) ) {
								itemsOnRemoveCallbacks.Add(metaEntry.Key.ToString(), (CacheItemRemovedCallback)metaEntry.Value);
							}
						}						
					}
				}
			}
			catch( Exception genException )
			{
				ExceptionManager.Publish( genException );
				throw;
			}
		}

		#endregion

		#endregion
	}
}