//===============================================================================
// Microsoft Caching Application Block for .NET
// http://msdn.microsoft.com/library/default.asp?url=/library/en-us/dnpag/html/CachingBlock.asp
//
// DefaultDataProtection.cs
// Sample data protection provider.  This provider uses DPAPI to encrypt items
// and SHA algorithm to validate them.
//
//===============================================================================
// Copyright (C) 2003 Microsoft Corporation
// All rights reserved.
// THIS CODE AND INFORMATION IS PROVIDED "AS IS" WITHOUT WARRANTY
// OF ANY KIND, EITHER EXPRESSED OR IMPLIED, INCLUDING BUT NOT
// LIMITED TO THE IMPLIED WARRANTIES OF MERCHANTABILITY AND/OR
// FITNESS FOR A PARTICULAR PURPOSE.
//===============================================================================


using System;
using System.Xml;
using System.Runtime.InteropServices; 
using System.Security.Cryptography;

// Used for all the caching operations
using Microsoft.ApplicationBlocks.Cache;
// Used for Handling Exceptions
using Microsoft.ApplicationBlocks.ExceptionManagement;

namespace Microsoft.ApplicationBlocks.Cache.DataProtection
{
	#region DataProtection provider
	/// <summary>
	///	Sample data protection provider. This provider uses DPAPI to encrypt 
	///	items and SHA algorithm to validate them.
	/// </summary>
	/// <remarks>
	///	This provider uses the following attributes on the XML file:
	/// <list type="">
	///	<item>
	///	<b>Assembly</b>
	///	Used to specify the provider assembly (Required)
	///	</item>
	///	<item>
	///	<b>ClassName</b> 
	///	Used to specify the provider class name (Required)
	///	</item>
	///	<item>
	///	<b>ValidationKey</b> 
	///	Used to specify validation key (Optional). 
	///	If this attribute isnt specified, then the
	///	provider uses a random key.
	///	</item>
	///	<item>
	///	<b>Validation</b> 
	///	Used to specify the typeof encryption used for
	///	data validation (Optional). 
	///	If this attribute isnt specified, then the provider
	///	uses a SHA1 algorithm.
	///	</item>
	/// </list>
	/// </remarks>
	public class DefaultDataProtection : IDataProtection
	{
		#region Private members

		Win32DPAPI win32DataProtectionAPI;
		KeyedHashAlgorithm validationAlgorithm;
		
		#endregion
		
		#region Public constructor
		
		public DefaultDataProtection()
		{}
		
		#endregion
		
		#region Implementation of IDataProtection

		#region Init
		
		/// <summary>
		///	Inits the data protection provider.
		/// </summary>
		/// <remarks>
		///	IDataProtection.Init( configSection );
		/// </remarks>
		/// <param name="configSection">
		///	Represents the "DataProtectionInfo" node in the app.config file
		/// </param>
		void IDataProtection.Init( XmlNode configSection )
		{
			try
			{
				// Variable Declarations
				string validationKey = null ;
				string validation = null;

				#region Gets the validation key	

				if(! Object.Equals(configSection, null) && 
					(! Object.Equals(configSection.Attributes[ 
					CacheResources.ResourceManager[ 
						"RES_ValidationKey" ] ], null) ) )
				{
					validationKey = configSection.Attributes[ 
						CacheResources.ResourceManager[ 
							"RES_ValidationKey" ] ].Value ;			
				}
				
				#endregion
				
				#region Gets the type of validation algorithm			
				
				if(! Object.Equals(configSection, null) && 
					(! Object.Equals(configSection.Attributes[
					CacheResources.ResourceManager[ "RES_Validation" ] ],
					null ) ) )
				{
					validation = configSection.Attributes[ 
						CacheResources.ResourceManager[ 
							"RES_Validation" ] ].Value;				
				}
				
				#endregion
				
				#region Create the validation algorithm
				
				if(! Object.Equals(validation, null) )
				{
					validationAlgorithm = CreateAlgorithmFromName( 
						validation );
					if( Object.Equals(validationAlgorithm, null) )
					{
						throw new Exception( 
							CacheResources.ResourceManager[ 
							"RES_ExceptionInvalidConfigurationDefaultDataProtectionValidation" ] );
					}
				}
				else
				{
					validationAlgorithm = new HMACSHA1();	
				}
				
				#endregion
			
				#region Conversion of validation key
				
				if(! Object.Equals(validationKey, null) )
				{
					// Throw exception if the validation key length is zero
					if(validationKey.Trim().Length == 0)
					{
						throw new Exception( 
							CacheResources.ResourceManager[ 
							"RES_InvalidBaseString" ] );
					}
					validationAlgorithm.Key = Convert.FromBase64String(
						validationKey );
				}
				
				#endregion
			
				win32DataProtectionAPI = new Win32DPAPI();		
			}
			catch( Exception genException )
			{
				ExceptionManager.Publish( genException );
				throw;
			}
		}

		#endregion
		
		#region Encrypt
		
		/// <summary>
		///	Encrypts a raw of bytes.
		/// </summary>
		/// <remarks>
		///	byte[] cipherText = IDataProtection.Encrypt( plainValue );
		/// </remarks>
		/// <param name="plainValue">
		///	The data to be encrypted
		/// </param>
		/// <returns>
		///	The encrypted data
		/// </returns>
		byte[] IDataProtection.Encrypt( byte[] plainValue )
		{
			try
			{
				byte[] encryptedData;

				if( Object.Equals(plainValue, null) )
				{
					throw new ArgumentNullException( "plainValue",
						CacheResources.ResourceManager[ 
						"RES_ExceptionNullPlainValue" ] );
				}
				encryptedData = win32DataProtectionAPI.Encrypt( plainValue ); 
				
				return encryptedData;
			}
			catch( Exception genException )
			{
				ExceptionManager.Publish( genException );
				throw;
			}
		}
		#endregion
		
		#region Decrypt

		/// <summary>
		///	Decrypts a raw of bytes.
		/// </summary>
		/// <remarks>
		///	byte[] plainText = IDataProtection.Decrypt( cipherValue );
		/// </remarks>
		/// <param name="cipherValue">
		///	The data to be decrypted
		/// </param>
		/// <returns>
		///	The decrypted data
		/// </returns>
		byte[] IDataProtection.Decrypt( byte[] cipherValue )
		{
			byte[] decryptedData = null;
			try
			{

				if( Object.Equals(cipherValue, null) )
				{
					throw new ArgumentNullException( "cipherValue",
						CacheResources.ResourceManager[ 
							"RES_ExceptionNullCipherValue" ] );
				}
				decryptedData = win32DataProtectionAPI.Decrypt( cipherValue ); 
				
			}
			catch( Exception genException )
			{
				ExceptionManager.Publish( genException );
			}
			return decryptedData;
		}
		#endregion
		
		#region ComputeHash

		/// <summary>
		///	Computes a hash for data validation.
		/// </summary>
		/// <remarks>
		///	byte[] hashCode = IDataProtection.ComputeHash( plainText );
		/// </remarks>
		/// <param name="plainText">
		///	The raw bytes to be converted into hash code
		/// </param>
		/// <returns>
		///	The hash code generated using "HMACSHA1" or "MACtripleDES" 
		///	algorithm
		/// </returns>
		byte[] IDataProtection.ComputeHash ( byte[] plainText )
		{
			byte[] hashCode = null;

			try
			{
				if( Object.Equals(plainText, null) )
				{
					throw new ArgumentNullException( "plainText",
						CacheResources.ResourceManager[ 
							"RES_ExceptionNullPlainText" ] );
				}
				hashCode = validationAlgorithm.ComputeHash( plainText );
				
			}
			catch( Exception genException )
			{
				ExceptionManager.Publish( genException );
			}
			return hashCode;
		}
		#endregion

		#endregion
		
		#region Private methods
		/// <summary>
		///	Create the algorithm for validation of data.
		/// </summary>
		/// <remarks>
		///	KeyedHashAlgorithm algorithm = CreateAlgorithmFromName( name );
		/// </remarks>
		/// <param name="name">
		///	The validation algorithm name read from the "DataProtectionInfo" 
		///	of the app.config file		
		/// </param>
		/// <returns>
		///	The hash algorithm name i.e. either "HMACSHA1" or "MACTripleDES"
		/// </returns>
		private KeyedHashAlgorithm CreateAlgorithmFromName ( string name )
		{
			try
			{
				if( Object.Equals(name, null) )
				{
					throw new ArgumentNullException( "name",
						CacheResources.ResourceManager[ 
							"RES_ExceptionNullName" ] );
				}
				if( Object.Equals(name, String.Empty) )
				{
					throw new ArgumentOutOfRangeException( "name",
						CacheResources.ResourceManager[ 
							"RES_ExceptionEmptyName" ] );
				}
				switch( name )
				{
					case "SHA1":
						return new HMACSHA1();
					case "3DES":
						return new MACTripleDES(); 
					default:
						return null;
				}
			}
			catch( Exception genException )
			{
				ExceptionManager.Publish( genException );
				throw;
			}			
		}

		#endregion
	}

	#endregion
}