//===============================================================================
// Microsoft Caching Application Block for .NET
// http://msdn.microsoft.com/library/default.asp?url=/library/en-us/dnpag/html/CachingBlock.asp
//
// Win32DPAPI.cs
// The DPAPI wrapper class.
//
//===============================================================================
// Copyright (C) 2003 Microsoft Corporation
// All rights reserved.
// THIS CODE AND INFORMATION IS PROVIDED "AS IS" WITHOUT WARRANTY
// OF ANY KIND, EITHER EXPRESSED OR IMPLIED, INCLUDING BUT NOT
// LIMITED TO THE IMPLIED WARRANTIES OF MERCHANTABILITY AND/OR
// FITNESS FOR A PARTICULAR PURPOSE.
//===============================================================================


using System;
using System.Xml;
using System.Runtime.InteropServices; 
using System.Security.Cryptography;

// Used for all the caching operations
using Microsoft.ApplicationBlocks.Cache;
// Used for Handling Exceptions
using Microsoft.ApplicationBlocks.ExceptionManagement;

namespace Microsoft.ApplicationBlocks.Cache.DataProtection
{
	#region Enum Declarations

	/// <summary>
	///		Specifies which key will be used to encrypt an item
	/// </summary>
	internal enum DPAPIStore {Machine = 1, User};

	#endregion	

	#region DPAPI Wrapper class

	/// <summary>
	///	The DPAPI wrapper class.
	/// </summary>
	internal class Win32DPAPI
	{
		#region Private Members

		private DPAPIStore dataProtectionStore = DPAPIStore.Machine;

		#endregion
		
		#region Constants

		private const int CRYPTPROTECT_UI_FORBIDDEN = 0x1;
		private const int CRYPTPROTECT_LOCAL_MACHINE = 0x4;	
	
		#endregion

		#region P/Invoke structures

		[StructLayout(LayoutKind.Sequential, CharSet = CharSet.Unicode)]
		internal struct DATA_BLOB
		{
			public int cipherBlobData;
			public IntPtr plainBlobData;
		}

		[StructLayout(LayoutKind.Sequential, CharSet = CharSet.Unicode)]
		internal struct CRYPTPROTECT_PROMPTSTRUCT
		{
			public int cipherBlobSize;
			public int dwPromptFlags;
			public IntPtr hwndApp;
			public String szPrompt;
		}

		#endregion

		#region External methods

		[DllImport("Crypt32.dll", SetLastError=true, CharSet = CharSet.Auto)]
		private static extern bool CryptProtectData(
			ref DATA_BLOB pDataIn, 
			String szDataDescr, 
			ref DATA_BLOB pOptionalEntropy,
			IntPtr pvReserved, 
			ref CRYPTPROTECT_PROMPTSTRUCT 
			pPromptStruct, 
			int dwFlag, 
			ref DATA_BLOB pDataOut);

		[DllImport("Crypt32.dll", SetLastError=true, CharSet = CharSet.Auto)]
		private static extern bool CryptUnprotectData(
			ref DATA_BLOB pDataIn, 
			String szDataDescr, 
			ref DATA_BLOB pOptionalEntropy, 
			IntPtr pvReserved, 
			ref CRYPTPROTECT_PROMPTSTRUCT 
			pPromptStruct, 
			int dwFlag, 
			ref DATA_BLOB pDataOut);
		
		[DllImport("kernel32.dll", CharSet = CharSet.Auto)]
		private unsafe static extern int FormatMessage(int dwFlag, 
			ref IntPtr lpSource, 
			int dwMessageId,
			int dwLanguageId, 
			ref String lpBuffer, 
			int nSize,
			IntPtr *Arguments);

		#endregion

		#region Constructor

		public Win32DPAPI()
		{}
		
		public Win32DPAPI( DPAPIStore store )
		{
			dataProtectionStore = store;
		}

		#endregion

		#region Public methods

		#region Encrypt

		/// <summary>
		///	Encrypts a raw of bytes.
		/// </summary>
		/// <remarks>
		///	byte[] cipherText = Encrypt( plainText );
		/// </remarks>
		/// <param name="plainText">
		///	The data to be encrypted
		/// </param>
		/// <returns>
		///	The encrypted data
		/// </returns>
		public byte[] Encrypt ( byte[] plainText )
		{		
			try
			{
				byte[] encryptedData;

				if( Object.Equals(plainText, null) )
				{
					throw new ArgumentNullException( "plainText",
						CacheResources.ResourceManager[ 
						"RES_ExceptionNullPlainText" ] );
				}
				encryptedData = Encrypt(plainText, null);
                
				return encryptedData;
			}
			catch( Exception genException )
			{
				ExceptionManager.Publish( genException );
				throw;
			}		
		}
		
		/// <summary>
		///	Encrypts a raw of bytes.
		/// </summary>
		/// <remarks>
		///	byte[] cipherText = Encrypt( plainText, optionalEntropy );
		/// </remarks>
		/// <param name="plainText">
		///	The data to be encrypted
		/// </param>
		/// <param name="optionalEntropy">
		///	Password or other additional entropy used when the data was 
		///	encrypted. If an optional entropy was used in the encryption phase,
		///	that same entropy	must be used for the decryption phase.
		/// </param>
		/// <returns>
		///	The encrypted data
		/// </returns>
		public byte[] Encrypt ( byte[] plainText, byte[] optionalEntropy )
		{
			DATA_BLOB plainTextBlob = new DATA_BLOB();
			DATA_BLOB cipherTextBlob = new DATA_BLOB();
			DATA_BLOB entropyBlob = new DATA_BLOB();

			try
			{
				#region Variable Declarations

				bool isValue;

				CRYPTPROTECT_PROMPTSTRUCT prompt = new CRYPTPROTECT_PROMPTSTRUCT();
				InitPromptstruct(ref prompt);
				int dwFlag;
				int bytesSize;
				byte[] cipherText;
				
				#endregion

				if( Object.Equals(plainText, null) )
				{
					throw new ArgumentNullException( "plainText",
						CacheResources.ResourceManager[ 
						"RES_ExceptionNullPlainText" ] );
				}
				try
				{
					bytesSize = plainText.Length;
					plainTextBlob.plainBlobData = Marshal.AllocHGlobal(
						bytesSize);
					if( Object.Equals( IntPtr.Zero,
						plainTextBlob.plainBlobData ) )
					{
						throw new Exception( CacheResources.ResourceManager[
							"RES_FailedToAllocatePlainTextBuffer" ] );							
					}
					plainTextBlob.cipherBlobData = bytesSize;
					// Copy to unmanaged area 
					Marshal.Copy( plainText,
						0, 
						plainTextBlob.plainBlobData,
						bytesSize );
				}
				catch( Exception encryptException )
				{
					throw new Exception( CacheResources.ResourceManager[
						"RES_ExceptionMarshallingData" ],
						encryptException );
				}
				if(! Object.Equals(DPAPIStore.Machine,dataProtectionStore))
				{
					// Using the machine store, should be providing entropy
					dwFlag = CRYPTPROTECT_LOCAL_MACHINE |
						CRYPTPROTECT_UI_FORBIDDEN;

					// Check to see if the entropy is null
					if( Object.Equals(optionalEntropy, null) )
					{
						// Allocate something
						optionalEntropy = new byte[0];
					}
					try
					{
						bytesSize = optionalEntropy.Length;
						entropyBlob.plainBlobData = Marshal.AllocHGlobal(
							optionalEntropy.Length);

						if(! Object.Equals(IntPtr.Zero,
							entropyBlob.plainBlobData) )
						{
							throw new Exception( 
								CacheResources.ResourceManager[
								"RES_FailedToAllocateEntropyDataBuffer" ] );								
						}
						// Copy to unmanaged area
						Marshal.Copy( optionalEntropy,
							0, 
							entropyBlob.plainBlobData, 
							bytesSize );
						entropyBlob.cipherBlobData = bytesSize;
					}
					catch( Exception encryptException )
					{
						throw new Exception( CacheResources.ResourceManager[ 
							"RES_ExceptionEntropyMarshallingData" ],
							encryptException );
					}
				}
				else
				{
					// Using the user store
					dwFlag = CRYPTPROTECT_UI_FORBIDDEN;
				}
				isValue = CryptProtectData( ref plainTextBlob, 
					"", 
					ref entropyBlob, 
					IntPtr.Zero, 
					ref prompt, 
					dwFlag, 
					ref cipherTextBlob);

				if( Object.Equals(isValue, false) )
				{
					throw new Exception( CacheResources.ResourceManager[
						"RES_EncryptionFailed" ] 
						+ GetErrorMessage(Marshal.GetLastWin32Error()) );
				}

				cipherText = new byte[ cipherTextBlob.cipherBlobData ];
				// Copy back to managed area
				Marshal.Copy( cipherTextBlob.plainBlobData,
					cipherText,
					0, 
					cipherTextBlob.cipherBlobData );
				return cipherText;
			}
			catch( Exception encryptException )
			{
				throw new Exception( CacheResources.ResourceManager[
					"RES_EncryptionFailed" ]
					+ encryptException.Message );
			}	
			finally
			{
				if(! Object.Equals(IntPtr.Zero,
					cipherTextBlob.plainBlobData))
				{
					Marshal.FreeHGlobal(cipherTextBlob.plainBlobData);
				}
				if(! Object.Equals(IntPtr.Zero,
					entropyBlob.plainBlobData))
				{
					Marshal.FreeHGlobal(entropyBlob.plainBlobData);
				}
				if(!Object.Equals(IntPtr.Zero,
					plainTextBlob.plainBlobData))
				{
					Marshal.FreeHGlobal(plainTextBlob.plainBlobData);
				}
			}
		}

		#endregion Encrypt

		#region Decrypt

		/// <summary>
		///	Decrypts a raw of bytes.
		/// </summary>
		/// <remarks>
		///	byte[] plainText = Decrypt( cipherText );
		/// </remarks>
		/// <param name="cipherText">
		///	The data to be decrypted
		/// </param>
		/// <returns>
		///	The decrypted data
		/// </returns>
		public byte[] Decrypt ( byte[] cipherText )
		{
			
			try
			{
				byte[] decryptedData;

				if( Object.Equals(cipherText, null) )
				{
					throw new ArgumentNullException( "cipherText",
						CacheResources.ResourceManager[ 
							"RES_ExceptionNullCipherText" ] );
				}
				decryptedData = Decrypt( cipherText, null );
			
				return decryptedData;
			}
			catch( Exception genException )
			{
				ExceptionManager.Publish( genException );
				throw;
			}
		}
		
		/// <summary>
		///	Decrypts a raw of bytes.
		/// </summary>
		/// <remarks>
		///	byte[] plainText = Decrypt(cipherText, optionalEntropy);
		/// </remarks>
		/// <param name="cipherText">
		///	The data to be decrypted
		/// </param>
		/// <param name="optionalEntropy">
		///	Password or other additional entropy used when the data was 
		///	encrypted. If an optional entropy was used in the encryption 
		///	phase, that same entropy	must be used for the decryption phase.
		/// </param>
		/// <returns>
		///	The decrypted data
		/// </returns>
		public byte[] Decrypt ( byte[] cipherText, byte[] optionalEntropy )
		{
			DATA_BLOB plainTextBlob = new DATA_BLOB();
			DATA_BLOB cipherBlob = new DATA_BLOB();
			DATA_BLOB entropyBlob = new DATA_BLOB();	

			try
			{
				#region Variable Declarations

				bool isValue;
				CRYPTPROTECT_PROMPTSTRUCT prompt = new CRYPTPROTECT_PROMPTSTRUCT();
				InitPromptstruct(ref prompt);
				int cipherTextSize;
				int dwFlag;
				int bytesSize;
				byte[] plainText;

				#endregion

				if( Object.Equals(cipherText, null) )
				{
					throw new ArgumentNullException( "cipherText",
						CacheResources.ResourceManager[ 
						"RES_ExceptionNullCipherText" ] );
				}
				try
				{
					cipherTextSize = cipherText.Length;
					cipherBlob.plainBlobData = Marshal.AllocHGlobal(
						cipherTextSize);

					if( Object.Equals(IntPtr.Zero, cipherBlob.plainBlobData) )
					{
						throw new Exception( CacheResources.ResourceManager[
							"RES_FailedToAllocateCipherTextBuffer" ] );					
					}
					cipherBlob.cipherBlobData = cipherTextSize;
					Marshal.Copy( cipherText,
						0, 
						cipherBlob.plainBlobData, 
						cipherBlob.cipherBlobData );
				}
				catch(Exception decryptException)
				{
					throw new Exception( CacheResources.ResourceManager[
						"RES_ExceptionMarshallingData" ],
						decryptException );
				}
//				DATA_BLOB entropyBlob = new DATA_BLOB();				

				if( Object.Equals(DPAPIStore.Machine, dataProtectionStore) )
				{
					// Using the machine store, should be providing entropy
					dwFlag = CRYPTPROTECT_LOCAL_MACHINE | 
						CRYPTPROTECT_UI_FORBIDDEN;

					// Check to see if the entropy is null
					if(optionalEntropy == null)
					{
						// Allocate something
						optionalEntropy = new byte[0];
					}
					try
					{
						bytesSize = optionalEntropy.Length;
						entropyBlob.plainBlobData = Marshal.AllocHGlobal(
							bytesSize);
						if( Object.Equals(IntPtr.Zero, 
							entropyBlob.plainBlobData) )
						{
							throw new Exception(CacheResources.ResourceManager
								["RES_FailedToAllocateEntropyDataBuffer"]);								
						}
						entropyBlob.cipherBlobData = bytesSize;
						// Copy to unmanaged area
						Marshal.Copy(optionalEntropy, 
							0, 
							entropyBlob.plainBlobData,
							bytesSize);
					}
					catch( Exception decryptException )
					{
						throw new Exception( CacheResources.ResourceManager[
							"RES_ExceptionEntropyMarshallingData" ],
							decryptException );
					}
				}
				else
				{
					// Using the user store
					dwFlag = CRYPTPROTECT_UI_FORBIDDEN;
				}
				isValue = CryptUnprotectData(ref cipherBlob,
					null,
					ref entropyBlob, 
					IntPtr.Zero,
					ref prompt,
					dwFlag, 
					ref plainTextBlob);

				if( Object.Equals(isValue, false) )
				{
					throw new Exception( CacheResources.ResourceManager [
						"RES_DecryptionFailed"  ]
						+ GetErrorMessage(Marshal.GetLastWin32Error()) );
				}

				plainText = new byte[plainTextBlob.cipherBlobData];
				// Copy to managed area
				Marshal.Copy( plainTextBlob.plainBlobData,
					plainText,
					0, 
					plainTextBlob.cipherBlobData );
				return plainText;
			}
			catch(Exception decryptException)
			{
				throw new Exception( CacheResources.ResourceManager [
					"RES_DecryptionFailed"  ]
					+ decryptException.Message );
			}
			finally
			{
				if(! Object.Equals(IntPtr.Zero,
					cipherBlob.plainBlobData))
				{
					Marshal.FreeHGlobal(cipherBlob.plainBlobData);
				}
				if(! Object.Equals(IntPtr.Zero,
					entropyBlob.plainBlobData))
				{
					Marshal.FreeHGlobal(entropyBlob.plainBlobData);
				}
				if(!Object.Equals(IntPtr.Zero,
					plainTextBlob.plainBlobData))
				{
					Marshal.FreeHGlobal(plainTextBlob.plainBlobData);
				}
			}
		}

		#endregion Decrypt

		#endregion
		
		#region Private methods

		#region InitPromptstruct

		/// <summary>
		///	Initialize the value for the Promptstruct.
		/// </summary>
		/// <remarks>
		///	InitPromptstruct( ref promptStruct );
		/// </remarks>
		/// <param name="promptStruct">
		///	Reference of the Promptstruct
		/// </param>
		private void InitPromptstruct( 
			ref CRYPTPROTECT_PROMPTSTRUCT promptStruct ) 
		{
			try
			{
				promptStruct.cipherBlobSize = Marshal.SizeOf( 
					typeof( CRYPTPROTECT_PROMPTSTRUCT ) );
				promptStruct.dwPromptFlags = 0;
				promptStruct.hwndApp = IntPtr.Zero;
				promptStruct.szPrompt = null;
			}
			catch( Exception genException )
			{
				ExceptionManager.Publish( genException );
				throw;
			}
		}

		#endregion 

		#region GetErrorMessage
		
		/// <summary>
		///	Gets the error message.
		/// </summary>
		/// <remarks>
		///	String lpMsgBuf = GetErrorMessage( errorCode );
		/// </remarks>
		/// <param name="errorCode">
		///	The error code for which the message has to be retrieved
		///	</param>
		/// <returns>
		///	The error message
		/// </returns>
		private unsafe static String GetErrorMessage ( int errorCode )
		{
			try
			{
				#region Variable declaration
			
				int FORMAT_MESSAGE_ALLOCATE_BUFFER = 0x00000100;
				int FORMAT_MESSAGE_IGNORE_INSERTS = 0x00000200;
				int FORMAT_MESSAGE_FROM_SYSTEM  = 0x00001000;
				int messageSize = 255;
				String lpMsgBuf = "";
				int dwFlag = FORMAT_MESSAGE_ALLOCATE_BUFFER | 
					FORMAT_MESSAGE_FROM_SYSTEM | 
					FORMAT_MESSAGE_IGNORE_INSERTS;
				IntPtr ptrlpSource = new IntPtr();
				IntPtr prtArguments = new IntPtr();
			
				#endregion

				int isValue = FormatMessage( dwFlag,
					ref ptrlpSource, 
					errorCode,
					0, 
					ref lpMsgBuf,
					messageSize, 
					&prtArguments );

				if( isValue != 0)
				{
					throw new Exception( CacheResources.ResourceManager[ 
						"RES_FailedToFormatMessage" ]
						+ errorCode );
				}

				return lpMsgBuf;
			}
			catch( Exception genException )
			{
				ExceptionManager.Publish( genException );
				throw;
			}			
		}

		#endregion

		#endregion
	}
	#endregion
}