//===============================================================================
// Microsoft Caching Application Block for .NET
// http://msdn.microsoft.com/library/default.asp?url=/library/en-us/dnpag/html/CachingBlock.asp
//
// DataProtectionManager.cs
// This class simplifies the use of the cryptographic functions.
//
//===============================================================================
// Copyright (C) 2003 Microsoft Corporation
// All rights reserved.
// THIS CODE AND INFORMATION IS PROVIDED "AS IS" WITHOUT WARRANTY
// OF ANY KIND, EITHER EXPRESSED OR IMPLIED, INCLUDING BUT NOT
// LIMITED TO THE IMPLIED WARRANTIES OF MERCHANTABILITY AND/OR
// FITNESS FOR A PARTICULAR PURPOSE.
//===============================================================================


using System;
using System.IO;
using System.Text;
using System.Runtime.Serialization.Formatters.Binary;

// Used for handling Exceptions
using Microsoft.ApplicationBlocks.ExceptionManagement;

namespace Microsoft.ApplicationBlocks.Cache
{
	#region DataProtection class

	/// <summary>
	///	This class simplifies the use of the cryptographic functions.
	/// </summary>
	public sealed class DataProtectionManager
	{
		// Private Members
		private static IDataProtection dataProtection;
		
		#region MACValue struct

		/// <summary>
		///	To store the value before and after using the MACTriples algorithm.
		/// </summary>
		[Serializable]
		private struct MACValue
		{
			public byte[] Value;
			public byte[] MAC;
		}

		#endregion
		
		#region Constructor

		/// <summary>
		///	Constructor to read the data protection information from the 
		///	config file.
		/// </summary>
		static DataProtectionManager()
		{
			try
			{
				// Read storage mode from config file
				if( ! Object.Equals(
					CacheConfiguration.Config.DataProtectionInformation, null))
				{
					// Initialize the Data Protection provider
					dataProtection = CacheConfiguration.Config.
						DataProtectionInformation.CreateInstance();
				}
				else 
				{
					// Initialize the default DataProtection provider	
					dataProtection 
						= new DataProtection.DefaultDataProtection(); 
					dataProtection.Init( null );
				}
			}
			catch( Exception genException )
			{
				ExceptionManager.Publish( genException );
				throw;
			}
		}
		
		// Prevents creation of the instance of this class
		private DataProtectionManager(){}

		#endregion
		
		#region Public Methods

		#region Encrypt
		
		/// <summary>
		///	Encrypts a raw of bytes.
		/// </summary>
		/// <remarks>
		///	byte[] encryptedData = Encrypt( plainValue );
		/// </remarks>
		/// <param name="plainValue">
		///	The data to be encrypted
		/// </param>
		/// <returns>
		///	The encrypted data
		/// </returns>
		public static byte[] Encrypt ( byte[] plainValue )
		{
			try
			{
				byte[] encryptedData;

				if( Object.Equals(plainValue, null) )
				{
					throw new ArgumentNullException( "plainValue",
						CacheResources.ResourceManager[ 
							"RES_ExceptionNullPlainValue" ] );
				}
				encryptedData = dataProtection.Encrypt( plainValue );

				return encryptedData;
			}
			catch( Exception genException )
			{
				ExceptionManager.Publish( genException );
				throw;
			}
			
		}
		#endregion
		
		#region Decrypt

		/// <summary>
		///	Decrypts a raw of bytes.
		/// </summary>
		/// <remarks>
		///	byte[] decryptedData = Decrypt( cipherValue );
		/// </remarks>
		/// <param name="cipherValue">
		///	The data to be decrypted
		/// </param>
		/// <returns>
		///	The decrypted data
		/// </returns>
		public static byte[] Decrypt ( byte[] cipherValue )
		{
			try
			{
				byte[] decryptedData;

				if( Object.Equals(cipherValue, null) )
				{
					throw new ArgumentNullException( "cipherValue",
						CacheResources.ResourceManager[ 
							"RES_ExceptionNullCipherValue" ] );
				}
				decryptedData = dataProtection.Decrypt( cipherValue ); 
				
				return decryptedData;
			}
			catch( Exception genException )
			{
				ExceptionManager.Publish( genException );
				throw;
			}
		}

		#endregion
		
		#region AppendMAC

		/// <summary>
		///	Appends a message authenticate code to plain value. Compute the 
		///	MAC by hashing the the value with the configured validation key. 
		/// </summary>
		/// <remarks>
		///	byte[] codeMAC = AppendMAC( plainValue );
		///	</remarks>
		///	<param name="plainValue">
		///	The data to be converted to the MAC code
		///	</param>
		///	<returns>
		///	The MAC code is used to ensure data integrity
		///	</returns>
		public static byte[] AppendMAC ( byte[] plainValue )
		{
			try
			{
				byte[] hashCode;

				if( Object.Equals(plainValue, null) )
				{
					throw new ArgumentNullException( "plainValue",
						CacheResources.ResourceManager[ 
							"RES_ExceptionNullPlainValue" ] );
				}
				MACValue macVal = new MACValue();
				macVal.Value = plainValue;
				macVal.MAC = dataProtection.ComputeHash( plainValue );
			
				MemoryStream memStream = new MemoryStream();
				new BinaryFormatter().Serialize( memStream, macVal );
			
				hashCode = memStream.ToArray();
				memStream.Close();

				return hashCode;
			}
			catch( Exception genException )
			{
				ExceptionManager.Publish( genException );
				throw;
			}
		}
		#endregion
		
		#region RemoveMAC
		/// <summary>
		///	Gets the plain value.  This method takes the plain value and 
		///	produces a hash.  The resulting hash is compared with the existing 
		///	hash and if they match the value is considered valid.
		/// </summary>
		/// <remarks>
		///	byte[] codeMAC = RemoveMAC( binaryMacValue );
		/// </remarks>
		/// <param name="binaryMacValue">
		///	The MAC code to be compared
		/// </param>
		/// <returns>
		///	The MAC code if the comparison of hashes is valid
		/// </returns>
		public static byte[] RemoveMAC ( byte[] binaryMacValue )
		{
			try
			{
				byte[] hashCode;
			
				if( Object.Equals(binaryMacValue, null) )
				{
					throw new ArgumentNullException( "binaryMacValue",
						CacheResources.ResourceManager[ 
							"RES_ExceptionNullBinaryMACValue" ] );
				}
				MemoryStream memStream = new MemoryStream( binaryMacValue );
				MACValue macVal = 
					(MACValue)new BinaryFormatter().Deserialize( memStream );
				memStream.Close();
						
				byte[] hash = dataProtection.ComputeHash( macVal.Value );
			
				// Compares the original hash with the new hash			
				for( int index = 0; index < hash.Length; index++ )
					if( hash[index] != macVal.MAC[index] )
						throw new Exception( 
							CacheResources.ResourceManager[ 
							"RES_ExceptionDataValidation" ] );	

				hashCode = macVal.Value;

				return hashCode;
			}
			catch( Exception genException )
			{
				ExceptionManager.Publish( genException );
				throw;
			}
		}
		#endregion

		#endregion
	}
	
	#endregion
}
