//===============================================================================
// Microsoft Caching Application Block for .NET
// http://msdn.microsoft.com/library/default.asp?url=/library/en-us/dnpag/html/CachingBlock.asp
//
// AbsoluteTime.cs
// This class tests if a data item was expired using a absolute time schema.
//
//===============================================================================
// Copyright (C) 2003 Microsoft Corporation
// All rights reserved.
// THIS CODE AND INFORMATION IS PROVIDED "AS IS" WITHOUT WARRANTY
// OF ANY KIND, EITHER EXPRESSED OR IMPLIED, INCLUDING BUT NOT
// LIMITED TO THE IMPLIED WARRANTIES OF MERCHANTABILITY AND/OR
// FITNESS FOR A PARTICULAR PURPOSE.
//===============================================================================


using System;
using System.Globalization;
using System.Runtime.Serialization;
using System.Security.Permissions;

// Used for Handling Exceptions
using Microsoft.ApplicationBlocks.ExceptionManagement;

namespace Microsoft.ApplicationBlocks.Cache.Expirations
{
	/// <summary>
	///	This class tests if a data item was expired using a absolute time 
	///	schema.
	/// </summary>
	[Serializable]
	public class AbsoluteTime : ICacheItemExpiration, ISerializable
	{
		#region Private members

		private DateTime absoluteExpirationTime;
		private bool isExpired;
		private string keyValue;	
	
		#endregion

		#region Constructors

		/// <summary>
		///	Create a new instance of the class.
		/// </summary>
		public AbsoluteTime()
		{}

		/// <summary>
		///	Create an instance of the class with a time value as input and 
		///	convert it to UTC.
		/// </summary>
		/// <param name="absoluteTime">
		///	The time to be checked for expiration
		/// </param>
		public AbsoluteTime( DateTime absoluteTime )
		{
			try
			{ 
				if( absoluteTime > DateTime.Now )
				{
					// Convert to UTC in order to compensate for time zones	
					absoluteExpirationTime = absoluteTime.ToUniversalTime();
				}
				else
				{
					throw(new ArgumentOutOfRangeException("absoluteTime",
						CacheResources.ResourceManager[ 
							"RES_ExceptionRangeAbsoluteTime"]));
				}

				absoluteExpirationTime = absoluteTime;

				// Convert to UTC in order to compensate for time zones		
				absoluteExpirationTime = 
					absoluteExpirationTime.ToUniversalTime();
			}
			catch( Exception genException )
			{
				ExceptionManager.Publish( genException );
				throw;
			}
		}
		#endregion

		#region ICacheItemExpiration Implementation

		/// <summary>
		///	This method sets the external dependency key.
		/// </summary>
		/// <param name="keyVal">
		/// Specifies the key value
		/// </param>
		void ICacheItemExpiration.Key(string keyVal)
		{
			try
			{
				if(Object.Equals(keyVal, null))
				{
					throw(new ArgumentNullException("keyVal",
						CacheResources.ResourceManager[
						"RES_ExceptionNullKey"]));
				}
				if(keyVal.Length == 0)
				{
					throw(new ArgumentOutOfRangeException("keyVal",
						CacheResources.ResourceManager[
						"RES_ExceptionEmptyKey"]));
				}
				keyValue = keyVal;
			}
			catch(Exception genException)
			{
				ExceptionManager.Publish(genException);
				throw;
			}
		}

		/// <summary>
		///	Specifies if item has expired or not.
		/// </summary>
		/// <remarks>
		///	bool isExpired = ICacheItemExpiration.HasExpired();
		/// </remarks>
		/// <returns>
		///	"True", if the data item has expired or "false", if the data item 
		///	has not expired
		/// </returns>
		bool ICacheItemExpiration.HasExpired()
		{
			try
			{
				// Convert to UTC in order to compensate for time zones		
				DateTime nowDateTime = DateTime.Now.ToUniversalTime();

				// Check expiration
				if(nowDateTime.Ticks > absoluteExpirationTime.Ticks)
				{
					isExpired = true;
				}
				else
				{
					isExpired = false;
				}
				return isExpired;
			}
			catch( Exception genException )
			{
				ExceptionManager.Publish( genException );
				throw;
			}			
		}

		/// <summary>
		///	Notifies that the item was recently used.
		/// </summary>
		void ICacheItemExpiration.Notify()
		{}

		/// <summary>
		///	Event to indicate the cache item expiration.
		/// </summary>
		public event ItemDependencyChangeEventHandler Change;

		#endregion

		#region Implementation of ISerializable

		/// <summary>
		///	This method performs the deserialziaton of members of the current 
		///	class.
		/// </summary>
		/// <param name="info">
		///	A SerializationInfo object which is deserialized by the formatter
		///	and then passed to current constructor
		/// </param>
		/// <param name="context">
		///	A StreamingContext that describes the source of the serialized 
		///	stream from where the Serialization object is retrieved
		/// </param>
		protected AbsoluteTime ( SerializationInfo info,
			StreamingContext context ) 
		{
			try
			{
				absoluteExpirationTime = Convert.ToDateTime(
					info.GetValue("absoluteExpiration",	typeof(DateTime)),
					DateTimeFormatInfo.CurrentInfo );
			}
			catch( Exception genException )
			{
				ExceptionManager.Publish( genException );
				throw;
			}
		}

		/// <summary>
		///	This method performs the serialziaton of members of the current 
		///	class.
		/// </summary>
		/// <param name="info">
		///	A SerializationInfo object which is deserialized by the formatter 
		///	and then passed to current constructor
		/// </param>
		/// <param name="context">
		///	A StreamingContext that describes the source of the serialized 
		///	stream from where the Serialization object is retrieved
		/// </param>
		[SecurityPermissionAttribute( SecurityAction.Demand,
			 SerializationFormatter = true )]
		[SecurityPermissionAttribute( SecurityAction.LinkDemand,
			 SerializationFormatter = true )]
		public void GetObjectData ( SerializationInfo info,
			StreamingContext context )
		{
			try
			{
				info.AddValue("absoluteExpiration", absoluteExpirationTime);
			}
			catch( Exception genException )
			{
				ExceptionManager.Publish( genException );
				throw;
			}
		}

		#endregion
	}
}
