//===============================================================================
// Microsoft Caching Application Block for .NET
// http://msdn.microsoft.com/library/default.asp?url=/library/en-us/dnpag/html/CachingBlock.asp
//
// ExtendedFormat.cs
// This class represents a extended format for the time to be checked for expirations.
//
//===============================================================================
// Copyright (C) 2003 Microsoft Corporation
// All rights reserved.
// THIS CODE AND INFORMATION IS PROVIDED "AS IS" WITHOUT WARRANTY
// OF ANY KIND, EITHER EXPRESSED OR IMPLIED, INCLUDING BUT NOT
// LIMITED TO THE IMPLIED WARRANTIES OF MERCHANTABILITY AND/OR
// FITNESS FOR A PARTICULAR PURPOSE.
//===============================================================================


using System;
using System.Globalization;

// Used for Handling Exceptions
using Microsoft.ApplicationBlocks.ExceptionManagement;

namespace Microsoft.ApplicationBlocks.Cache.Expirations
{
	#region Extended Format class
	/// <summary>
	///	This class represents a extended format for the time to be checked for 
	///	expirations.
	/// </summary>
	internal class ExtendedFormat
	{
		#region Private Members

		private static readonly char ARGUMENT_DELIMITER = Convert.ToChar(
			CacheResources.ResourceManager [ "RES_ArgumentDelimiter" ],
			CultureInfo.CurrentCulture );
		private static readonly char WILDCARD_ALL = Convert.ToChar( 
			CacheResources.ResourceManager [ "RES_WildcardAll" ],
			CultureInfo.CurrentCulture );
		const char REFRESH_DELIMITER = ' ';

		// Variables
		private int[] minutes;
		private int[] hours;
		private int[] days;
		private int[] months;
		private int[] daysOfWeek;
		private bool isAlwaysExpired;

		#endregion

		#region Properties

		/// <summary>
		///	Gets the status whether metadata has to expire every minute.
		/// </summary>
		public bool IsExpireEveryMinute
		{
			get { return Object.Equals(minutes[ 0 ], -1); }
		}

		/// <summary>
		///	Gets the status whether metadata has to expire every day.
		/// </summary>
		public bool IsExpireEveryDay
		{
			get { return Object.Equals(days[ 0 ], -1); }
		}

		/// <summary>
		///	Gets the status whether metadata has to expire every hour.
		/// </summary>
		public bool IsExpireEveryHour
		{
			get { return Object.Equals(hours[ 0 ], -1); }
		}

		/// <summary>
		///	Gets the status whether metadata has to expire every month.
		/// </summary>
		public bool IsExpireEveryMonth
		{
			get { return Object.Equals(months[ 0 ], -1); }
		}

		/// <summary>
		///	Gets the status whether metadata has to expire every week.
		/// </summary>
		public bool IsExpireEveryDayOfWeek
		{
			get { return Object.Equals(daysOfWeek[ 0 ], -1); }
		}

		/// <summary>
		///	Gets the status whether the metadata is always expire.
		/// </summary>
		public bool IsAlwaysExpire
		{
			get{ return isAlwaysExpired; }
		}

		#endregion
        
		#region Constructor

		/// <summary>
		///	Constructor to convert the string format and spilt using delimiters
		///	to retrieve the time format.
		/// </summary>
		/// <param name="format">
		/// The time format for the item to get expired
		/// </param>
		public ExtendedFormat( string timeFormat )
		{
			try
			{
				string[] parsedFormat = timeFormat.Split( REFRESH_DELIMITER );
	           
				#region Throwing Argument Exception

				if(Object.Equals(timeFormat, null))
				{
					throw(new ArgumentNullException("timeFormat",
						CacheResources.ResourceManager[
						"RES_ExceptionNullTimeFormat"]));
				}
				if(timeFormat.Length == 0)
				{
					throw(new ArgumentOutOfRangeException("timeFormat",
						CacheResources.ResourceManager[
						"RES_ExceptionRangeTimeFormat"]));
				}
				if( ! Object.Equals(parsedFormat.Length,
					Convert.ToInt32(CacheResources.ResourceManager[ 
						"RES_ParsedFormatLength" ] ) ) )
				{				
					throw new Exception( CacheResources.ResourceManager[ 
						"RES_ExceptionInvalidExtendedFormatArguments" ] );
				}

				#endregion

				#region Minutes

				minutes = ParseValueToInt( parsedFormat[ 
					Convert.ToInt32(CacheResources.ResourceManager[ 
						"RES_MinutesIndex" ] ) ] );

				foreach (int minute in minutes)
				{
					if(minute > 59)
					{
						throw(new ArgumentOutOfRangeException("minutes",
							CacheResources.ResourceManager[
							"RES_ExceptionRangeMinute"]));
					}
				}

				#endregion

				#region Hours

				hours = ParseValueToInt( parsedFormat[ 
					Convert.ToInt32(CacheResources.ResourceManager[ 
						"RES_HoursIndex" ] ) ] );

				foreach(int hour in hours)
				{
					if(hour > 23)
					{
						throw(new ArgumentOutOfRangeException("hours",
							CacheResources.ResourceManager[
							"RES_ExceptionRangeHour"]));
					}
				}

				#endregion

				#region Days

				days = ParseValueToInt( parsedFormat[ 
					Convert.ToInt32(CacheResources.ResourceManager[ 
						"RES_DaysIndex" ] ) ] );

				foreach(int day in days)
				{
					if((day < 1 || day > 31) && day != -1)
					{
						throw(new ArgumentOutOfRangeException("days",
							CacheResources.ResourceManager[
							"RES_ExceptionRangeDayOfMonth"]));
					}
				}

				#endregion

				#region Months

				months = ParseValueToInt( parsedFormat[ 
					Convert.ToInt32(CacheResources.ResourceManager[ 
						"RES_MonthsIndex" ] ) ] );
				
				foreach(int month in months)
				{
					if((month < 1 || month > 12) && (month != -1))
					{
						throw(new ArgumentOutOfRangeException("months",
							CacheResources.ResourceManager[
							"RES_ExceptionRangeMonth"]));
					}
				}

				#endregion

				#region DaysOfWeek

				daysOfWeek = ParseValueToInt( parsedFormat[ 
					Convert.ToInt32(CacheResources.ResourceManager[ 
					"RES_DaysOfWeekIndex" ] ) ] );

				foreach(int day in daysOfWeek)
				{
					if(day > 6)
					{
						throw(new ArgumentOutOfRangeException("daysOfWeek",
							CacheResources.ResourceManager[
							"RES_ExceptionRangeDay"]));
					}
				}

				#endregion

				for(int dayLength = 0; dayLength <= days.Length - 1;
					dayLength++)
				{
					if(! (days[dayLength] == -1 || months[dayLength] == -1))
					{
						if(days[dayLength] > DateTime.DaysInMonth(
							DateTime.Now.Year, months[dayLength]))
						{
							throw(new Exception(CacheResources.ResourceManager[
								"RES_ExcetionInvalidDay"]));
						}
					}
				}

				if( IsExpireEveryMinute && IsExpireEveryHour && 
					IsExpireEveryDay && IsExpireEveryMonth && 
					IsExpireEveryDayOfWeek )
				{
					isAlwaysExpired = true;
				}
			}
			catch( Exception genException )
			{
				ExceptionManager.Publish( genException );
				throw;
			}
		}

		#endregion

		#region Public Methods

		/// <summary>
		///	This method gets the extended format minute argument.
		/// </summary>
		public int[] MinutesFormat()
		{
			return minutes; 
		}

		/// <summary>
		///	This method gets the extended format hour argument.
		/// </summary>
		public int[] HoursFormat()
		{
			return hours; 
		}

		/// <summary>
		///	This method gets the extended format day of month argument.
		/// </summary>
		public int[] DaysFormat()
		{
			return days; 
		}

		/// <summary>
		///	This method gets the extended format month argument.
		/// </summary>
		public int[] MonthsFormat()
		{
			return months;
		}

		/// <summary>
		///	This method gets the extended format day of week argument.
		/// </summary>
		public int[] DaysOfWeekFormat()
		{
			return daysOfWeek;
		}

		#endregion

		#region Private Methods

		/// <summary>
		///	Parse a string value to int.
		/// </summary>
		/// <remarks>
		///	int[] result = ParseValueToInt( timeFormat );
		/// </remarks>
		/// <param name="timeFormat">
		///	The string value
		/// </param>
		/// <returns>
		///	Int array
		/// </returns>
		private int[] ParseValueToInt( string timeFormat )
		{
			int[] result = null;

			try
			{
				#region Throwing Argument Exception

				if(Object.Equals(timeFormat, null))
				{
					throw(new ArgumentNullException("timeFormat",
						CacheResources.ResourceManager[
						"RES_ExceptionNullTimeFormat"]));
				}
				if(timeFormat.Length == 0)
				{
					throw(new ArgumentOutOfRangeException("timeFormat",
						CacheResources.ResourceManager[
						"RES_ExceptionRangeTimeFormat"]));
				}

				#endregion

				if( ! Object.Equals(timeFormat.IndexOf( WILDCARD_ALL ), -1) )
				{
					result = new int[1];
					result[0] = -1;
				}
				else
				{
					string[] values = timeFormat.Split( ARGUMENT_DELIMITER );
					result = new int[ values.Length ]; 
					for( int counter = 0; counter < values.Length; counter++)
					{
						result[ counter ] =  int.Parse( values[ counter ],
							CultureInfo.CurrentCulture );
						if(result[counter] < 0)
						{
							throw(new ArgumentException(CacheResources.
								ResourceManager[
								"RES_ExceptionInvalidExtendedFormatTime"]));
						}
					}
				}
			}
			catch( Exception genException )
			{
				ExceptionManager.Publish( genException );
				throw;
			}
			return result;
		}

		#endregion
	}
    #endregion
}