//===============================================================================
// Microsoft Caching Application Block for .NET
// http://msdn.microsoft.com/library/default.asp?url=/library/en-us/dnpag/html/CachingBlock.asp
//
// ExtendedFormatHelper.cs
// This class tests if a item was expired using a extended format.
//
//===============================================================================
// Copyright (C) 2003 Microsoft Corporation
// All rights reserved.
// THIS CODE AND INFORMATION IS PROVIDED "AS IS" WITHOUT WARRANTY
// OF ANY KIND, EITHER EXPRESSED OR IMPLIED, INCLUDING BUT NOT
// LIMITED TO THE IMPLIED WARRANTIES OF MERCHANTABILITY AND/OR
// FITNESS FOR A PARTICULAR PURPOSE.
//===============================================================================


using System;
using System.Collections;
using System.Runtime.InteropServices;

// Used for Handling Exceptions
using Microsoft.ApplicationBlocks.ExceptionManagement;

namespace Microsoft.ApplicationBlocks.Cache.Expirations
{
	#region Extended Format Helper class
	/// <summary>
	///	This class tests if a item was expired using a extended format.
	/// </summary>
	/// <remarks>
	/// Extended format sintax : <br/><br/>
	/// 
	/// Minute       - 0-59 <br/>
	/// Hour         - 0-23 <br/>
	/// Day of month - 1-31 <br/>
	/// Month        - 1-12 <br/>
	/// Day of week  - 0-7 (Sunday is 0 or 7) <br/>
	/// Wildcards    - * means run every <br/>
	/// Examples: <br/>
	/// * * * * *    - expire every minute of every day <br/>
	/// 5 * * * *    - expire 5th minute of every hour or every minute
	///				   after 1 hour from the last used time, whichever
	///				   is earlier <br/>
	/// * 21 * * *   - expire every minute of the 21st hour of every day 
	///				   or one hour after every day<br/>
	/// 31 15 * * *  - expire 3:31 PM every day <br/>
	/// 7 4 * * 6    - expire Saturday 4:07 AM <br/>
	/// 15 21 4 7 *  - expire 9:15 PM on 4 July <br/>
	/// </remarks>
	[ComVisible(false)]
	internal class ExtendedFormatHelper
	{
		#region Private Members

		private static Hashtable parsedFormatCache = new Hashtable();

		#endregion

		#region Constructor

		/// <summary>
		/// Default constructor.
		/// </summary>
		private ExtendedFormatHelper() {}

		#endregion
 
		#region Public Methods

		#region IsExtendedExpired
		/// <summary>
		///	Test the extended format with a given date.
		/// </summary>
		/// <remarks>
		///	bool isExpired = IsExtendedExpired( format, getTime, nowTime );
		/// </remarks>
		/// <param name="format">
		///	The extended format arguments
		///	</param>
		/// <param name="getTime">
		///	The time when the item has been refreshed
		///	</param>
		/// <param name="nowTime">
		///	Always DateTime.Now, or the date to test with
		///	</param>
		/// <returns>
		///	True if the item was expired, otherwise false
		///	</returns>
		public static bool IsExtendedExpired ( string format,
			DateTime getTime, 
			DateTime nowTime )
		{
			try
			{
				#region Validate arguments

				if( Object.Equals(format, null) )
				{
					throw new ArgumentNullException( "format",
						CacheResources.ResourceManager [ 
							"RES_ExceptionNullFormat" ] );
				}
				if( Object.Equals(format, String.Empty) )
				{
					throw new  ArgumentOutOfRangeException( "format",
						CacheResources.ResourceManager [ 
							"RES_ExceptionEmptyFormat" ] );
				}

				#endregion

				ExtendedFormat parsedFormat = 
					(ExtendedFormat)parsedFormatCache[ format ];
				if( Object.Equals(parsedFormat, null) )
				{
					parsedFormat = new ExtendedFormat( format );
					lock(parsedFormatCache.SyncRoot)
					{
						parsedFormatCache[ format ] = parsedFormat;
					}
				}

				// Validate the format arguments
				bool isDataExpired = parsedFormat.IsAlwaysExpire || ( 
					( parsedFormat.IsExpireEveryMinute || 
					ValidateMinute( parsedFormat.MinutesFormat(),
						getTime,
						nowTime) ) &&
					( parsedFormat.IsExpireEveryHour || 
					ValidateHour( parsedFormat.HoursFormat(),
						getTime,
						nowTime) ) &&
					( parsedFormat.IsExpireEveryDay || 
					ValidateDayOfMonth( parsedFormat.DaysFormat(),
						getTime, 
						nowTime) ) &&
					( parsedFormat.IsExpireEveryMonth || 
					ValidateMonth( parsedFormat.MonthsFormat(),
						getTime, 
						nowTime) ) &&
					( parsedFormat.IsExpireEveryDayOfWeek || 
					ValidateDayOfWeek( parsedFormat.DaysOfWeekFormat(),
						getTime, 
						nowTime) ) );

					return isDataExpired;
			}
			catch( Exception genException )
			{
				ExceptionManager.Publish( genException );
				throw;
			}			
		}
		#endregion IsExtendedExpired

		#endregion

		#region Private Methods

		#region ValidateMinute
		/// <summary>
		///	Validates the extended format minute argument. 
		/// </summary>
		/// <remarks>
		///	bool isValid = ValidateMinute( minutes, getTime, nowTime );
		/// </remarks>
		/// <param name="minutes">
		///	The extended format minute argument
		/// </param>
		/// <param name="getTime">
		///	The time when the item has been refreshed
		/// </param>
		/// <param name="nowTime">
		///	Always DateTime.Now, or the date to test with
		/// </param>
		/// <returns>
		///	True if the item was expired, otherwise false
		/// </returns>
		private static bool ValidateMinute ( int[] minutes,
			DateTime getTime, 
			DateTime nowTime )
		{
			try
			{
				if( Object.Equals(minutes, null) )
				{
					throw new ArgumentNullException( "minutes",
						CacheResources.ResourceManager [
						"RES_ExceptionNullMinutes"] );
				}

				// Validates the minute argument
				TimeSpan span = nowTime.Subtract( getTime );
				foreach( int minute in minutes )
				{				
					if( span.TotalMinutes >= 60 )
					{
						return true;
					}
					else 
					{
						if( ! Object.Equals(getTime.Hour, nowTime.Hour) && 
							nowTime.Minute >= minute  ||
							( getTime.Minute < minute && 
							nowTime.Minute >= minute ) )
						{
							return true;
						}
					}
				}
			}
			catch( Exception genException )
			{
				ExceptionManager.Publish( genException );
				throw;
			}
			return false;
		}

		#endregion ValidateMinute

		#region ValidateHour

		/// <summary>
		///	Validates the extended format hour argument. 
		/// </summary>
		/// <remarks>
		///	bool isValid = ValidateHour( hours, getTime, nowTime );
		/// </remarks>
		/// <param name="hours">
		///	The extended format hour argument
		/// </param>
		/// <param name="getTime">
		///	The time when the item has been refreshed
		/// </param>
		/// <param name="nowTime">
		///	Always DateTime.Now, or the date to test with
		/// </param>
		/// <returns>
		///	True if the item was expired, otherwise false
		/// </returns>
		private static bool ValidateHour ( int[] hours,
			DateTime getTime, 
			DateTime nowTime )
		{
			TimeSpan span;
			try
			{
				if( Object.Equals(hours, null) )
				{
					throw new ArgumentNullException( "hours",
						CacheResources.ResourceManager [
							"RES_ExceptionNullHours"] );
				}

				// Validates the hour argument			
				foreach( int hour in hours )
				{
					span = nowTime.Subtract( getTime );
					if( span.TotalHours >= 24 )
					{
						return true;
					}
					else if(! Object.Equals(nowTime.Day, getTime.Day) &&
						nowTime.Hour >= hour  ||
						( getTime.Hour <= hour && nowTime.Hour >= hour ))
					{
						return true;
					}
				}
			}
			catch( Exception genException )
			{
				ExceptionManager.Publish( genException );
				throw;
			}
			return false;			
		}

		#endregion ValidateHour

		#region ValidateDayOfMonth
		
		/// <summary>
		///	Validates the extended format day of month argument.  
		/// </summary>
		/// <remarks>
		///	bool isValid = ValidateDayOfMonth( days, getTime, nowTime );
		/// </remarks>
		/// <param name="days">
		///	The extended format day of month argument
		/// </param>
		/// <param name="getTime">
		///	The time when the item has been refreshed
		/// </param>
		/// <param name="nowTime">
		///	Always DateTime.Now, or the date to test with
		/// </param>
		/// <returns>
		///	True if the item was expired, otherwise false
		/// </returns>
		private static bool ValidateDayOfMonth ( int[] days,
			DateTime getTime,
			DateTime nowTime )
		{
			TimeSpan span;
			try
			{
				if( Object.Equals(days, null) )
				{
					throw new ArgumentNullException( "days",
						CacheResources.ResourceManager [
							"RES_ExceptionNullDays" ] );
				}

				// Validates the day of month argument			
				foreach( int day in days )
				{
					span = nowTime.Subtract( getTime );
					if( span.TotalDays >= 30 )
					{
						return true;
					}
					else 
					{
						if( ! Object.Equals(getTime.Month, nowTime.Month) &&
							nowTime.Day >= day  ||
							( getTime.Day <= day && 
							nowTime.Day >= day )) 
							
						{
							return true;
						}				
					}
				}
			}
			catch( Exception genException )
			{
				ExceptionManager.Publish( genException );
				throw;
			}
			return false;			
		}
		#endregion ValidateDayOfMonth

		#region ValidateMonth
		/// <summary>
		///	Validates the extended format month argument.  
		/// </summary>
		/// <remarks>
		///	bool isValid = ValidateMonth( months, getTime, nowTime );
		/// </remarks>
		/// <param name="months">
		///	The extended format month argument
		/// </param>
		/// <param name="getTime">
		///	The time when the item has been refreshed
		/// </param>
		/// <param name="nowTime">
		///	Always DateTime.Now, or the date to test with
		/// </param>
		/// <returns>
		///	True if the item was expired, otherwise false
		/// </returns>
		private static bool ValidateMonth ( int[] months,
			DateTime getTime,
			DateTime nowTime )
		{
			TimeSpan span;
			try
			{
				if( Object.Equals(months, null) )
				{
					throw new ArgumentNullException( "months",
						CacheResources.ResourceManager [
						"RES_ExceptionNullMonths" ] );
				}

				// Validates the month argument			
				foreach( int month in months )
				{
					span = nowTime.Subtract( getTime );
					if( span.TotalDays >= 364 )
					{
						return true;
					}
					else
					{
						if( ! Object.Equals(nowTime.Year, getTime.Year) &&
							nowTime.Month >= month ||
							( getTime.Month <= month && 
							nowTime.Month >= month ) )
						{
							return true;
						}				
					}
				}
			}
			catch( Exception genException )
			{
				ExceptionManager.Publish( genException );
				throw;
			}
			return false;			
		}
		#endregion ValidateMonth

		#region ValidateDayOfWeek
		/// <summary>
		///	Validates the extended format day of week argument.  
		/// </summary>
		/// <remarks>
		///	bool isValid = ValidateDayOfWeek(daysOfWeek, getTime, nowTime);
		/// </remarks>
		/// <param name="daysOfWeek">
		///	The extended format day of week argument
		/// </param>
		/// <param name="getTime">
		///	The time when the item has been refreshed
		/// </param>
		/// <param name="nowTime">
		///	Always DateTime.Now, or the date to test with
		/// </param>
		/// <returns>
		///	True if the item was expired, otherwise false
		/// </returns>
		private static bool ValidateDayOfWeek ( int[] daysOfWeek,
			DateTime getTime, 
			DateTime nowTime )
		{
			DateTime tempDateTime;
			TimeSpan span;
			try
			{
				if( Object.Equals(daysOfWeek, null) )
				{
					throw new ArgumentNullException( "daysOfWeek",
						CacheResources.ResourceManager [
							"RES_ExceptionNullDaysOfWeek" ] );
				}

				// Validates the day of week argument

				foreach(int day in daysOfWeek )
				{
					span = nowTime.Subtract( getTime );
					if( span.TotalDays >= 7 )
					{
						return true;
					}
					else
					{
						tempDateTime = nowTime;
						for( int counter = 0; 
							tempDateTime <= tempDateTime.AddDays( 0 );
							counter++ )
						{
							if( Object.Equals((int)tempDateTime.DayOfWeek,
								day) )
							{
								return true;
							}
						}
						return false;
					}
				}
			}
			catch( Exception genException )
			{
				ExceptionManager.Publish( genException );
				throw;
			}
			return false;			
		}
		#endregion ValidateDayOfWeek

		#endregion
	}
    #endregion
}
