//===============================================================================
// Microsoft Caching Application Block for .NET
// http://msdn.microsoft.com/library/default.asp?url=/library/en-us/dnpag/html/CachingBlock.asp
//
// ExtendedFormatTime.cs
// This class provides tests if a data item was expired using a extended format.
//
//===============================================================================
// Copyright (C) 2003 Microsoft Corporation
// All rights reserved.
// THIS CODE AND INFORMATION IS PROVIDED "AS IS" WITHOUT WARRANTY
// OF ANY KIND, EITHER EXPRESSED OR IMPLIED, INCLUDING BUT NOT
// LIMITED TO THE IMPLIED WARRANTIES OF MERCHANTABILITY AND/OR
// FITNESS FOR A PARTICULAR PURPOSE.
//===============================================================================


using System;
using System.Globalization;
using System.Text;
using System.Runtime.Serialization;
using System.Security.Permissions;

// Used for Handling Exceptions
using Microsoft.ApplicationBlocks.ExceptionManagement;

namespace Microsoft.ApplicationBlocks.Cache.Expirations
{
	/// <summary>
	///	This provider tests if a item was expired using a extended format.
	/// </summary>
	[Serializable]
	public class ExtendedFormatTime : ICacheItemExpiration, ISerializable
	{
		#region Private Members
		private string extendedFormat;
		private DateTime lastUsedTime;
		private bool isExpired = false;
		#endregion

		#region Constructor
		/// <summary>
		///	Creates an instance of the class.
		/// </summary>
		public ExtendedFormatTime()
		{}

		/// <summary>
		///	Convert the input format to the extented time format.
		/// </summary>
		/// <param name="format">
		///	This contains the expiration information
		/// </param>
		public ExtendedFormatTime( string timeFormat )
		{
			try
			{
				#region Throwing Argument Exception

				if(Object.Equals(timeFormat, null))
				{
					throw(new ArgumentNullException("timeFormat",
						CacheResources.ResourceManager[
						"RES_ExceptionNullTimeFormat"]));
				}
				if(timeFormat.Length == 0)
				{
					throw(new ArgumentOutOfRangeException("timeFormat",
						CacheResources.ResourceManager[
						"RES_ExceptionRangeTimeFormat"]));
				}

				#endregion

				// Validate the time format passed by the client
				ExtendedFormat format = new ExtendedFormat(timeFormat);

				// Get the modified extended format
				extendedFormat = GetModifiedFormat(timeFormat);

				// Convert to UTC in order to compensate for time zones		
				lastUsedTime = DateTime.Now.ToUniversalTime();
			}	
			catch(Exception genException)
			{
				ExceptionManager.Publish(genException);
				throw;
			}
		}
		#endregion

		#region Implementation of ICacheItemExpiration

		/// <summary>
		///	Specifies if item has expired or not.
		/// </summary>
		/// <remarks>
		///	bool ICacheItemExpiration.HasExpired();
		/// </remarks>
		/// <returns>
		///	Returns true if the data is expired otherwise false
		/// </returns>
		bool ICacheItemExpiration.HasExpired()
		{
			try
			{
				// Convert to UTC in order to compensate for time zones		
				DateTime nowDateTime = DateTime.Now.ToUniversalTime();

				// Check expiration
				if( ExtendedFormatHelper.IsExtendedExpired( 
					extendedFormat, lastUsedTime, nowDateTime ) )
				{
					isExpired = true;
				}
				else
				{
					isExpired = false;
				}
			}
			catch( Exception genException )
			{
				ExceptionManager.Publish( genException );
				throw;
			}
			return isExpired;
		}

		/// <summary>
		///	Notifies that the item was recently used.
		/// </summary>
		void ICacheItemExpiration.Notify()
		{}

		/// <summary>
		///	This method sets the external dependency key.
		/// </summary>
		void ICacheItemExpiration.Key(string keyVal)
		{ }

		/// <summary>
		///	Event to indicate the cache item expiration.
		/// </summary>
		public event ItemDependencyChangeEventHandler Change;

		#endregion

		#region ISerializable implementation 
		
		#region Public Method
		/// <summary>
		///	This method performs the serialziaton of members of the 
		///	current class.
		/// </summary>
		/// <param name="info">
		///	A SerializationInfo object which is deserialized by the 
		///	formatter and then passed to current constructor.
		/// </param>
		/// <param name="context">
		///	A StreamingContext that describes the source of the 
		///	serialized stream from where the Serialization object 
		///	is retrieved.
		/// </param>
		[SecurityPermissionAttribute( 
			 SecurityAction.Demand,
			 SerializationFormatter = true )]
		[SecurityPermissionAttribute(
			 SecurityAction.LinkDemand, 
			 SerializationFormatter = true )]
		public void GetObjectData( SerializationInfo info,
			StreamingContext context )
		{
			try
			{
				info.AddValue("extendedFormat", extendedFormat );
				info.AddValue("lastUsedTime", lastUsedTime );
			}
			catch( Exception genException )
			{
				ExceptionManager.Publish( genException );
				throw;
			}
		}
		#endregion

		#region Protected Method
		/// <summary>
		///	This method performs the deserialziaton of members of the 
		///	current class.
		/// </summary>
		/// <param name="info">
		///	A SerializationInfo object which is deserialized by the 
		///	formatter and then passed to current constructor
		/// </param>
		/// <param name="context">
		///	A StreamingContext that describes the source of the 
		///	serialized stream from where the Serialization object 
		///	is retrieved
		/// </param>
		protected ExtendedFormatTime ( SerializationInfo info,
			StreamingContext context ) 
		{
			try
			{
				extendedFormat = info.GetString( "extendedFormat" );
				lastUsedTime = Convert.ToDateTime( 
					info.GetValue( "lastUsedTime", typeof(DateTime) ),
					DateTimeFormatInfo.CurrentInfo );
			}
			catch( Exception genException )
			{
				ExceptionManager.Publish( genException );
				throw;
			}
		}
		#endregion

		#region Private methods
		
		/// <summary>
		///	This function modifies the format passed from the client. If the 
		///	minute format is wild card character, it converts it to current minute
		///	plus one so that the cache item expires after a minute.
		/// </summary>
		/// <remarks>
		///	string modifiedFormat = GetModifiedFormat(format);
		/// </remarks>
		/// <param name="format">
		/// The format to be modified
		/// </param>
		/// <returns>
		///	Returns the modified extended format
		/// </returns>
		private string GetModifiedFormat(string format)
		{
			const int MAX_MINUTE = 59;
			StringBuilder modifiedFormat = new StringBuilder();

			// Check whether the first character is wild card
			if(format.IndexOf( 
				CacheResources.ResourceManager["RES_WildcardAll"]) == 0)
			{
				int nowMinute = DateTime.Now.ToUniversalTime().Minute + 1;

                // Replace the first character with the current 
                // minute value + 1. If it is greater than 59 use zero. 
                // i.e it will expire in the next hour.
				if(nowMinute > MAX_MINUTE)
				{
					nowMinute = 0;
				}
				modifiedFormat.Append(nowMinute.ToString());
				modifiedFormat.Append(format.Substring(1, format.Length - 1));
                return modifiedFormat.ToString();
			}
			else
			{
				return format;
			}
		}

		#endregion

		#endregion
	}    
}
