//===============================================================================
// Microsoft Caching Application Block for .NET
// http://msdn.microsoft.com/library/default.asp?url=/library/en-us/dnpag/html/CachingBlock.asp
//
// FileDependency.cs
// This class tracks a file cache dependency.
//
//===============================================================================
// Copyright (C) 2003 Microsoft Corporation
// All rights reserved.
// THIS CODE AND INFORMATION IS PROVIDED "AS IS" WITHOUT WARRANTY
// OF ANY KIND, EITHER EXPRESSED OR IMPLIED, INCLUDING BUT NOT
// LIMITED TO THE IMPLIED WARRANTIES OF MERCHANTABILITY AND/OR
// FITNESS FOR A PARTICULAR PURPOSE.
//===============================================================================


using System;
using System.IO;
using System.Runtime.Serialization;
using System.Security.Permissions;

// Used for Handling Exceptions
using Microsoft.ApplicationBlocks.ExceptionManagement;

namespace Microsoft.ApplicationBlocks.Cache.Expirations
{
	/// <summary>
	///	This class tracks a file cache dependency.
	/// </summary>
	[Serializable]
	public class FileDependency : ICacheItemExpiration, ISerializable
	{
		#region Private members

		private string keyValue;	
		private string dependencyFileName;
		private DateTime lastAccessedTime;

		#endregion

		#region Constructor
		/// <summary>
		///	Constructor with one argument.
		/// </summary>
		/// <param name="fullFileName">
		///	Indicates the file name of the file
		/// </param>
		public FileDependency( string fullFileName )
		{
			try
			{
				
				string path;
				string fileName;

				#region Throwing Argument Exception

				if(Object.Equals(fullFileName, null))
				{
					throw new ArgumentNullException("fullFileName",
						CacheResources.ResourceManager[
						"RES_ExceptionNullFileName"]);
				}
				if(fullFileName.Length == 0)
				{
					throw(new ArgumentOutOfRangeException("fullFileName",
						CacheResources.ResourceManager[
						"RES_ExceptionEmptyFileName"]));
				}
				if(! File.Exists(fullFileName))
				{
					throw(new ArgumentException(CacheResources.
						ResourceManager["RES_ExceptionInvalidFileName"], 
						"fullFileName"));
				}

				#endregion

				// Validate File
				FileInfo fileDependencyInfo = new FileInfo(fullFileName);
				if (!fileDependencyInfo.Exists)
				{
					throw new FileNotFoundException();
				}

				// Get Path from full file name
				path = Path.GetDirectoryName(fullFileName);
				// Get file name from full file name
				fileName = Path.GetFileName(fullFileName);

				dependencyFileName = fullFileName;
				// Set the last accessed time
				lastAccessedTime = File.GetLastAccessTime(fullFileName);

			}
			catch( Exception genException )
			{
				ExceptionManager.Publish( genException );
				throw;
			}
		}
		#endregion

		#region ICacheItemExpiration Implementation

		/// <summary>
		///	Event to indicate the cache item expiration.
		/// </summary>
		public event ItemDependencyChangeEventHandler Change;
		
		/// <summary>
		///	This method sets the external dependency key.
		/// </summary>
		void ICacheItemExpiration.Key(string keyVal)
		{
			try
			{
				#region Throwing Argument Exception

				if( Object.Equals(keyVal, null) )
				{
					throw new ArgumentNullException( "keyVal",
						CacheResources.ResourceManager[ 
						"RES_ExceptionNullKey" ] );
				}
				if( keyVal.Length == 0)
				{
					throw new ArgumentOutOfRangeException( "keyVal",
						CacheResources.ResourceManager[ 
						"RES_ExceptionEmptyKey" ] );
				}

				#endregion

				keyValue = keyVal;
			}
			catch(Exception genException)
			{
				ExceptionManager.Publish(genException);
				throw;
			}
		}
		
		/// <summary>
		///	Specifies if the item has expired or not.
		/// </summary>
		bool ICacheItemExpiration.HasExpired()
		{
			try
			{
				// Compare the Filedependency object's last accessed time 
				// value with the last accessed time value. If they are not
				// equal return true, otherwise false.
				if(DateTime.Compare(lastAccessedTime, 
					File.GetLastAccessTime(dependencyFileName)) != 0)
				{
					return true;
				}
			}
			catch(Exception genException)
			{
				ExceptionManager.Publish(genException);
			}
			return false;

		}
		
		/// <summary>
		///	Notifies that the item was recently used.
		/// </summary>
		void ICacheItemExpiration.Notify()
		{}

		#endregion
		
		#region Serialization functions

		#region Public method
		/// <summary>
		///	This method performs the serialization of members of the 
		///	current class.
		/// </summary>
		/// <param name="info">
		///	A SerializationInfo object which is deserialized by the 
		///	formatter and then passed to current constructor
		/// </param>
		/// <param name="context">
		///	A StreamingContext that describes the source of the 
		///	serialized stream from where the Serialization object 
		///	is retrieved
		/// </param>
		[SecurityPermissionAttribute( 
			 SecurityAction.Demand, SerializationFormatter = true)]
		[SecurityPermissionAttribute( 
			 SecurityAction.LinkDemand, SerializationFormatter = true)]
		public void GetObjectData( SerializationInfo info,
			StreamingContext context )
		{
			// Adds the file name and last accessed time 
			// into the SerializationInfo, 
            // where it is associated with the name key
			info.AddValue("fileName", dependencyFileName);
			info.AddValue("lastAccessedTime", lastAccessedTime);
		}
		#endregion

		#region Constructor
		/// <summary>
		///	This method performs the deserialization of members of the 
		///	current class.
		/// </summary>
		/// <param name="info">
		///	A SerializationInfo object which is deserialized by the 
		///	formatter and then passed to current constructor
		/// </param>
		/// <param name="context">
		///	A StreamingContext that describes the source of the 
		///	serialized stream from where the Serialization object 
		///	is retrieved
		/// </param>
		protected FileDependency ( SerializationInfo info,
			StreamingContext context ) 
		{
			try
			{
				// Getting the value of file name and
				// last accessed time
				dependencyFileName = info.GetString("fileName");
				lastAccessedTime = Convert.ToDateTime(
					info.GetValue("lastAccessedTime", 
					typeof(DateTime)));
				
			}
			catch( Exception genException )
			{
				ExceptionManager.Publish( genException );
			}
		}
		#endregion
		
		#endregion
}
}
