//===============================================================================
// Microsoft Caching Application Block for .NET
// http://msdn.microsoft.com/library/default.asp?url=/library/en-us/dnpag/html/CachingBlock.asp
//
// SlidingTime.cs
// This class provides tests if a item was expired using a time slice schema.
//
//===============================================================================
// Copyright (C) 2003 Microsoft Corporation
// All rights reserved.
// THIS CODE AND INFORMATION IS PROVIDED "AS IS" WITHOUT WARRANTY
// OF ANY KIND, EITHER EXPRESSED OR IMPLIED, INCLUDING BUT NOT
// LIMITED TO THE IMPLIED WARRANTIES OF MERCHANTABILITY AND/OR
// FITNESS FOR A PARTICULAR PURPOSE.
//===============================================================================


using System;
using System.Globalization;
using System.Runtime.Serialization;
using System.Security.Permissions;

// Used for Handling Exceptions
using Microsoft.ApplicationBlocks.ExceptionManagement;

namespace Microsoft.ApplicationBlocks.Cache.Expirations
{
	/// <summary>
	///	This provider tests if a item was expired using a time slice schema.
	/// </summary>
	[Serializable]
	public class SlidingTime : ICacheItemExpiration, ISerializable
	{
		#region Private members

		private DateTime timeLastUsed;
		private TimeSpan itemSlidingExpiration;
		private long expirationTicks = 0;
		private string keyValue;

		#endregion

		#region Constructor

		/// <summary>
		///	Create an instance of this class with the timespan for expiration.
		/// </summary>
		/// <param name="slidingExpiration">
		///	Expiration time span
		/// </param>
		public SlidingTime ( TimeSpan slidingExpiration )
		{
			try
			{
				// Check the expiration for null
				if(object.Equals(slidingExpiration, null))
				{
                    throw new ArgumentNullException( 
                        "slidingExpiration", CacheResources.ResourceManager[ 
                        "RES_ExceptionNullSlidingExpiration"]);
				}
				if(!(slidingExpiration.TotalSeconds >= 1))
				{
					throw(new ArgumentOutOfRangeException("slidingExpiration",
						CacheResources.ResourceManager[
						"RES_ExceptionRangeSlidingExpiration"]));
				}
				itemSlidingExpiration = slidingExpiration;
			}
			catch(Exception genexception)
			{
				ExceptionManager.Publish(genexception);
				throw;
			}
		}

		#endregion

		#region Implementation of ICacheItemExpiration
		
		/// <summary>
		///	This method sets the external dependency key.
		/// </summary>
		void ICacheItemExpiration.Key(string keyVal)
		{
			try
			{
				#region Throwing Argument Exception

				if( Object.Equals(keyVal, null) )
				{
					throw new ArgumentNullException( "keyVal",
						CacheResources.ResourceManager[ 
						"RES_ExceptionNullKey" ] );
				}
				if( keyVal.Length == 0)
				{
					throw new ArgumentOutOfRangeException( "keyVal",
						CacheResources.ResourceManager[ 
						"RES_ExceptionEmptyKey" ] );
				}

				#endregion

				keyValue = keyVal;
			}
			catch(Exception genException)
			{
				ExceptionManager.Publish(genException);
				throw;
			}
		}

		/// <summary>
		///	Event to indicate the cache item expiration.
		/// </summary>
		public event ItemDependencyChangeEventHandler Change;

		/// <summary>
		///	Specifies if item has expired or not.
		/// </summary>
		bool ICacheItemExpiration.HasExpired ()
		{
			bool isItemexpired = CheckSlidingExpiration ( DateTime.Now,
				timeLastUsed,
				itemSlidingExpiration );
			return isItemexpired;	
		}
		
		/// <summary>
		///	Notifies that the item was recently used.
		/// </summary>
		void ICacheItemExpiration.Notify()
		{
			timeLastUsed = DateTime.Now;			
		}

		#endregion

		#region Private functions

		/// <summary>
		///	Check whether the sliding time has expired.
		/// </summary>
		/// <remarks>
		/// bool isSlidingExpiration = CheckSlidingExpiration(nowDateTime,
		/// lastUsed, slidingExpiration);
		/// </remarks>
		/// <param name="nowDateTime">
		///	Current time
		/// </param>
		/// <param name="lastUsed">
		///	The last time when the item has been used
		/// </param>
		/// <param name="slidingExpiration">
		///	The span of sliding expiration
		/// </param>
		/// <returns>
		///	True if the item was expired, otherwise false
		/// </returns>
		private bool CheckSlidingExpiration( DateTime nowDateTime,
			DateTime lastUsed, 
			TimeSpan slidingExpiration )
		{
			// Variable Declarations
			DateTime tmpNowDateTime;
			DateTime tmpLastUsed;
			bool isExpired = false;
			
			try
			{
				// Convert to UTC in order to compensate for time zones
				tmpNowDateTime = nowDateTime.ToUniversalTime();

				// Calculate the ticks only once
				if( ! Object.Equals(expirationTicks, 0) )
				{
					// Convert to UTC in order to compensate for time zones
					tmpLastUsed = lastUsed.ToUniversalTime();

					expirationTicks = tmpLastUsed.Ticks 
						+ slidingExpiration.Ticks;
				}

				#region Check expiration
				
				if(tmpNowDateTime.Ticks > expirationTicks )
				{
					isExpired = true;
				}
				else
				{
					isExpired = false;
				}

				#endregion
			}
			catch( Exception genException )
			{
				ExceptionManager.Publish( genException );
				throw;
			}
			return isExpired;
		}

		#endregion

		#region ISerializable implementation 
		
		/// <summary>
		///	This method performs the deserialziaton of members of the current 
		///	class.
		/// </summary>
		/// <param name="info">
		///	A SerializationInfo object which is deserialized by the formatter 
		///	and then passed to current constructor
		/// </param>
		/// <param name="context">
		///	A StreamingContext that describes the source of the serialized 
		///	stream from where the Serialization object is retrieved
		/// </param>
		protected SlidingTime ( SerializationInfo info,
			StreamingContext context ) 
		{
			try
			{
				timeLastUsed = Convert.ToDateTime(
					info.GetValue( "lastUsed", typeof(DateTime) ), 
					DateTimeFormatInfo.CurrentInfo );
				itemSlidingExpiration = (TimeSpan)info.GetValue( 
					"slidingExpiration",
					typeof(TimeSpan) );
			}
			catch(Exception genException)
			{
				ExceptionManager.Publish(genException);
				throw;
			}
		}

		/// <summary>
		///	This method performs the serialziaton of members of the current 
		///	class.
		/// </summary>
		/// <param name="info">
		///	A SerializationInfo object which is deserialized by the formatter 
		///	and then passed to current constructor
		/// </param>
		/// <param name="context">
		///	A StreamingContext that describes the source of the serialized 
		///	stream from where the Serialization object is retrieved
		/// </param>
		[SecurityPermissionAttribute( SecurityAction.Demand, 
			 SerializationFormatter = true )]
		[SecurityPermissionAttribute( SecurityAction.LinkDemand,
			 SerializationFormatter = true )]
		public void GetObjectData( SerializationInfo info,
			StreamingContext context )
		{
			try
			{
				info.AddValue("slidingExpiration", itemSlidingExpiration);
				info.AddValue("lastUsed", timeLastUsed);
			}
			catch(Exception genException)
			{
				ExceptionManager.Publish(genException);
				throw;
			}
		}
		#endregion
	}
}