//===============================================================================
// Microsoft Caching Application Block for .NET
// http://msdn.microsoft.com/library/default.asp?url=/library/en-us/dnpag/html/CachingBlock.asp
//
// InterfaceDefinitions.cs
// This class has all the interfaces with their method definitions required for
// the caching operations.
//
//===============================================================================
// Copyright (C) 2003 Microsoft Corporation
// All rights reserved.
// THIS CODE AND INFORMATION IS PROVIDED "AS IS" WITHOUT WARRANTY
// OF ANY KIND, EITHER EXPRESSED OR IMPLIED, INCLUDING BUT NOT
// LIMITED TO THE IMPLIED WARRANTIES OF MERCHANTABILITY AND/OR
// FITNESS FOR A PARTICULAR PURPOSE.
//===============================================================================


using System;
using System.Collections;
using System.Xml;

namespace Microsoft.ApplicationBlocks.Cache
{
	#region ICacheItemExpiration

	/// <summary>
	///	Allows end users to implement their own cache item expiration schema.
	/// </summary>
	public interface ICacheItemExpiration
	{
		/// <summary>
		///	This method is used to provide two types of expirations. The 
		///	expirations based on an asynchronous notification (which are not 
		///	polled), and the expirations based on a polling thread which calls 
		///	the HasExpired method.
		/// </summary>
		
		#region Synchronous expiration

		/// <summary>
		///	Specifies if item has expired or not.
		/// </summary>
		bool HasExpired();
		
		/// <summary>
		///	Notifies that the item was recently used.
		/// </summary>
		void Notify();	
	
		#endregion

		#region Asynchronous expiration (user for external dependencies)

		/// <summary>
		///	This method sets the external dependency key.
		/// </summary>
		void Key(string keyVal);
				
		/// <summary>
		///	Event to indicate the cache item expiration.
		/// </summary>
		event ItemDependencyChangeEventHandler Change;

		#endregion
	}

	#endregion

	#region ICacheMetadata

	/// <summary>
	///	Allows end users to implement their own cache metadata management 
	///	schema.
	/// </summary> 
	public interface ICacheMetadata
	{
		/// <summary>
		///	Adds new data to the cache metadata storage.
		/// </summary>
		void Add ( string key, 
			ICacheItemExpiration[]	expirations,
			CacheItemPriority priority );
		
		/// <summary>
		///	Removes the element with the specified key from the metadata 
		///	storage.
		/// </summary>
		void Remove	( string key );
		
		/// <summary>
		///	Gets all metadata from the metadata storage.
		/// </summary>
		Hashtable GetMetadata ();
		
		/// <summary>
		///	Removes all metadata from the metadata storage.
		/// </summary>
		void Flush ();
	}

	/// <summary>
	///	Allows end users to implement their own cache metadata management 
	///	schema and serialize the callback.
	/// </summary> 
	/// <remarks>
	/// If the call back is to be serialized it must be a reference to a static member becaue instance based members can not be called back.
	/// </remarks>
	public interface ICacheMetadataWithCallback : ICacheMetadata {
		/// <summary>
		///	Adds new data to the cache metadata storage.
		/// </summary>
		void Add ( string key, 
			ICacheItemExpiration[]	expirations,
			CacheItemPriority priority,
			CacheItemRemovedCallback itemRemovedCallBack);
	}

	#endregion

	#region IScavengingAlgorithm

	/// <summary>
	///	Allows end users to implement their own scavenging algorithm.
	/// </summary>
	public interface IScavengingAlgorithm
	{
		/// <summary>
		///	Inits the scavenging algorithm.
		/// </summary>
		void Init ( CacheService cacheService,
			ICacheStorage cacheStorage, 
			ICacheMetadata cacheMetadata,
			XmlNode config );
		
		/// <summary>
		///	Notifies that the element with the specified key was recently used.
		/// </summary>
		void Notify	( string key );
		
		/// <summary>
		///	Executes the algorithm.
		/// </summary>
		void Execute ();
		
		/// <summary>
		///	Adds a new element to the item algorithm list. This list is used 
		///	when the algorithm is executed.
		/// </summary>
		void Add ( string key, CacheItemPriority priority );
		
		/// <summary>
		///	Removes the element with the specified key from the item 
		///	algorithm list.
		/// </summary>
		void Remove ( string key );
		
		/// <summary>
		///	Removes all elements from the item algorithm list.
		/// </summary>
		void Flush ();
	}

	#endregion

	#region ICacheStorage

	/// <summary>
	///	Allows end users to implement their own cache management storage. All 
	///	storage providers must implement this interface.
	/// </summary>
	public interface ICacheStorage
	{
		/// <summary>
		///	Inits the storage provider.
		/// </summary>
		void Init( XmlNode config );
		
		/// <summary>
		///	Adds an element with the specified key and value into the storage.
		/// </summary>
		void Add ( string key, object keyData);
		
		/// <summary>
		///	Removes all elements from the storage.
		/// </summary>
		void Flush ();
		
		/// <summary>
		///	Gets the element with the specified key.
		/// </summary>
		object GetData ( string key);
		
		/// <summary>
		///	Removes the element with the specified key.
		/// </summary>
		void Remove	( string key);
		
		/// <summary>
		///	Updates the element with the specified key.
		/// </summary>
		void Update	( string key, object keyData);
		
		/// <summary>
		///	Gets the number of elements actually contained in the storage.
		/// </summary>
		int Size{ get; }
	}

	#endregion 

	#region IDataProtection

	/// <summary>
	///	Allows end users to implement their own cache item protection schema.
	/// </summary>
	public interface IDataProtection
	{
		/// <summary>
		///	Inits the data protection provider.
		/// </summary>
		void Init( XmlNode config );
		
		/// <summary>
		///	Encrypts a raw of bytes.
		/// </summary>
		byte[] Encrypt( byte[] plainValue );
		
		/// <summary>
		///	Decrypts a raw of bytes.
		/// </summary>
		byte[] Decrypt( byte[] cipherValue );
		
		/// <summary>
		///	Computes a hash for data validation.
		/// </summary>
		byte[] ComputeHash(byte[] plainValue );
	}

	#endregion

	#region IMmfReference

	/// <summary>
	///	To increment or decrement the reference count of the 
	///	memory mapped file object.
	/// </summary>
	public interface IMmfReference
	{

		/// <summary>
		///	Increase the reference count value by one.
		/// </summary>
		object AddReference(string key);

		/// <summary>
		///	Decrease the reference count by one.
		/// </summary>
		void RemoveReference(object mmfs);

	}

	#endregion
}