//===============================================================================
// Microsoft Caching Application Block for .NET
// http://msdn.microsoft.com/library/default.asp?url=/library/en-us/dnpag/html/CachingBlock.asp
//
// ProviderBaseClass.cs
// Base class for all providers settings within the cache configuration settings
// in the config file.
//
//===============================================================================
// Copyright (C) 2003 Microsoft Corporation
// All rights reserved.
// THIS CODE AND INFORMATION IS PROVIDED "AS IS" WITHOUT WARRANTY
// OF ANY KIND, EITHER EXPRESSED OR IMPLIED, INCLUDING BUT NOT
// LIMITED TO THE IMPLIED WARRANTIES OF MERCHANTABILITY AND/OR
// FITNESS FOR A PARTICULAR PURPOSE.
//===============================================================================


using System;
using System.Configuration;
using System.Globalization;
using System.Reflection;
using System.Xml;

// Used for Handling exceptions
using Microsoft.ApplicationBlocks.ExceptionManagement;

namespace Microsoft.ApplicationBlocks.Cache
{
	#region Provider class

	/// <summary>
	///	Base class for all providers settings within the cache configuration 
	///	settings in the config file.
	/// </summary>
	internal abstract class ProviderBaseClass
	{
		#region Declare field members
	
		/// <summary>
		///	The XmlNode for the provider definition on the XML configuration 
		///	file.
		/// </summary>
		protected XmlNode configProviderInfo;

		/// <summary>
		///	The provider's assembly name.
		/// </summary>
		protected string providerAssemblyName;

		/// <summary>
		///	The provider's class name.
		/// </summary>
		protected string providerClassName;

		/// <summary>
		///	The provider's type.
		/// </summary>
		protected Type providerType;

		#endregion

		#region Constructor
		
		/// <summary>
		///	Constructor to read all the provider settings from the config file.
		/// </summary>
		protected ProviderBaseClass( XmlNode providerInfo )
		{
			try
			{
				if(Object.Equals(providerInfo, null))
				{
					throw(new ArgumentNullException("providerInfo",
						CacheResources.ResourceManager[
						"RES_ExceptionNullProviderInfo"]));
				}
				configProviderInfo = providerInfo;

				// Check assembly name
				if( Object.Equals( configProviderInfo.Attributes[ 
					CacheResources.ResourceManager[ "RES_AssemblyName" ] ],
					null ) )
				{
					throw new ConfigurationException( 
						string.Format( 
						CultureInfo.CurrentCulture, 
						CacheResources.ResourceManager[ 
						"RES_ExceptionInvalidConfigurationProviderTypeInformationAssembly" ], 
						this.GetType().Name ) );
				}

				// Check class name
				if( Object.Equals( configProviderInfo.Attributes[ 
					CacheResources.ResourceManager[ "RES_ClassName" ] ],
					null ) )
				{
					throw new ConfigurationException( 
						string.Format( 
						CultureInfo.CurrentCulture, 
						CacheResources.ResourceManager[ 
						"RES_ExceptionInvalidConfigurationProviderTypeInformationClassName" ], 
						this.GetType().Name ) );
				}

				providerAssemblyName = configProviderInfo.Attributes[ 
					CacheResources.ResourceManager[ 
					"RES_AssemblyName" ] ].Value;
				// Check the assembly name provider
				if( Object.Equals(providerAssemblyName.Length, 0) )
				{
					throw new ConfigurationException( 
						string.Format(
						CultureInfo.CurrentCulture,
						CacheResources.ResourceManager[ 
						"RES_ExceptionInvalidConfigurationProviderTypeInformationAssembly" ],
						this.GetType().Name ) );
				}

				// Check the class name provider
				providerClassName = configProviderInfo.Attributes[ 
					CacheResources.ResourceManager[ "RES_ClassName" ] ].Value;
				if( providerClassName.Length == 0 )
				{
					throw new ConfigurationException( 
						string.Format( 
						CultureInfo.CurrentCulture,
						CacheResources.ResourceManager[ 
						"RES_ExceptionInvalidConfigurationProviderTypeInformationClassName" ],
						this.GetType().Name ) );
				}

				// Check the provider type
				Assembly assemblyInstance = 
					Assembly.LoadWithPartialName( 
					providerAssemblyName );
				providerType = assemblyInstance.GetType( 
					providerClassName,
					true );
			}
			catch( Exception genException )
			{
				throw new ConfigurationException( 
					string.Format( 
					CultureInfo.CurrentCulture,
					CacheResources.ResourceManager[ 
					"RES_ExceptionInvalidConfigurationProviderInformation" ], 
					this.GetType().Name, 
					providerAssemblyName, 
					providerClassName ), 
					genException );
			}
		}

		#endregion

		#region Public properties
		
		/// <summary>
		///	Gets the provider's assembly name.
		/// </summary>
		public string AssemblyName
		{ get{ return providerAssemblyName; } }
		
		/// <summary>
		///	Gets the provider's class name.
		/// </summary>
		public string ClassName
		{ get{ return providerClassName; } }

		/// <summary>
		///	Gets the value of the property.
		/// </summary>
		public string this[ string propertyName ]
		{
			get
			{
				try
				{
					XmlNode customProperty;
					string propertyValue = "";

					if( Object.Equals(propertyName, String.Empty) ) 						
					{
						throw new ArgumentOutOfRangeException( 
							"propertyName" ,
							CacheResources.ResourceManager[ 
								"RES_ExceptionEmptyPropertyName" ] );
					}
					if( Object.Equals(propertyName, null) )
					{
						throw new ArgumentNullException( "propertyName" ,
							CacheResources.ResourceManager[ 
							"RES_ExceptionNullPropertyName" ] );
					}

					customProperty = configProviderInfo.SelectSingleNode( 
						CacheResources.ResourceManager[ 
							"RES_CharacterForPropertyValue" ] 
						+ propertyName );

					if(! Object.Equals(customProperty, null) )
					{
						propertyValue = customProperty.Value;
					}

					customProperty = configProviderInfo.SelectSingleNode( 
						CacheResources.ResourceManager[ 
							"RES_CharacterForPropertyInnerText" ]
						+ propertyName );
					if(! Object.Equals(customProperty,null) )
					{
						propertyValue = customProperty.InnerText;
					}
					return propertyValue;
				}
				catch( Exception genException )
				{
					ExceptionManager.Publish( genException );
					throw;
				}				
			}
		}

		#endregion
	}

	#endregion
}