//===============================================================================
// Microsoft Caching Application Block for .NET
// http://msdn.microsoft.com/library/default.asp?url=/library/en-us/dnpag/html/CachingBlock.asp
//
// LruScavenging.cs
// This class manages scavenging operations by removing the cache items using
// the LRU (Least Recently Used) algorithm to ensure that the cache storage size
// does not exceed a certain limit.
//
//===============================================================================
// Copyright (C) 2003 Microsoft Corporation
// All rights reserved.
// THIS CODE AND INFORMATION IS PROVIDED "AS IS" WITHOUT WARRANTY
// OF ANY KIND, EITHER EXPRESSED OR IMPLIED, INCLUDING BUT NOT
// LIMITED TO THE IMPLIED WARRANTIES OF MERCHANTABILITY AND/OR
// FITNESS FOR A PARTICULAR PURPOSE.
//===============================================================================


using System;
using System.Collections;
using System.Configuration;
using System.Globalization;
using System.Xml;
using System.Collections.Specialized;

// Used for Handling Exceptions
using Microsoft.ApplicationBlocks.ExceptionManagement;

namespace Microsoft.ApplicationBlocks.Cache.Scavenging
{
	/// <summary>
	///	A Last recently used scavenging algorithm implementation.
	/// </summary>
	public class LruScavenging : IScavengingAlgorithm
	{
		#region Member variables

		private double storageUtilization;
		private HybridDictionary itemsLastUsed;
		private HybridDictionary itemsPriority;
		private CacheService cachingService;
		private ICacheStorage cacheStorage;
		private ICacheMetadata cacheMetadata;

		#endregion

		#region Constructor
		/// <summary>
		///	Constructor.
		/// </summary>
		public LruScavenging()
		{
		}
		#endregion

		#region Implementation of IScavengingAlgorithm

		#region Init
		/// <summary>
		///	Inits the scavenging algorithm. 
		/// </summary>
		/// <remarks>
		///	IScavengingAlgorithm.Init ( cachingService, cacheStorage, 
		///	cacheMetadata, configSection );
		/// </remarks>
		/// <param name="cachingService">
		///	Instance of CacheService
		/// </param>
		/// <param name="cacheStorage">
		///	Instance of ICacheStorage
		/// </param>
		/// <param name="cacheMetadata">
		///	Instance of ICacheMetadata
		/// </param>
		/// <param name="configSection">
		///	Represents a single node in the XML document
		/// </param>
		void IScavengingAlgorithm.Init ( CacheService cachingService, 
			ICacheStorage cacheStorage, 
			ICacheMetadata cacheMetadata, 
			XmlNode configSection )
		{

			try
			{
				int maxCacheStorageSize = 0;
				int cacheUtilizationToScavenge = 0;
				const int PERCENTAGE_CONVERTER = 100;
				
				#region Throwing Argument Exception

				if(Object.Equals(cachingService, null))
				{
					throw(new ArgumentNullException("cachingService",
						CacheResources.ResourceManager[
						"RES_ExceptionNullCacheService"]));
				}
				if(Object.Equals(cacheStorage, null))
				{
					throw(new ArgumentNullException("cacheStorage",
						CacheResources.ResourceManager[
						"RES_ExceptionNullCacheStorage"]));
				}

				#endregion

				this.cachingService = cachingService;
				this.itemsPriority = cachingService.cacheItemsPriority;
				this.cacheStorage = cacheStorage;
				this.cacheMetadata = cacheMetadata;

				itemsLastUsed = new HybridDictionary(false);
				if(! Object.Equals(configSection, null))
				{
					#region MaxStorageSize

					XmlNode nodeValue = configSection.SelectSingleNode(
						CacheResources.ResourceManager["RES_MaximumSize"]);
					string storageSize;
					double storage;

					if(Object.Equals(nodeValue, null))
					{
						nodeValue = configSection.SelectSingleNode(
							CacheResources.ResourceManager[
							"RES_TempMaximumSize"]);
						if(! Object.Equals(nodeValue, null))
						{
							storageSize = nodeValue.InnerText;
						}
						else
						{
							throw(new ConfigurationException(CacheResources.
								ResourceManager[
								"RES_ExceptionNullMaxStorageSize"]));
						}
					}
					else
					{
						storageSize = nodeValue.Value;
					}
					
					if(storageSize.Length == 0)
					{
						throw(new ConfigurationException(CacheResources.
							ResourceManager[
							"RES_ExceptionEmptyMaxStorageSize"]));
					}

					if(double.TryParse(storageSize, NumberStyles.Integer, 
						null, out storage))
					{
						if(storage > int.MaxValue)
						{
							throw(new ConfigurationException(CacheResources.
								ResourceManager[
								"RES_ExceptionMaxMaxStorageSize"]));
						}
						maxCacheStorageSize = int.Parse(storageSize);
						if(maxCacheStorageSize < 1)
						{
							throw(new ConfigurationException(CacheResources.
								ResourceManager[
								"RES_ExceptionMinMaxStorageSize"]));
						}
					}
					else
					{
						throw(new ConfigurationException(CacheResources.
							ResourceManager[
							"RES_ExceptionInvalidMaxStorageSize"]));
					}

					#endregion
		
					#region UtilationForscavenging

					nodeValue = configSection.SelectSingleNode( 
						CacheResources.ResourceManager[ 
							"RES_UtilizationForScavenging" ] );
					string utilization;
					double scavengingUtilization;
					if(Object.Equals(nodeValue, null))
					{
						nodeValue = configSection.SelectSingleNode(
							CacheResources.ResourceManager[
							"RES_TempUtilizationForScavenging"]);
						if(! Object.Equals(nodeValue, null))
						{
							utilization = nodeValue.InnerText;
						}
						else
						{
							throw(new ConfigurationException(CacheResources.
								ResourceManager[
								"RES_ExceptionNullUtilization"]));
						}
					}
					else
					{
						utilization = nodeValue.Value;
					}
					if(utilization.Length == 0)
					{
						throw(new ConfigurationException(CacheResources.
							ResourceManager["RES_ExceptionEmptyUtilization"]));
					}
					if(double.TryParse(utilization, NumberStyles.Integer, 
						null, out scavengingUtilization))
					{
						if(scavengingUtilization > int.MaxValue)
						{
							throw(new ConfigurationException(CacheResources.
								ResourceManager[
								"RES_ExceptionMaxUtilization"]));
						}
						cacheUtilizationToScavenge = int.Parse(utilization);
						if(cacheUtilizationToScavenge < 1)
						{
							throw(new ConfigurationException(CacheResources.
								ResourceManager[
								"RES_ExceptionMinUtilization"]));
						}
					}
					else
					{
						throw(new ConfigurationException(CacheResources.
							ResourceManager[
							"RES_ExceptionInvalidUtilization"]));
					}
				}
				#endregion

				else
				{
					// Default values
					maxCacheStorageSize = Convert.ToInt32(
						CacheResources.ResourceManager[
						"RES_MaxCacheStorageSize" ]);
					cacheUtilizationToScavenge = Convert.ToInt32(
						CacheResources.ResourceManager[
						"RES_CacheUtilizationToScavenge" ]);
				}
				storageUtilization = ( maxCacheStorageSize * 
					cacheUtilizationToScavenge / 
					PERCENTAGE_CONVERTER);
			}
			catch( Exception genException )
			{
				ExceptionManager.Publish( genException );
				throw;
			}
		}
		#endregion

		#region Execute

		/// <summary>
		///	Executes the algorithm.
		/// </summary>
		/// <remarks>
		///	IScavengingAlgorithm.Execute();
		/// </remarks>
		void IScavengingAlgorithm.Execute() 
		{
			// Get storage size
			int storageSize = cacheStorage.Size;

			try
			{
				if ( storageSize >= storageUtilization )
				{
					// Start removing
					while ( storageSize >= storageUtilization )
					{
						string key = GetLruItem();
						// Remove the item
						cacheStorage.Remove(key);
						if (! Object.Equals(cacheMetadata, null) )
						{
							lock(cacheMetadata)
							{
								cacheMetadata.Remove(key);
							}
						}

						cachingService.RemoveItem(key,
							CacheItemRemoveCause.Scavenged);

						lock(itemsLastUsed)
						{
							itemsLastUsed.Remove(key);
						}
						lock(itemsPriority)
						{
							itemsPriority.Remove(key);
						}
						storageSize = cacheStorage.Size;
					}				
				}
			}
			catch( Exception genException )
			{
				ExceptionManager.Publish( genException );
				throw;
			}
		}
		#endregion

		#region Notify
		/// <summary>
		///	Notifies that the element with the specified 
		///	key was recently used.
		/// </summary>
		/// <remarks>
		///	IScavengingAlgorithm.Notify( string key );
		/// </remarks>
		/// <param name="key">
		///	The unique key to identify the value
		/// </param>
		void IScavengingAlgorithm.Notify ( string key ) 
		{
			try
			{
				#region Validation for key
				if( Object.Equals(key, null) )
				{
					throw new ArgumentNullException( "key",
						CacheResources.ResourceManager[ 
							"RES_ExceptionNullKey" ] );
				}
				if( Object.Equals(key, String.Empty) )
				{
					throw new ArgumentOutOfRangeException( "key",
						CacheResources.ResourceManager[ 
							"RES_ExceptionEmptyKey" ] );
				}
				#endregion

				lock(itemsLastUsed)
				{
					if (itemsLastUsed.Contains(key))
					{
						itemsLastUsed[key] = DateTime.Now;
					}
				}
			}
			catch( Exception genException )
			{
				ExceptionManager.Publish( genException );
				throw;
			}
		}
		#endregion

		#region Add
		/// <summary>
		///	Adds a new element to the item algorithm list. 
		///	This list is used when the algorithm is executed.
		/// </summary>
		/// <remarks>
		///	IScavengingAlgorithm.Add ( key, priority );
		/// </remarks>
		/// <param name="key">
		///	The unique to identify the value
		/// </param>
		/// <param name="priority">
		///	Specifies the cache item priority levels
		/// </param>
		void IScavengingAlgorithm.Add ( string key, 
			CacheItemPriority priority ) 
		{
			try
			{
				#region Validation for input parameters
				if( Object.Equals(key, null) )
				{
					throw new ArgumentNullException( "key",
						CacheResources.ResourceManager[ 
						"RES_ExceptionNullKey" ] );
				}
				if( Object.Equals(key, String.Empty) )
				{
					throw new ArgumentOutOfRangeException( "key",
						CacheResources.ResourceManager[ 
						"RES_ExceptionEmptyKey" ] );
				}
				#endregion

				lock(itemsLastUsed)
				{
					itemsLastUsed[key] = DateTime.Now;
				}
				lock(itemsPriority)
				{
					itemsPriority[key] = priority;
				}
			}
			catch( Exception genException )
			{
				ExceptionManager.Publish( genException );
				throw;
			}
		}
		#endregion

		#region Remove
		/// <summary>
		///	Removes the element with the specied key from the 
		///	item algorithm list.
		/// </summary>
		/// <remarks>
		///	IScavengingAlgorithm.Remove ( string key );
		/// </remarks>
		/// <param name="key">
		///	The unique key to identify the value
		/// </param>
		void IScavengingAlgorithm.Remove ( string key ) 
		{
			try
			{
				#region Validation for key
				if( Object.Equals(key, null) )
				{
					throw new ArgumentNullException( "key",
						CacheResources.ResourceManager[ 
						"RES_ExceptionNullKey" ] );
				}
				if( Object.Equals(key, String.Empty) )
				{
					throw new ArgumentOutOfRangeException( "key",
						CacheResources.ResourceManager[ 
						"RES_ExceptionEmptyKey" ] );
				}
				#endregion

				lock(itemsLastUsed)
				{
					itemsLastUsed.Remove(key);
				}
				lock(itemsPriority)
				{
					itemsPriority.Remove(key);
				}
			}
			catch( Exception genException )
			{
				ExceptionManager.Publish( genException );
				throw;
			}
		}
		#endregion

		#region Flush
		/// <summary>
		///	Removes all elements from the item algorithm list.
		/// </summary>
		/// <remarks>
		///	IScavengingAlgorithm.Flush ();
		/// </remarks>
		void IScavengingAlgorithm.Flush() 
		{
			try
			{
				lock(itemsLastUsed)
				{
					itemsLastUsed.Clear();
				}
				lock(itemsPriority)
				{
					itemsPriority.Clear();
				}
			}
			catch( Exception genException )
			{
				ExceptionManager.Publish( genException );
				throw;
			}
		}
		#endregion

		#endregion

		#region Private method
		/// <summary>
		///	Gets the Least Recently used Item.
		/// </summary>
		/// <remarks>
		///	string lruItemKey = GetLruItem ();
		/// </remarks>
		/// <returns>
		///	The key that has to be removed
		/// </returns>
		private string GetLruItem ()
		{
			try
			{			
				string lruItemKey = "";
				DateTime tmpDateTime = DateTime.Now;

				lock(itemsLastUsed)
				{
					foreach ( DictionaryEntry dictEntry in itemsLastUsed )
					{
						if( DateTime.Compare(tmpDateTime,
							(DateTime)dictEntry.Value) > 0 )
						{
							tmpDateTime = (DateTime)dictEntry.Value;
							lruItemKey = dictEntry.Key.ToString();
						}
					}
				}
				return lruItemKey;
			}
			catch( Exception genException )
			{
				ExceptionManager.Publish( genException );
				throw;
			}			
		}
		#endregion
	}
}
