//===============================================================================
// Microsoft Caching Application Block for .NET
// http://msdn.microsoft.com/library/default.asp?url=/library/en-us/dnpag/html/CachingBlock.asp
//
// MmfCacheStorage.cs
// This class is used to cache data into a shared memory file stream.
//
//===============================================================================
// Copyright (C) 2003 Microsoft Corporation
// All rights reserved.
// THIS CODE AND INFORMATION IS PROVIDED "AS IS" WITHOUT WARRANTY
// OF ANY KIND, EITHER EXPRESSED OR IMPLIED, INCLUDING BUT NOT
// LIMITED TO THE IMPLIED WARRANTIES OF MERCHANTABILITY AND/OR
// FITNESS FOR A PARTICULAR PURPOSE.
//===============================================================================


using System;
using System.Collections;
using System.ComponentModel;
using System.Configuration;
using System.Globalization;
using System.IO;
using System.Threading;
using System.Xml;
using System.Collections.Specialized;
using System.Runtime.InteropServices; 
using System.Runtime.Serialization.Formatters.Binary;

// Used for Memory Mapped Files
using Microsoft.ApplicationBlocks.MemoryMappedFile;
// Used for Caching Operations
using Microsoft.ApplicationBlocks.Cache;
// Used for Data Protection
using Microsoft.ApplicationBlocks.Cache.DataProtection;
// Used for Handling Exception
using Microsoft.ApplicationBlocks.ExceptionManagement;

namespace Microsoft.ApplicationBlocks.Cache.Storages
{
	/// <summary>
	///	Sample storage provider used to cache data into a shared memory 
	///	file stream.
	/// </summary>
	/// <remarks>
	///	This provider uses the following attributes on the XML file:
	/// <list type="">
	///	<item>
	///	<b>Assembly</b> 
	///	Used to specify the provider assembly (Required)
	///	</item>
	///	<item>
	///	<b>ClassName</b>
	///	Used to specify the provider class name (Required)
	///	</item>
	///	<item>
	///	<b>BasePath</b> 
	///	Used to specify the folder where the files will be stored
	///	</item>
	///	<item>
	///	<b>Encrypted</b>
	///	Specifies if the cache item must be encrypted before
	///	it is stored (Optional)
	///	</item>
	///	<item>
	///	<b>Validated</b> 
	///	Specifies if a cache item validation must be done to 
	///	ensure data integrity (Optional)
	///	</item>
	/// </list>
	/// </remarks>
	[Serializable]
	public class MmfCacheStorage : ICacheStorage, IMmfReference
	{
		#region Constants
		private const int LENGTH = 4;
		private const string MMF_OBJECT_NAME = "mmf_Dictionary";
		#endregion

        #region Field members

		bool isEncrypted;
		bool isValidated;
		// Dictionary to store mmf objects
		HybridDictionary handleTable;
		MemoryMappedFileStream dictionaryMmfs;
		uint hybridDictionarySize = 0;

		#endregion

		#region Properties

		/// <summary>
		///	Gets the storage size.
		/// </summary>
		int ICacheStorage.Size
		{
			get
			{
				try
				{
					lock(handleTable)
					{
						handleTable = (HybridDictionary)
							GetDictionary(MMF_OBJECT_NAME);

						return handleTable.Count;
					}
				}
				catch( Exception genException )
				{
					ExceptionManager.Publish( genException );
					throw;
				}
			}
		}

		#endregion

		#region Constructor
		/// <summary>
		///	Default Constructor.
		/// </summary>
		public MmfCacheStorage()
		{
			
		}
		#endregion

		#region Implementation of ICacheStorage

		#region Init
		/// <summary>
		///	Inits the storage provider.
		/// </summary>
		/// <remarks>
		///	ICacheStorage.Init( config );
		/// </remarks>
		/// <param name="config">
		///	The configuration file node
		/// </param>
		void ICacheStorage.Init ( System.Xml.XmlNode config )
		{
			try
			{
				string validatedString;
				string encryptedString;
				string mmfDictionarySize;
				double storage;
		
				const string CONST_TRUE = "true";
				const string CONST_FALSE = "false";

				if(object.Equals(config, null))
				{
					throw new ArgumentNullException("config", 
						CacheResources.ResourceManager[ 
						"RES_ExceptionNullConfigSection"]);
				}

				#region Gets the Signed attribute

				if( ! Object.Equals( config.Attributes[
					CacheResources.ResourceManager[ 
					"RES_ConfigValidated" ] ], null ) ) 
				{
					validatedString = config.Attributes[CacheResources.
						ResourceManager["RES_ConfigValidated"]].Value ;
				}
				else
				{
					throw(new ConfigurationException(CacheResources.
						ResourceManager["RES_ExceptionEmptyNodeValidated"]));
				}
				if(Object.Equals(validatedString, null))
				{
					throw(new ConfigurationException(CacheResources.
						ResourceManager["RES_ExceptionNullValidated"]));
				}
				else if (validatedString.Length == 0)
				{
					throw(new ConfigurationException(CacheResources.
						ResourceManager["RES_ExceptionEmptyValidated"]));
				}
				else if(!(Object.Equals(validatedString.ToLower(), CONST_TRUE)
					|| Object.Equals(validatedString.ToLower(), CONST_FALSE)))
				{
					throw(new ConfigurationException(CacheResources.
						ResourceManager["RES_ExceptionInvalidValidated"]));
				}
				else
				{
					isValidated = bool.Parse(validatedString.ToLower(
						CultureInfo.CurrentCulture));
				}				 

				#endregion
			
				#region Gets the Encrypted attribute
				if( ! Object.Equals(config.Attributes[
					CacheResources.ResourceManager[
					"RES_ConfigEncrypted" ] ], null) ) 
				{
					encryptedString = config.Attributes[CacheResources.
						ResourceManager["RES_ConfigEncrypted"]].Value;
				}
				else
				{
					throw(new ConfigurationException(CacheResources.
						ResourceManager["RES_ExceptionEmptyNodeEncrypted"]));
				}

				if(Object.Equals(encryptedString, null))
				{
					throw(new ConfigurationException(CacheResources.
						ResourceManager["RES_ExceptionNullEncrypted"]));
				}
				else if(encryptedString.Length == 0)
				{
					throw(new ConfigurationException(CacheResources.
						ResourceManager["RES_ExceptionEmptyEncrypted"]));
				}
				else if(!(Object.Equals(encryptedString.ToLower(), CONST_TRUE)
					|| Object.Equals(encryptedString.ToLower(), CONST_FALSE)))
				{
					throw(new ConfigurationException(CacheResources.
						ResourceManager["RES_ExceptionInvalidEncrypted"]));
				}
				else
				{
					isEncrypted = bool.Parse(encryptedString.ToLower(
						CultureInfo.CurrentCulture));
				}
				#endregion

				#region Gets the MmfDictionarySize attribute

				if( ! Object.Equals(config.Attributes[
					CacheResources.ResourceManager[
					"RES_ConfigMmfDictionarySize" ] ], null) ) 
				{
					mmfDictionarySize = config.Attributes[CacheResources.
						ResourceManager["RES_ConfigMmfDictionarySize"]].Value;
				}
				else
				{
					throw(new ConfigurationException(CacheResources.
						ResourceManager[
						"RES_ExceptionEmptyNodeMmfDictionarySize"]));
				}
				if(mmfDictionarySize.Length == 0)
				{
					throw(new ConfigurationException(CacheResources.
						ResourceManager[
						"RES_ExceptionEmptyDictionarySize"]));
				}

				if(double.TryParse(mmfDictionarySize, NumberStyles.Integer, 
					null, out storage))
				{
					if(storage > uint.MaxValue)
					{
						throw(new ConfigurationException(CacheResources.
							ResourceManager[
							"RES_ExceptionMaxDictionarySize"]));
					}
					hybridDictionarySize = uint.Parse(mmfDictionarySize);
					if(hybridDictionarySize < 1)
					{
						throw(new ConfigurationException(CacheResources.
							ResourceManager[
							"RES_ExceptionMinDictionarySize"]));
					}
				}
				else
				{
					throw(new ConfigurationException(CacheResources.
						ResourceManager[
						"RES_ExceptionInvalidDictionarySize"]));
				}

				#endregion

				// Check whether the dictionary object is available in
				// the shared memory (created by some other process)

				// Synchronize the access
				lock(this)
				{
					IntPtr dictPtr = 
						MemoryMappedFileStream.OpenDictionary(
						MMF_OBJECT_NAME,
						MemoryProtection.PageReadWrite);

					handleTable = new HybridDictionary();

					if (dictPtr == IntPtr.Zero)
					{
						// Add the dictionary to shared memory because 
						// it is not available already
						AddDictionary(MMF_OBJECT_NAME);
					}
				}

			}
			catch( Exception genException )
			{
				ExceptionManager.Publish( genException );
				throw;
			}
		}
		#endregion

		#region Remove

		/// <summary>
		///	Removes the element with the specified key.
		/// </summary>
		/// <remarks>
		///	ICacheStorage.Remove( key );
		/// </remarks>
		/// <param name="key">
		///	The unique key to identify the value
		/// </param>
		void ICacheStorage.Remove ( string key )
		{
			if (object.Equals(key, null))
			{
				throw new ArgumentNullException("key", 
					CacheResources.ResourceManager["RES_ExceptionNullKey"]);
			}

			if( key.Length == 0)
			{
				throw new ArgumentOutOfRangeException("key", 
					CacheResources.ResourceManager[ 
						"RES_ExceptionEmptyKey"]);
			}

			// Synchronize access to the memory just in case 
			// some other thread is reading while another 
			// removes the cache item
			Mutex mutexSync = new Mutex(false, GetHandleName(key));
			try
			{
				mutexSync.WaitOne();
				// Get the memory mapped file stream object
				lock(handleTable)
				{
					handleTable = (HybridDictionary)
						GetDictionary(MMF_OBJECT_NAME);
					MemoryMappedFileStream mmfs = 
						(MemoryMappedFileStream)handleTable[key];

					// Call the CloseHandle method
					if(!object.Equals(mmfs, null))
					{
						mmfs.CloseMapHandle();
						mmfs.Dispose();
					}

					// Remove the key from the dictionary
					handleTable.Remove(key);

					AddDictionary(MMF_OBJECT_NAME);
				}
			}
			catch( Exception genException )
			{
				ExceptionManager.Publish( genException );
				throw;
			}
			finally
			{
				mutexSync.ReleaseMutex();
			}

		}
		#endregion

		#region GetData
		/// <summary>
		///	Gets the element with the specified key.
		/// </summary>
		/// <remarks>
		///	ICacheStorage.GetData(key);
		/// </remarks>
		/// <param name="key">
		///	The unique key to identify the value
		/// </param>
		/// <returns>
		///	The value retrieved
		/// </returns>
		object ICacheStorage.GetData ( string key )
		{
			try
			{
				object returnValue = null;
				
				#region Throwing Argument Exception

				if(Object.Equals(key, null)) 
				{
					throw(new ArgumentNullException("key",
						CacheResources.ResourceManager[
						"RES_ExceptionNullKey"]));
				}
				if(key.Length == 0)
				{
					throw(new ArgumentOutOfRangeException("key",
						CacheResources.ResourceManager[
						"RES_ExceptionEmptyKey"]));
				}

				#endregion

				int totalSize = 0;

				// Check whether the data is available in the cache
				lock(handleTable)
				{
					handleTable = (HybridDictionary)
						GetDictionary(MMF_OBJECT_NAME);
				}
				if(! handleTable.Contains(key))
				{
					return null;
				}
				
				Mutex mutexSync = new Mutex(false, GetHandleName(key));
				try
				{
					mutexSync.WaitOne();
					using( MemoryMappedFileStream mmfs = 
							   MemoryMappedFileStream.OpenExisting(
							   GetKeyName(key),
							   MemoryProtection.PageReadWrite))
					{
						#region Read the size of data
						try
						{
							mmfs.MapViewToProcessMemory( 0,	LENGTH );

							using(BinaryReader memoryReader = 
									  new BinaryReader(mmfs))
							{
								// Read the data length
								totalSize = memoryReader.ReadInt32();
							}
						}
						finally
						{
							mmfs.CloseMapHandle();
							mmfs.Dispose();
						}
						#endregion
					}

					using( MemoryMappedFileStream mmfs = 
							   MemoryMappedFileStream.OpenExisting(
							   GetKeyName(key), 
							   MemoryProtection.PageReadWrite))
					{
						try
						{
							mmfs.MapViewToProcessMemory( 0,	
								totalSize + LENGTH );

							#region Read the Mapped memory contents
						
							using(BinaryReader memoryReader = new BinaryReader(
									  mmfs))
							{

								// Read the data length
								totalSize = memoryReader.ReadInt32();
								if( totalSize == 0) 
								{
									return null;
								}

								MemoryStream memStream = new MemoryStream( 
									totalSize );
					
								// Read the data from the stream
								CopyStreams( mmfs, memStream, totalSize );

								#region Encryption and Validation

								if( isEncrypted || isValidated )
								{
									byte[] binaryObject = memStream.ToArray();

									if( isEncrypted )
									{
										binaryObject = 
											DataProtectionManager.Decrypt(
											binaryObject );
									}
						
									if( isValidated )
									{
										binaryObject = 
											DataProtectionManager.RemoveMAC( 
											binaryObject );
									}
								
									memStream = new MemoryStream(binaryObject);
								}
								#endregion

								#region Deserialization

								memStream.Seek(0, SeekOrigin.Begin);
								returnValue = new BinaryFormatter().
									Deserialize(memStream);
								memStream.Close();

								#endregion						

								return returnValue;
							}
						#endregion

						}
						finally
						{
							mmfs.CloseMapHandle();
							mmfs.Dispose();
						}
					}
				}
				finally
				{
					mutexSync.ReleaseMutex();
				}
			}
			catch( Exception genException )
			{
				ExceptionManager.Publish( genException );
				throw;
			}
		}
		#endregion
		
		#region Update
		/// <summary>
		///	Updates the element with the specified key.
		/// </summary>
		/// <remarks>
		///	ICacheStorage.Update(key, keyData);
		/// </remarks>
		/// <param name="key">
		///	The unique key to identify the value
		/// </param>
		/// <param name="keyData">
		///	The value to be updated
		/// </param>
		void ICacheStorage.Update ( string key, object keyData )
		{
			MemoryStream memStream = new MemoryStream(); 
			byte[] binaryObject = null;

			try
			{
				#region Throwing Argument Exception

				if(Object.Equals(key, null)) 
				{
					throw(new ArgumentNullException("key",
						CacheResources.ResourceManager[
						"RES_ExceptionNullKey"]));
				}
				if(key.Length == 0)
				{
					throw(new ArgumentOutOfRangeException("key",
						CacheResources.ResourceManager[
						"RES_ExceptionEmptyKey"]));
				}
				if( Object.Equals(keyData, null) )
				{
					throw new ArgumentNullException( "keyData",
						CacheResources.ResourceManager[ 
						"RES_ExceptionNullValue" ] );
				}

				#endregion

				// Serialization
				new BinaryFormatter().Serialize( memStream, keyData );
				memStream.Seek( 0, SeekOrigin.Begin );

				#region Encryption and Validation
				if( isEncrypted || isValidated )
				{
					binaryObject = memStream.ToArray();
				
					if( isValidated )
					{
						binaryObject = DataProtectionManager.AppendMAC( 
							binaryObject );
					}
					
					if( isEncrypted )
					{
						binaryObject = DataProtectionManager.Encrypt( 
							binaryObject ); 
					}
				
					memStream = new MemoryStream( binaryObject );
				}
				#endregion
			
				using( MemoryMappedFileStream mmfs = 
							new MemoryMappedFileStream( 
							GetKeyName( key ), 
							(uint)memStream.Length + LENGTH, 
							MemoryProtection.PageReadWrite ) )
				{
					#region Synchronize access to memory
					
					Mutex mutexSync = new Mutex(false, GetHandleName(key));

					try
					{
						mutexSync.WaitOne();

						mmfs.MapViewToProcessMemory( 0, 
							(int)memStream.Length + LENGTH );

						// Write the data length
						using( BinaryWriter memoryWriter = 
									new BinaryWriter( mmfs ) )
						{
							memoryWriter.Write( (int)memStream.Length );

							// Write the data
							CopyStreams( memStream,
								mmfs,
								(int)memStream.Length );
						}
					}
					finally
					{	
						mutexSync.ReleaseMutex();
						mmfs.CloseMapHandle();
						mmfs.Dispose();
					}

					#endregion
				}
			
				memStream.Close();
			}
			catch( Exception genException )
			{
				ExceptionManager.Publish( genException );
				throw;
			}
		}
		#endregion

		#region Add
		
		/// <summary>
		///	Adds an element with the specified key and value into the shared 
		///	memory stream.
		/// </summary>
		/// <remarks>
		///	ICacheStorage.Add(key, keyData);
		/// </remarks>
		/// <param name="key">
		///	The unique key to identify the value
		/// </param>
		/// <param name="keyData">
		///	The value to be added
		/// </param>
		void ICacheStorage.Add ( string key, object keyData )
		{
			try
			{
				#region Throwing Argument Exception

				if(Object.Equals(key, null)) 
				{
					throw(new ArgumentNullException("key",
						CacheResources.ResourceManager[
						"RES_ExceptionNullKey"]));
				}
				if(key.Length == 0)
				{
					throw(new ArgumentOutOfRangeException("key",
						CacheResources.ResourceManager[
						"RES_ExceptionEmptyKey"]));
				}
				if( Object.Equals(keyData, null) )
				{
					throw new ArgumentNullException( "keyData",
						CacheResources.ResourceManager[ 
						"RES_ExceptionNullValue" ] );
				}

				#endregion

				// Serializes the value in order to store it 
				MemoryStream memStream = new MemoryStream();
				byte[] binaryObject = null;
				new BinaryFormatter().Serialize( memStream, keyData );
			
				#region Encryption and Validation
				if( isEncrypted || isValidated )
				{
					binaryObject = memStream.ToArray();
				
					// Appends a mac value for data validation
					if( isValidated )
					{
						binaryObject = DataProtectionManager.AppendMAC(
							binaryObject ); 
					}
					
					// Encrypts the value
					if( isEncrypted )
					{
						binaryObject = DataProtectionManager.Encrypt( 
							binaryObject ); 
					}
				
					memStream = new MemoryStream( binaryObject );
				}
				#endregion
			
				memStream.Seek( 0, SeekOrigin.Begin );

				using( MemoryMappedFileStream mmfs = 
							new MemoryMappedFileStream(
							GetKeyName( key ), 
							(uint)memStream.Length + LENGTH, 
							MemoryProtection.PageReadWrite ) )
				{
					#region Synchronize the access to memory

					Mutex mutexSync = new Mutex(false, GetHandleName(key));
					try
					{
						mutexSync.WaitOne();

						mmfs.MapViewToProcessMemory( 0, 
							(int)memStream.Length + LENGTH );

						// Write the data length
						using( BinaryWriter memoryWriter = 
									new BinaryWriter( mmfs ) )
						{
							memoryWriter.Write((int)memStream.Length );

							// Write the data
							CopyStreams( memStream,
								mmfs, 
								(int)memStream.Length );
						}

						lock(handleTable)
						{
							// Add the memory mapped object to the hash table
							handleTable = (HybridDictionary)
								GetDictionary(MMF_OBJECT_NAME);

							if(! handleTable.Contains(key))
							{
								handleTable.Add(key, mmfs);
							}

							AddDictionary(MMF_OBJECT_NAME);
						}
					}
					finally
					{
						mutexSync.ReleaseMutex();
						mmfs.Dispose();
					}

					#endregion
				}

				memStream.Close();
			}
			catch( Exception genException )
			{
				ExceptionManager.Publish( genException );
				throw;
			}
		}
		#endregion
		
		#region Flush
		/// <summary>
		///	Removes all elements from the storage.
		/// </summary>
		/// <remarks>
		///	ICacheStorage.Flush();
		/// </remarks>
		void ICacheStorage.Flush ()
		{
			try
			{
				RemoveAllKeys();
			}
			catch( Exception genException )
			{
				ExceptionManager.Publish( genException );
				throw;
			}
		}

		#endregion
		
		#endregion

		#region Implementation of IMmfReference

		/// <summary>
		///	Opens the already existing file mapping object. It returns
		///	the MemoryMappedFileStream object which contains a 
		///	handle to the opened file mapping object.
		/// </summary>
		/// <remarks>
		///	IMmfReference.AddReference(key)
		/// </remarks>
		/// <param name="key">
		///	The unique key to identify the value
		/// </param>
		object IMmfReference.AddReference(string key)
		{
			Mutex mutexSync = new Mutex(false, GetHandleName(key));
			try
			{
				mutexSync.WaitOne();
				using( MemoryMappedFileStream mmfs = 
						   MemoryMappedFileStream.OpenExisting(
						   GetKeyName(key),
						   MemoryProtection.PageReadWrite))
				{
					return mmfs;
				}
			}
			finally
			{
				mutexSync.ReleaseMutex();
			}
		}

		/// <summary>
		///	Removes the reference by calling the closehandle function.
		/// </summary>
		/// <remarks>
		///	IMmfReference.RemoveReference(memoryMap);
		/// </remarks>
		/// <param name="memoryMap">
		/// The MemoryMappedFileStream object which 
		/// contains the handle to close
		/// </param>
		void IMmfReference.RemoveReference(object memoryMap)
		{
			MemoryMappedFileStream mmfs = (MemoryMappedFileStream)memoryMap;
			if(mmfs != null)
			{
				mmfs.CloseMapHandle();
			}
		}

		#endregion
		
		#region Private methods

		/// <summary>
		///	Gets the complete key name.
		/// </summary>
		/// <remarks>
		///	string fileName = GetKeyName( key );
		/// </remarks>
		/// <param name="key">
		///	The unique key to identify the value
		/// </param>
		string GetKeyName( string key )
		{
			try
			{
				string fileName;
				const string INVALID_CHAR = "*?:|/\"\\";
				if(key.IndexOfAny(INVALID_CHAR.ToCharArray()) >= 0)
				{
					throw(new Exception(CacheResources.
						ResourceManager["RES_ExceptionInvalidChar"]));
				}

				fileName = 
					CacheResources.ResourceManager["RES_KeyFileName"] + key;

				return fileName;
			}
			catch( Exception genException )
			{
				ExceptionManager.Publish( genException );
				throw;
			}
		}

		/// <summary>
		///	Gets a mutex handle name.
		/// </summary>
		/// <remarks>
		///	string filepath = GetHandleName( key );
		/// </remarks>
		/// <param name="key">
		///	The unique key to identify the value
		/// </param>
		string GetHandleName( string key )
		{
			try
			{
				return CacheResources.ResourceManager["RES_MutexName"] + key;
			}
			catch( Exception genException )
			{
				ExceptionManager.Publish( genException );
				throw;
			}
		}

		/// <summary>
		///	Copies a number of elements from the first stream to the second 
		///	stream.
		/// </summary>
		/// <remarks>
		///	CopyStreams( from, to, count );
		/// </remarks>
		/// <param name="from">
		///	Indicates the start point of file
		/// </param>
		/// <param name="to">
		///	Indicates the end point of file
		/// </param>
		/// <param name="count">
		///	Length of the byte stream to be copied
		/// </param>
		void CopyStreams ( Stream from, Stream to, int count )
		{
			int readed = 0;
			int totalReaded = 0;
			byte[] tempBuffer = new byte[ Convert.ToInt32(
				CacheResources.ResourceManager[ "RES_BufferSize" ]) ];

			try
			{
				while( totalReaded < from.Length && totalReaded < count )
				{
					readed = from.Read( 
						tempBuffer, 
						0, 
						tempBuffer.Length );

					to.Write( 
						tempBuffer, 
						0, 
						(readed > count) ? count : readed );

					totalReaded += readed;
				}
			}
			catch( Exception genException )
			{
				ExceptionManager.Publish( genException);
				throw;
			}
		}

		/// <summary>
		///	This method removes all the cache files in the disk.
		/// </summary>
		/// <remarks>
		///	RemoveAllKeys();
		/// </remarks>
		private void RemoveAllKeys()
		{
			// Synchronize access
			Mutex mutexSync = new Mutex();
			try
			{
				mutexSync.WaitOne();
				lock(handleTable)
				{
					handleTable = (HybridDictionary) 
						GetDictionary(MMF_OBJECT_NAME);

					foreach(object key in handleTable.Keys)
					{
						MemoryMappedFileStream mmfs = 
							(MemoryMappedFileStream)handleTable[key];
						if(!object.Equals(mmfs, null))
						{
							mmfs.CloseMapHandle();
							mmfs.Close();
						}
					}

					handleTable.Clear();
					AddDictionary(MMF_OBJECT_NAME);
				}
				
			}
			catch( Exception genException )
			{
				ExceptionManager.Publish( genException );
				throw;
			}
			finally
			{
				mutexSync.ReleaseMutex();
			}
		}

		/// <summary>
		///	Adds the Hybrid dictionary object to the shared memory.
		/// </summary>
		/// <remarks>
		///	AddDictionary(key);
		/// </remarks>
		/// <param name="key">
		///	The unique key to identify the dictionary
		/// </param>
		private void AddDictionary (string key)
		{
			// Close the dictionary Handle
			if(!object.Equals(dictionaryMmfs, null))
			{
				dictionaryMmfs.CloseMapHandle();
				dictionaryMmfs.Dispose();
			} 

			// Serializes the value in order to store it 
			MemoryStream memStream = new MemoryStream();
			new BinaryFormatter().Serialize( memStream, handleTable );
	
			memStream.Seek( 0, SeekOrigin.Begin );

			using( MemoryMappedFileStream dictMmfs = 
					   new MemoryMappedFileStream(
					   key, 
					   hybridDictionarySize, 
					   MemoryProtection.PageReadWrite ) )
			{
				dictionaryMmfs = dictMmfs;

				#region Synchronize the access to memory
				
				Mutex mutexSync = new Mutex(false, GetHandleName(key));
				try
				{
					mutexSync.WaitOne();

					dictMmfs.MapViewToProcessMemory( 0, 
						(int)memStream.Length + LENGTH );

					// Write the data length
					using( BinaryWriter memoryWriter = 
							   new BinaryWriter( dictMmfs ) )
					{
						memoryWriter.Write((int)memStream.Length );

						// Write the data
						CopyStreams( memStream,
							dictMmfs, 
							(int)memStream.Length );
					}
				}
				finally
				{
					mutexSync.ReleaseMutex();
					dictMmfs.Dispose();
				}

				#endregion
			}

			memStream.Close();
		}

		/// <summary>
		///	Gets the Hybrid dictionary object from the shared memory.
		/// </summary>
		/// <remarks>
		///	HybridDictionary dictionary = GetDictionary(key);
		/// </remarks>
		/// <param name="key">
		///	The unique key to identify dictionary
		/// </param>
		private object GetDictionary(string key)
		{

			object returnValue = null;
			int totalSize;
			Mutex mutexSync = new Mutex(false, GetHandleName(key));
			try
			{
				mutexSync.WaitOne();
				using( MemoryMappedFileStream dictMmfs = 
						   MemoryMappedFileStream.OpenExisting(
						   key,
						   MemoryProtection.PageReadWrite))
				{
					#region Read the size of dictionary

					try
					{
						dictMmfs.MapViewToProcessMemory( 0,	LENGTH );

						using(BinaryReader memoryReader = 
								  new BinaryReader(dictMmfs))
						{
							// Read the data length
							totalSize = memoryReader.ReadInt32();
						}
					}
					finally
					{
						dictMmfs.CloseMapHandle();
						dictMmfs.Dispose();
					}
					#endregion
				}

				using( MemoryMappedFileStream dictMmfs = 
						   MemoryMappedFileStream.OpenExisting(
						   key,
						   MemoryProtection.PageReadWrite))
				{
					try
					{
						dictMmfs.MapViewToProcessMemory( 0,
							totalSize + LENGTH );

						#region Read the Mapped memory contents
						
						using(BinaryReader memoryReader = new BinaryReader(
								  dictMmfs))
						{

							// Read the data length
							totalSize = memoryReader.ReadInt32();
							if( totalSize == 0) 
							{
								return null;
							}

							MemoryStream memStream = new MemoryStream( 
								totalSize );
					
							// Read the data from the stream
							CopyStreams( dictMmfs, memStream, totalSize );

							#region Deserialization

							memStream.Seek(0, SeekOrigin.Begin);
							returnValue = new BinaryFormatter().
								Deserialize(memStream);
							memStream.Close();

							#endregion						

							return returnValue;
						}
						#endregion
					}
					finally
					{
						dictMmfs.CloseMapHandle();
						dictMmfs.Dispose();
					}
				}			
			}
			finally
			{
				mutexSync.ReleaseMutex();
			}
		}

		#endregion
	}
}