//===============================================================================
// Microsoft Caching Application Block for .NET
// http://msdn.microsoft.com/library/default.asp?url=/library/en-us/dnpag/html/CachingBlock.asp
//
// SingletonCacheStorage.cs
// This class is used to cache data into memory.
//
//===============================================================================
// Copyright (C) 2003 Microsoft Corporation
// All rights reserved.
// THIS CODE AND INFORMATION IS PROVIDED "AS IS" WITHOUT WARRANTY
// OF ANY KIND, EITHER EXPRESSED OR IMPLIED, INCLUDING BUT NOT
// LIMITED TO THE IMPLIED WARRANTIES OF MERCHANTABILITY AND/OR
// FITNESS FOR A PARTICULAR PURPOSE.
//===============================================================================


using System;
using System.Collections;
using System.Configuration;
using System.Globalization;
using System.IO;
using System.Xml;
using System.Collections.Specialized;
using System.Runtime.Serialization.Formatters.Binary;

// Used for Data Protection of the data being cached
using Microsoft.ApplicationBlocks.Cache.DataProtection;
// Used for Handling Exceptions
using Microsoft.ApplicationBlocks.ExceptionManagement;

namespace Microsoft.ApplicationBlocks.Cache.Storages
{
	/// <summary>
	///	Sample storage provider used to cache data into memory.
	/// </summary>
	/// <remarks>
	///	This provider uses the following attributes on the XML file:
	/// <list type="">
	///	<item>
	///	<b>Assembly</b>. 
	///	Used to specify the provider assembly (Required)
	///	</item>
	///	<item><b>ClassName</b>. 
	///	Used to specify the provider class name (Required)
	///	</item>
	///	<item><b>Encrypted</b>. 
	///	Specifies if the cache item must be encrypted before 
	///	it is stored (Optional)
	///	</item>
	///	<item><b>Validated</b>. 
	///	Specifies if a cache item validation must be done 
	///	to ensure data integrity (Optional)
	///	</item>
	/// </list>
	/// </remarks>
	public class SingletonCacheStorage : ICacheStorage
	{
		#region Private Members

		private HybridDictionary cacheStorage = new HybridDictionary();
		private bool isValidated = false;
		private bool isEncrypted = false;

		#endregion
		
		#region Constructor
		/// <summary>
		///	Default Constructor.
		/// </summary>
		public SingletonCacheStorage()
		{
		}
		#endregion

		#region Properties
		/// <summary>
		///	Gets the number of elements actually contained in storage.
		/// </summary>
		int ICacheStorage.Size
		{
			get{return cacheStorage.Values.Count;}
		}
		#endregion

		#region Implementation of ICacheStorage

		#region Init
		/// <summary>
		///	Inits the storage provider.
		/// </summary>
		/// <remarks>
		///	ICacheStorage.Init( configSection );
		/// </remarks>
		/// <param name="configSection">
		///	Represents a single node in the XML document
		/// </param>
		void ICacheStorage.Init ( XmlNode configSection ) 
		{
			// Variable Declarations
			string validatedString;
			string encryptedString;

			const string CONST_TRUE = "true";
			const string CONST_FALSE = "false";

			try
			{
				#region Throwing Argument Exception

				if(Object.Equals(configSection, null))
				{
					throw(new ArgumentNullException("configSection",
						CacheResources.ResourceManager[
						"RES_ExceptionNullConfigSection"]));
				}

				#endregion

				#region Gets the Signed attribute
			
				if( ! Object.Equals( configSection.Attributes[
					CacheResources.ResourceManager[ 
					"RES_ConfigValidated" ] ], null ) ) 
				{
					validatedString = configSection.Attributes[CacheResources.
						ResourceManager["RES_ConfigValidated"]].Value ;
				}
				else
				{
					throw(new ConfigurationException(CacheResources.
						ResourceManager["RES_ExceptionEmptyNodeValidated"]));
				}

				if(Object.Equals(validatedString, null))
				{
					throw(new ConfigurationException(CacheResources.
						ResourceManager["RES_ExceptionNullValidated"]));
				}
				else if (validatedString.Length == 0)
				{
					throw(new ConfigurationException(CacheResources.
						ResourceManager["RES_ExceptionEmptyValidated"]));
				}
				else if(!(Object.Equals(validatedString.ToLower(), CONST_TRUE)
					|| Object.Equals(validatedString.ToLower(), CONST_FALSE)))
				{
					throw(new ConfigurationException(CacheResources.
						ResourceManager["RES_ExceptionInvalidValidated"]));
				}
				else
				{
					isValidated = bool.Parse(validatedString.ToLower(
						CultureInfo.CurrentCulture));
				}				
				
				#endregion
			
				#region Gets the Encrypted attribute
				
				if( ! Object.Equals(configSection.Attributes[
					CacheResources.ResourceManager[
					"RES_ConfigEncrypted" ] ], null) ) 
				{
					encryptedString = configSection.Attributes[CacheResources.
						ResourceManager["RES_ConfigEncrypted"]].Value;
				}
				else
				{
					throw(new ConfigurationException(CacheResources.
						ResourceManager["RES_ExceptionEmptyNodeEncrypted"]));
				}

				if(Object.Equals(encryptedString, null))
				{
					throw(new ConfigurationException(CacheResources.
						ResourceManager["RES_ExceptionNullEncrypted"]));
				}
				else if(encryptedString.Length == 0)
				{
					throw(new ConfigurationException(CacheResources.
						ResourceManager["RES_ExceptionEmptyEncrypted"]));
				}
				else if(!(Object.Equals(encryptedString.ToLower(),CONST_TRUE)
					|| Object.Equals(encryptedString.ToLower(), CONST_FALSE)))
				{
					throw(new ConfigurationException(CacheResources.
						ResourceManager["RES_ExceptionInvalidEncrypted"]));
				}
				else
				{
					isEncrypted = bool.Parse(encryptedString.ToLower(
						CultureInfo.CurrentCulture));
				}
				
				#endregion
			}
			catch( Exception genException )
			{
				ExceptionManager.Publish( genException );
				throw;
			}				
		}

		#endregion

		#region Add
		
		/// <summary>
		///	Adds an element with the specified key and value into memory.
		/// </summary>
		/// <remarks>
		///	ICacheStorage.Add( key, keyData );
		/// </remarks>
		/// <param name="key">
		///	The unique key to identify a value
		/// </param>
		/// <param name="keyData">
		///	The value to be stored
		/// </param>
		void ICacheStorage.Add ( string key, object keyData )
		{
			try
			{
				#region Throwing Argument Exception

				if(Object.Equals(key, null)) 
				{
					throw(new ArgumentNullException("key",
						CacheResources.ResourceManager[
						"RES_ExceptionNullKey"]));
				}
				if(key.Length == 0)
				{
					throw(new ArgumentOutOfRangeException("key",
						CacheResources.ResourceManager[
						"RES_ExceptionEmptyKey"]));
				}
				if( Object.Equals(keyData, null) )
				{
					throw new ArgumentNullException( "keyData",
						CacheResources.ResourceManager[ 
						"RES_ExceptionNullValue" ] );
				}

				#endregion

				((ICacheStorage)this).Update( key, keyData );
			}
			catch( Exception genException )
			{
				ExceptionManager.Publish( genException );
				throw;
			}				
		}

		#endregion
		
		#region Flush
		
		/// <summary>
		///	Removes all elements from the storage.
		/// </summary>
		/// <remarks>
		///	ICacheStorage.Flush();
		/// </remarks>
		void ICacheStorage.Flush ()
		{
			try
			{
				// The HybridDictionary must be synchronized on write access
				lock( cacheStorage )
				{
					cacheStorage.Clear();
				}
			}
			catch( Exception genException )
			{
				ExceptionManager.Publish( genException );
				throw;
			}		
		}

		#endregion
		
		#region GetData
		
		/// <summary>
		///	Gets the element with the specified key.
		/// </summary>
		/// <remarks>
		///	object returnValue = ICacheStorage.GetData ( key );
		/// </remarks>
		/// <param name="key">
		///	The unique key to identify a value
		/// </param>
		object ICacheStorage.GetData ( string key )
		{
			try
			{
				object returnValue;

				#region Throwing Argument Exception

				if(Object.Equals(key, null)) 
				{
					throw(new ArgumentNullException("key",
						CacheResources.ResourceManager[
						"RES_ExceptionNullKey"]));
				}
				if(key.Length == 0)
				{
					throw(new ArgumentOutOfRangeException("key",
						CacheResources.ResourceManager[
						"RES_ExceptionEmptyKey"]));
				}

				#endregion

				returnValue = cacheStorage[key];
				if ( (! Object.Equals(returnValue, null)) && 
					( isEncrypted || isValidated ) )
				{
					// The value is serialized and must be restored
					byte[] binaryObject = (byte[])returnValue;
				
					#region Decrypts the value
			
					if( isEncrypted )
					{
						binaryObject = 
							DataProtectionManager.Decrypt( binaryObject );
					}
					
					#endregion
					 
					#region Validates the value
					
					if( isValidated )
					{
						binaryObject = 
							DataProtectionManager.RemoveMAC( binaryObject );
					}
					
					#endregion
					
					#region Deserialization
					
					using ( MemoryStream memStream = 
								new MemoryStream( binaryObject ) )
					{
						returnValue = 
							new BinaryFormatter().Deserialize( memStream );
						memStream.Close();
					}
					
					#endregion
				}
									
				return returnValue;
			}
			catch( Exception genException )
			{
				ExceptionManager.Publish( genException );
				throw;
			}
		}

		#endregion

		#region Remove
		
		/// <summary>
		///	Removes the element with the specified key.
		/// </summary>
		/// <remarks>
		///	ICacheStorage.Remove( key);
		/// </remarks>
		/// <param name="key">
		///	The unique key to identify a value
		/// </param>
		void ICacheStorage.Remove( string key)
		{
			try
			{
				// The HybridDictionary must be synchronized on write access
				lock( cacheStorage )
				{
					cacheStorage.Remove(key);
				}
			}
			catch( Exception genException )
			{
				ExceptionManager.Publish( genException );
				throw;
			}	
		}

		#endregion
		
		#region Update
		
		/// <summary>
		///	Updates the element with the specified key.
		/// </summary>
		/// <remarks>
		///	ICacheStorage.Update( key, keyData );
		/// </remarks>
		/// <param name="key">
		///	The unique key to identify a value
		/// </param>
		/// <param name="keyData">
		///	The value to be updated
		/// </param>
		void ICacheStorage.Update( string key, object keyData)
		{
			try
			{
				#region Throwing Argument Exception

				if(Object.Equals(key, null)) 
				{
					throw(new ArgumentNullException("key",
						CacheResources.ResourceManager[
						"RES_ExceptionNullKey"]));
				}
				if(key.Length == 0)
				{
					throw(new ArgumentOutOfRangeException("key",
						CacheResources.ResourceManager[
						"RES_ExceptionEmptyKey"]));
				}
				if( Object.Equals(keyData, null) )
				{
					throw new ArgumentNullException( "keyData",
						CacheResources.ResourceManager[ 
						"RES_ExceptionNullValue" ] );
				}

				#endregion

				if ( isEncrypted || isValidated )
				{
					#region Serialization
					// If the value is encrypted or signed then 
					// must be serialized
					byte[] binaryObject;
					using ( MemoryStream memStream = new MemoryStream() )
					{
						new BinaryFormatter().Serialize( memStream, keyData );
						binaryObject = memStream.ToArray(); 
						memStream.Close();
					}
			
					#endregion
						
					#region Data validation
					
					// Appends a mac value for data validation
					if( isValidated )
					{
						binaryObject = 
							DataProtectionManager.AppendMAC( binaryObject );
					}
					
					#endregion
				
					#region Encrypts the value	
					
					if( isEncrypted )
					{
						binaryObject = 
							DataProtectionManager.Encrypt( binaryObject );
					}
					
					#endregion
						
					// The HybridDictionary must be synchronized on
					// write access
					lock( cacheStorage )
					{
						cacheStorage[ key ] = binaryObject;
					}
				}
				else
				{
					// The HybridDictionary must be synchronized
					// on write access
					lock( cacheStorage )
					{
						cacheStorage[ key ] = keyData;
					}
				}
			}
			catch( Exception genException )
			{
				ExceptionManager.Publish( genException );
				throw;
			}	
		}	

		#endregion

		#endregion

	}
}