//===============================================================================
// Microsoft Caching Application Block for .NET
// http://msdn.microsoft.com/library/default.asp?url=/library/en-us/dnpag/html/CachingBlock.asp
//
// SqlServerCacheStorage.cs
// This class is used to cache data into a Sqldatabase
//
//===============================================================================
// Copyright (C) 2003 Microsoft Corporation
// All rights reserved.
// THIS CODE AND INFORMATION IS PROVIDED "AS IS" WITHOUT WARRANTY
// OF ANY KIND, EITHER EXPRESSED OR IMPLIED, INCLUDING BUT NOT
// LIMITED TO THE IMPLIED WARRANTIES OF MERCHANTABILITY AND/OR
// FITNESS FOR A PARTICULAR PURPOSE.
//===============================================================================


using System;
using System.Collections;
using System.Configuration;
using System.Data;
using System.Globalization;
using System.IO;
using System.Text;
using System.Xml;
using System.Data.SqlClient;
using System.Runtime.Serialization.Formatters.Binary;

// Used for Caching Management
using Microsoft.ApplicationBlocks.Cache;
// Used for Data Access Management
using Microsoft.ApplicationBlocks.Data;
// Used for Handling Exceptions
using Microsoft.ApplicationBlocks.ExceptionManagement;
// Used for Data Protection of the data being cached
using Microsoft.ApplicationBlocks.Cache.DataProtection;

namespace Microsoft.ApplicationBlocks.Cache.Storages
{
	/// <summary>
	///	Sample SqlServer storage provider used to cache data into a database.
	/// </summary>
	/// <remarks>
	///	This provider uses the following attributes on the XML file:
	/// <list type="">
	///	<item>
	///	<b>Assembly</b> 
	///	Used to specify the provider assembly (Required)
	///	</item>
	///	<item>
	///	<b>ClassName</b> 
	///	Used to specify the provider class name (Required)
	///	</item>
	///	<item>
	///	<b>ApplicationName</b> 
	///	Used to specify the application name that will use this provider 
	///	(Optional)
	///	</item>
	///	<item>
	///	<b>ConnectionString</b> 
	///	Used to specify the connection string for a SqlServer database 
	///	(Required)
	///	</item>
	///	<item>
	///	<b>Encrypted</b> 
	///	Specifies if the cache item must be encrypted before it is stored 
	///	(Optional)
	///	</item>
	///	<item>
	///	<b>Validated</b> 
	///	Specifies if a cache item validation must be done to ensure data 
	///	integrity (Optional)
	///	</item>
	/// </list>
	/// </remarks>
	[Serializable]
	public class SqlServerCacheStorage : ICacheStorage, ICacheMetadataWithCallback
	{
		#region Private constants

		private static readonly string APPLICATION_PARAMETER = 
			"@piApplication";
		private static readonly string KEY_PARAMETER = "@piKey";
		private static readonly string VALUE_PARAMETER = "@piValue";
		private static readonly string PRIORITY_PARAMETER = "@piPriority";
		private static readonly string EXPIRY_PARAMETER = "@piExpiryValue";
		private static readonly string CALLBACKMETHODNAME_PARAMETER = "@piCallbackMethodName";
		private static readonly string CALLBACKTYPE_PARAMETER = "@piCallbackType";
		#endregion

		#region Private members

		string connectionString;			
		string applicationName;
		bool isEncrypted;
		bool isValidated;
		int rowCount;
		byte[] binaryObject;
		static int keySize;
		static int applicationSize;

		#endregion

		#region Public constructor

		/// <summary>
		///	Default constructor.
		/// </summary>
		public SqlServerCacheStorage() {}

		#endregion

		#region Properties

		/// <summary>
		///	Gets the number of elements actually contained in the database.
		/// </summary>
		int ICacheStorage.Size
		{
			get
			{
				try
				{
					object itemCount = SqlHelper.ExecuteScalar( 
						connectionString,
						CommandType.StoredProcedure,
						CacheResources.ResourceManager [ 
						"RES_StoredProcedureItemCount" ], 
						new SqlParameter[] { new SqlParameter( 
											   APPLICATION_PARAMETER,
											   applicationName ) } );
					
					return (int)itemCount;
				}
				catch( Exception genException )
				{
					throw new Exception(
						string.Format( 
						CultureInfo.CurrentCulture,
						CacheResources.ResourceManager[ 
						"RES_ExceptionSqlStorageErrorChechingCount" ] ), 
						genException );
				}
			}
		}

		#endregion

		#region Implementation of ICacheStorage

		#region Init

		/// <summary>
		///	Inits the storage provider.
		/// </summary>
		/// <remarks>
		///	ICacheStorage.Init( configSection );
		/// </remarks>
		/// <param name="configSection">
		///	Xml node to read the attributes from the config file
		/// </param>		
		void ICacheStorage.Init ( XmlNode configSection ) 
		{
			try
			{
				string validatedString;
				string encryptedString;			
				XmlAttribute applName;
				XmlAttribute conString;

				const string CONST_TRUE = "true";
				const string CONST_FALSE = "false";

				#region Get application name

				if (Object.Equals(configSection, null))
				{
					throw new ArgumentNullException("configSection", 
						CacheResources.ResourceManager[ 
						"RES_ExceptionNullConfigSection"]);
				}

				// Read the application name from the App.config file
				applName = configSection.Attributes[
					CacheResources.ResourceManager[ 
					"RES_ApplicationName" ] ];
				
				conString = configSection.Attributes[ 
					CacheResources.ResourceManager[ 
					"RES_ConnectionString" ] ];

				#endregion

				#region Get connection String

				// Read the connection string from the App.config file
				if( Object.Equals(conString, null) )
				{
					throw new ConfigurationException( 
						CacheResources.ResourceManager[ 
						"RES_ExceptionInvalidConfigurationSqlStorageProviderConnectionString" ] );
				}

				connectionString = conString.Value;

				if( connectionString.Length == 0 )
				{
					throw new ConfigurationException( 
						CacheResources.ResourceManager[ 
						"RES_ExceptionInvalidConfigurationSqlStorageProviderConnectionString" ] );
				}

				#endregion

				#region Get validated attribute

				// Read the Validated attribute from the App.config file
				if( ! Object.Equals( configSection.Attributes[
					CacheResources.ResourceManager[ 
					"RES_ConfigValidated" ] ], null ) ) 
				{
					validatedString = configSection.Attributes[CacheResources.
						ResourceManager["RES_ConfigValidated"]].Value ;
				}
				else
				{
					throw(new ConfigurationException(CacheResources.
						ResourceManager["RES_ExceptionEmptyNodeValidated"]));
				}

				if(Object.Equals(validatedString, null))
				{
					throw(new ConfigurationException(CacheResources.
						ResourceManager["RES_ExceptionNullValidated"]));
				}
				else if (validatedString.Length == 0)
				{
					throw(new ConfigurationException(CacheResources.
						ResourceManager["RES_ExceptionEmptyValidated"]));
				}
				else if(!(Object.Equals(validatedString.ToLower(), CONST_TRUE)
					|| Object.Equals(validatedString.ToLower(), CONST_FALSE)))
				{
					throw(new ConfigurationException(CacheResources.
						ResourceManager["RES_ExceptionInvalidValidated"]));
				}
				else
				{
					isValidated = bool.Parse(validatedString.ToLower(
						CultureInfo.CurrentCulture));
				}				 

				#endregion
			
				#region Get encrypted attribute

				// Read the Encrypted attribute from the App.config file
				if( ! Object.Equals(configSection.Attributes[
					CacheResources.ResourceManager[
					"RES_ConfigEncrypted" ] ], null) ) 
				{
					encryptedString = configSection.Attributes[CacheResources.
						ResourceManager["RES_ConfigEncrypted"]].Value;
				}
				else
				{
					throw(new ConfigurationException(CacheResources.
						ResourceManager["RES_ExceptionEmptyNodeEncrypted"]));
				}
				
				if(Object.Equals(encryptedString, null))
				{
					throw(new ConfigurationException(CacheResources.
						ResourceManager["RES_ExceptionNullEncrypted"]));
				}
				else if(encryptedString.Length == 0)
				{
					throw(new ConfigurationException(CacheResources.
						ResourceManager["RES_ExceptionEmptyEncrypted"]));
				}
				else if(!(Object.Equals(encryptedString.ToLower(), CONST_TRUE)
					|| Object.Equals(encryptedString.ToLower(), CONST_FALSE)))
				{
					throw(new ConfigurationException(CacheResources.
						ResourceManager["RES_ExceptionInvalidEncrypted"]));
				}
				else
				{
					isEncrypted = bool.Parse(encryptedString.ToLower(
						CultureInfo.CurrentCulture));
				}

				#endregion
						
				#region Test the database connection

				try
				{
					SqlConnection connection = 
						new SqlConnection( connectionString );

					object keyDbSize = SqlHelper.ExecuteScalar(connection,
						CommandType.StoredProcedure,
						CacheResources.ResourceManager["RES_GetKeySize"]);
					
					keySize = Convert.ToInt32(keyDbSize);

					object appSize = SqlHelper.ExecuteScalar(connection,
						CommandType.StoredProcedure,
						CacheResources.ResourceManager[
						"RES_GetApplicationSize"]);
					
					applicationSize = Convert.ToInt32(appSize);
				}
				catch( Exception genException )
				{
					throw new Exception( String.Format( 
						CultureInfo.CurrentCulture,
						CacheResources.ResourceManager[ 
						"RES_ExceptionCantOpenConnection" ],
						connectionString ),
						genException );
				}

				#endregion

				#region Validate application name
				
				if( ! Object.Equals(applName, null) )
				{
					if(applName.Value.Length == 0 || 
						applName.Value.Length > applicationSize)
					{
						throw(new ConfigurationException(CacheResources.
							ResourceManager[
							"RES_ExceptionRangeApplicationName"]));
					}
					applicationName = applName.Value;
				}
				else
				{
					applicationName = CacheResources.ResourceManager[ 
						"RES_NoApplicationName" ];
				}

				#endregion

			}
			catch( Exception genException )
			{
				ExceptionManager.Publish( genException );
				throw;
			}
				
		}
		#endregion

		#region Add
		/// <summary>
		///	Adds an element with the specified key and value into the database.
		/// </summary>
		/// <remarks>
		///	ICacheStorage.Add( string key, object keyData );
		/// </remarks>
		/// <param name="key">
		///	The unique key to identify the value
		/// </param>
		/// <param name="keyData">
		///	The value to be stored
		/// </param>
		void ICacheStorage.Add ( string key, object keyData )
		{            
			try
			{
				// Throw exception if key size is greater than the 
				// maximum size in the database
				if (key.Length > keySize)
				{
					throw new Exception(CacheResources.ResourceManager[
						"RES_ExceptionKeySize"]);
				}
				SqlParameter valueParameter = ConvertData(keyData);

				#region Database call 

				// Database call to the stored procedure to add or
				// update the cache item
				rowCount = SqlHelper.ExecuteNonQuery( connectionString,
					CommandType.StoredProcedure, 
					CacheResources.ResourceManager[ 
					"RES_StoredProcedureSetItem" ], 
					new SqlParameter[] { new SqlParameter( 
										   APPLICATION_PARAMETER,
										   applicationName ),
										   new SqlParameter( 
										   KEY_PARAMETER, 
										   key ),
										   valueParameter } );
				#endregion
										 
				if( rowCount != 1 )
				{
					throw new Exception( 
						CacheResources.ResourceManager[ 
						"RES_ExceptionErrorUpdatingTheDatabaseStorage" ] );
				}
			}
			catch( Exception genException )
			{
				ExceptionManager.Publish(genException);
				throw;
			}			
		}
		#endregion

		#region Flush
		/// <summary>
		///	Removes all elements from the storage.
		/// </summary>
		/// <remarks>
		///	ICacheStorage.Flush ();
		/// </remarks>
		void ICacheStorage.Flush ()
		{
			try
			{
				#region Database call
			
				// Database call to the stored procedure 
				// to flush the cache items
				SqlHelper.ExecuteNonQuery( connectionString,
					CommandType.StoredProcedure,
					CacheResources.ResourceManager[ 
						"RES_StoredProcedureFlushItem" ], 
					new SqlParameter[] { new SqlParameter( 
										   APPLICATION_PARAMETER,
										   applicationName ) } );
				#endregion
			}
			catch( Exception genException )
			{
				throw new Exception( 
					CacheResources.ResourceManager[ 
					"RES_ExceptionSqlStorageErrorFlushingValue" ],
					genException );
			}
		}
		#endregion

		#region GetData
		/// <summary>
		/// Gets the element with the specified key.
		/// </summary>
		/// <remarks>
		///	object result = ICacheStorage.GetData(string key);
		/// </remarks>
		/// <param name="key">
		///	The unique key to identify the value
		/// </param>
		/// <returns>
		///	The cache item value
		/// </returns>
		object ICacheStorage.GetData ( string key )
		{
			try
			{
				object result = null;
			
				if( Object.Equals(key, null) || (key.Length == 0) )
				{
					throw new ArgumentNullException( "key", 
						CacheResources.ResourceManager[ 
						"RES_ExceptionTheKeyCantBeNullOrEmptyString" ] );
				}

				#region Database call
			
				// Database call to the stored procedure to get
				// the cache item value
				Object reader = SqlHelper.ExecuteScalar ( 
					connectionString,
					CommandType.StoredProcedure,
					CacheResources.ResourceManager[ 
					"RES_StoredProcedureGetItem" ],
					new SqlParameter[] { new SqlParameter( 
										   APPLICATION_PARAMETER, 
										   applicationName ),
										   new SqlParameter( 
										   KEY_PARAMETER, 
										   key ) } );
	
				#endregion

				if(!object.Equals(reader, null))
				{
					#region Restores the value
				
					string base64CacheItem = reader.ToString();
					byte[] cacheItemBinary = Convert.FromBase64String( 
						base64CacheItem );
					
					#endregion
					
					#region Decrypts the value
				
					if( isEncrypted )
					{
						cacheItemBinary = DataProtectionManager.Decrypt( 
							cacheItemBinary );
					}
				
					#endregion
					
					#region Validates the value
				
					if( isValidated )
					{
						cacheItemBinary = DataProtectionManager.RemoveMAC( 
							cacheItemBinary ); 
					}
				
					#endregion
				
					#region Deserialize the data
				
					result = new BinaryFormatter().Deserialize( 
						new MemoryStream( cacheItemBinary ) );
				
					#endregion					
				}
				return result;
			}
			catch( Exception genException )
			{
				throw new Exception( string.Format( 
					CultureInfo.CurrentCulture,
					CacheResources.ResourceManager[ 
					"RES_ExceptionSqlStorageErrorGettingValue" ],
					key ), 
					genException );
			}			
		}
		#endregion

		#region Remove
		/// <summary>
		///	Removes the element with the specified key.
		/// </summary>
		/// <remarks>
		///	ICacheStorage.Remove( key );
		/// </remarks>
		/// <param name="key">
		///	The unique key to identify the value
		/// </param>
		void ICacheStorage.Remove ( string key )
		{
			try
			{				
				// Database call to the stored procedure to
				// remove a cache item from the store
				SqlHelper.ExecuteNonQuery( connectionString,
					CommandType.StoredProcedure, 
					CacheResources.ResourceManager[ 
						"RES_StoredProcedureDeleteItem" ], 
					new SqlParameter[] { new SqlParameter( 
										   APPLICATION_PARAMETER, 
										   applicationName ),
										   new SqlParameter( 
										   KEY_PARAMETER, 
										   key ) } );									
			}
			catch( Exception genException )
			{
				throw new Exception( 
					string.Format( 
					CultureInfo.CurrentCulture,
					CacheResources.ResourceManager[ 
					"RES_ExceptionSqlStorageErrorRemovingValue" ],
					key ),
					genException );
			}			
		}
		#endregion

		#region Update
		/// <summary>
		///	Updates the element with the specified key.
		/// </summary>
		/// <remarks>
		///	ICacheStorage.Update( key, keyData );
		/// </remarks>
		/// <param name="key">
		///	The unique to identify the value
		/// </param>
		/// <param name="keyData">
		///	The value to be updated
		/// </param>
		void ICacheStorage.Update ( string key, object keyData )
		{
			try
			{
				SqlParameter valueParameter = ConvertData(keyData);
				
				#region Database call
				
				// Database call to stored procedure
				// update the data into the store
				rowCount = SqlHelper.ExecuteNonQuery( connectionString,
					CommandType.StoredProcedure, 
					CacheResources.ResourceManager[ 
						"RES_StoredProcedureSetItem" ], 
				new SqlParameter[] { new SqlParameter( 
										   APPLICATION_PARAMETER,
										   applicationName ),
										   new SqlParameter( 
										   KEY_PARAMETER, 
										   key ),
										   valueParameter } );
				if( rowCount != 1 )
				{
					throw new Exception( 
						CacheResources.ResourceManager[ 
						"RES_ExceptionErrorUpdatingTheDatabaseStorage" ] );
				}
				
				#endregion
			}
			catch( Exception genException )
			{
				throw new Exception( 
					string.Format( 
					CultureInfo.CurrentCulture,
					CacheResources.ResourceManager[ 
					"RES_ExceptionSqlStorageErrorUpdatingValue" ], 
					key ), 
					genException );
			}				
		}		
		
		#endregion

		#region ConvertData
		
		/// <summary>
		/// This method is a helper function for adding and updating the 
		/// element with specified key.
		/// </summary>
		/// <remarks>
		/// sqlParameter value = ConvertData(keyData);
		/// </remarks>
		/// <param name="keyData">
		///	The value to be updated
		/// </param>
		private SqlParameter ConvertData(object keyData)
		{
			MemoryStream memStream = null;
			try
			{
				SqlParameter valueParameter;

				if(Object.Equals(keyData, null))
				{
					throw(new ArgumentNullException( "keyData",
						CacheResources.ResourceManager[
						"RES_ExceptionNullValue"]));
				}

				#region Serialize
			
				// Serializes the value in order to store it in the database
				memStream = new MemoryStream();
				new BinaryFormatter().Serialize( memStream, keyData );
								
				binaryObject = memStream.ToArray();
				
				#endregion

				#region Data Validation
				
				// Appends a mac value for data validation
				if( isValidated )
				{
					binaryObject = DataProtectionManager.AppendMAC( 
						binaryObject );
				}
				
				#endregion
				
				#region Encryption
				
				// Encrypts the value
				if( isEncrypted )
					binaryObject = DataProtectionManager.Encrypt( 
						binaryObject );
				
				#endregion
				
				#region Convert to Base64 string
				
				// Stores the value in the database
				valueParameter = new SqlParameter( 
					VALUE_PARAMETER, SqlDbType.Text );
				valueParameter.Value = Convert.ToBase64String( 
					binaryObject );

				return valueParameter;
				#endregion
			}
			catch(Exception genException)
			{
				ExceptionManager.Publish(genException);
				throw;
			}
			finally 
			{
				memStream.Close();
			}
		}

		#endregion

		#endregion

		#region Implementation of ICacheMetaData

		/// <summary>
		///	Adds new data to the cache metadata storage.
		/// </summary>
		/// <remarks>
		///	ICacheStorage.Add(key, exp, priority);
		/// </remarks>
		/// <param name="key">
		///	The unique key to identify the value
		/// </param>
		/// <param name="expirations">
		///	An array of expirations for the cache data
		/// </param>
		/// <param name="priority">
		///	Priority of the data in cache
		/// </param>
		void ICacheMetadata.Add(string key, 
			ICacheItemExpiration[] expirations,
			CacheItemPriority priority )
		{
//			try
//			{
//				#region Database call
//				// Add the priority in the database
//				rowCount = SqlHelper.ExecuteNonQuery( connectionString,
//					CommandType.StoredProcedure, 
//					CacheResources.ResourceManager[ 
//					"RES_SetPriority" ], 
//					new SqlParameter[] { new SqlParameter( 
//										   APPLICATION_PARAMETER,
//										   applicationName ),
//										   new SqlParameter( 
//										   KEY_PARAMETER, 
//										   key ),
//										   new SqlParameter(
//										   PRIORITY_PARAMETER,
//										   (int)priority)} );
//				if( rowCount != 1 )
//				{
//					throw new Exception( 
//						CacheResources.ResourceManager[ 
//						"RES_ExceptionErrorUpdatingTheDatabaseStorage" ] );
//				}
//
//				// Add the expirations array in the database
//				if(!object.Equals(expirations, null))
//				{
//					byte[] byteStream = null;
//					// Serialize the expirations array
//					byteStream = SerializeData(expirations);
//
//					rowCount = SqlHelper.ExecuteNonQuery( connectionString,
//						CommandType.StoredProcedure, 
//						CacheResources.ResourceManager[ 
//						"RES_SetExpiration" ], 
//						new SqlParameter[] { new SqlParameter( 
//											   APPLICATION_PARAMETER,
//											   applicationName ),
//											   new SqlParameter( 
//											   KEY_PARAMETER, 
//											   key ),
//											   new SqlParameter(
//											   EXPIRY_PARAMETER,
//											   byteStream)} );
//					if( rowCount != 1 )
//					{
//						throw new Exception( 
//							CacheResources.ResourceManager[ 
//							"RES_ExceptionErrorUpdatingTheDatabaseStorage" ] );
//					}
//				}
//
//				#endregion
//			}
//			catch( Exception genException )
//			{
//				throw new Exception( 
//					string.Format( 
//					CultureInfo.CurrentCulture,
//					CacheResources.ResourceManager[ 
//					"RES_ExceptionSqlStorageErrorUpdatingValue" ], 
//					key ), 
//					genException );
//			}
			((ICacheMetadataWithCallback)this).Add(key, expirations, priority, null);
		}
		
		void ICacheMetadataWithCallback.Add(string key, 
			ICacheItemExpiration[]	expirations,
			CacheItemPriority priority,
			CacheItemRemovedCallback itemRemovedCallBack) {
			try {
				#region Database call
				// Add the priority in the database
				rowCount = SqlHelper.ExecuteNonQuery( connectionString,
					CommandType.StoredProcedure, 
					CacheResources.ResourceManager[ 
					"RES_SetPriority" ], 
					new SqlParameter[] { new SqlParameter( 
										   APPLICATION_PARAMETER,
										   applicationName ),
										   new SqlParameter( 
										   KEY_PARAMETER, 
										   key ),
										   new SqlParameter(
										   PRIORITY_PARAMETER,
										   (int)priority)} );
				if( rowCount != 1 ) {
					throw new Exception( 
						CacheResources.ResourceManager[ 
						"RES_ExceptionErrorUpdatingTheDatabaseStorage" ] );
				}

				// Add the expirations array in the database
				if(!object.Equals(expirations, null)) {
					byte[] byteStream = null;
					// Serialize the expirations array
					byteStream = SerializeData(expirations);

					rowCount = SqlHelper.ExecuteNonQuery( connectionString,
						CommandType.StoredProcedure, 
						CacheResources.ResourceManager[ 
						"RES_SetExpiration" ], 
						new SqlParameter[] { new SqlParameter( 
											   APPLICATION_PARAMETER,
											   applicationName ),
											   new SqlParameter( 
											   KEY_PARAMETER, 
											   key ),
											   new SqlParameter(
											   EXPIRY_PARAMETER,
											   byteStream)} );
					if( rowCount != 1 ) {
						throw new Exception( 
							CacheResources.ResourceManager[ 
							"RES_ExceptionErrorUpdatingTheDatabaseStorage" ] );
					}
				}

				// Add the type and method name to the database
				if(!object.Equals(itemRemovedCallBack, null) && (itemRemovedCallBack.Method.IsStatic)) {
					byte[] byteStream = null;
					// Serialize the expirations array
					byteStream = SerializeData(itemRemovedCallBack.Method.DeclaringType);

					rowCount = SqlHelper.ExecuteNonQuery( connectionString,
						CommandType.StoredProcedure, 
						CacheResources.ResourceManager[ 
						"RES_SetCallback" ], 
						new SqlParameter[] { new SqlParameter( 
											   APPLICATION_PARAMETER,
											   applicationName ),
											   new SqlParameter( 
											   KEY_PARAMETER, 
											   key ),
											   new SqlParameter(
											   CALLBACKMETHODNAME_PARAMETER,
											   itemRemovedCallBack.Method.Name),
											   new SqlParameter(
											   CALLBACKTYPE_PARAMETER,
											   byteStream)} );
					if( rowCount != 1 ) {
						throw new Exception( 
							CacheResources.ResourceManager[ 
							"RES_ExceptionErrorUpdatingTheDatabaseStorage" ] );
					}
				}

				#endregion
			}
			catch( Exception genException ) {
				throw new Exception( 
					string.Format( 
					CultureInfo.CurrentCulture,
					CacheResources.ResourceManager[ 
					"RES_ExceptionSqlStorageErrorUpdatingValue" ], 
					key ), 
					genException );
			}
		}
		


		/// <summary>
		///	Gets all metadata from the metadata storage.
		/// </summary>
		/// <remarks>
		///	Hashtable table = ICacheMetadata.GetMetadata();
		/// </remarks>
		/// <returns>
		/// The Hashtable containing all the meta data
		/// </returns>
		Hashtable ICacheMetadata.GetMetadata()
		{	
			try
			{
				DataSet metadataSet = null;
				DataTable parseTable = null;
				DictionaryEntry dataEntry;
				Hashtable metadataTable = new Hashtable();

				metadataSet = SqlHelper.ExecuteDataset( connectionString,
					CommandType.StoredProcedure, 
					CacheResources.ResourceManager[ 
					"RES_GetMetadata" ], 
					new SqlParameter[] { new SqlParameter( 
										   APPLICATION_PARAMETER,
										   applicationName)});

				// Parse through the dataset 
				parseTable = metadataSet.Tables[0];
				foreach(DataRow row in parseTable.Rows)
				{
					// Add the cache item priority
					if(! object.Equals(row[1], System.DBNull.Value))
					{
						dataEntry = new DictionaryEntry(
							row[0],(CacheItemPriority)row[1]);
						metadataTable.Add(metadataTable.Count.ToString(),
							dataEntry);
					}

					// Add the expiration data
					if(! object.Equals(row[2], System.DBNull.Value))
					{
						dataEntry = new DictionaryEntry(
							row[0], 
							(ICacheItemExpiration[])
							DeSerializeData((byte[])row[2]));
						metadataTable.Add(metadataTable.Count.ToString(),
							dataEntry);
					}


					// Add the callback method by building
					if(!object.Equals(row[3], System.DBNull.Value) && !object.Equals(row[4], System.DBNull.Value)) {						
						string methodName = (string)row[3];
						Type type = (Type)DeSerializeData((byte[])row[4]);
						if (type != null && methodName != null && methodName.Length > 0) {
							Delegate d = Delegate.CreateDelegate(typeof(CacheItemRemovedCallback), type, methodName);
							CacheItemRemovedCallback callback = new CacheItemRemovedCallback((CacheItemRemovedCallback)d);						
							dataEntry = new DictionaryEntry(
								row[0], callback);
							metadataTable.Add(metadataTable.Count.ToString(),
								dataEntry);
						}						
					}
				}

				return metadataTable;
			}
			catch(Exception genException)
			{
				ExceptionManager.Publish(genException);
			}
			return null;
		}

		/// <summary>
		///	Removes all metadata from the metadata storage.
		/// </summary>
		void ICacheMetadata.Flush()
		{
			// No implementation is required because
			// when the cache data is deleted, the metadata
			// are also deleted with it
		}

		/// <summary>
		///	Removes the element with the specified key from the metadata 
		///	storage.
		/// </summary>
		/// <param name="key">
		/// The unique key which identifies the data
		/// </param>
		void ICacheMetadata.Remove(string key)
		{
			// No implementation is required because
			// when the cache data is deleted, the metadata
			// are also deleted with it
		}

		/// <summary>
		///	This method is used to serialize the object.
		/// </summary>
		/// <remarks>
		///	byte[] byteStream = SerializeData(data);
		/// </remarks>
		/// <param name="dataValue">
		/// The object to be serialized
		/// </param>
		/// <returns>
		/// A byte array of the serialized object
		/// </returns>
		private byte[] SerializeData(object dataValue)
		{
			MemoryStream memStream = null;
			try
			{
				memStream = new MemoryStream();
				new BinaryFormatter().Serialize(memStream, dataValue);
				return memStream.ToArray();
			}
			catch(Exception genException)
			{
				ExceptionManager.Publish(genException);
				throw;
			}
			finally
			{
				memStream.Close();
			}
		}

		/// <summary>
		///	This method is used to deserialize the byte array.
		/// </summary>
		/// <remarks>
		///	object data = DeSerializeData(binaryData);
		/// </remarks>
		/// <param name="binaryData">
		/// The byte array to be deserialized
		/// </param>
		/// <returns>
		/// The deserialized object
		/// </returns>
		private object DeSerializeData(byte[] binaryData)
		{
			MemoryStream memStream = null;
			try
			{
				memStream = new MemoryStream(binaryData);
				return new BinaryFormatter().Deserialize(memStream);
			}
			catch(Exception genException)
			{
				ExceptionManager.Publish(genException);
				throw;
			}
			finally
			{
				memStream.Close();
			}
		}

		#endregion
	}
}