//===============================================================================
// Copyright (C) 2003 Microsoft Corporation
// All rights reserved.
// THIS CODE AND INFORMATION IS PROVIDED "AS IS" WITHOUT WARRANTY
// OF ANY KIND, EITHER EXPRESSED OR IMPLIED, INCLUDING BUT NOT
// LIMITED TO THE IMPLIED WARRANTIES OF MERCHANTABILITY AND/OR
// FITNESS FOR A PARTICULAR PURPOSE.
//===============================================================================


using System;
using System.Configuration;
using System.Diagnostics;
using System.Runtime.InteropServices; 
using System.Security.Cryptography;
using System.Xml;
using System.Xml.XPath;

namespace Microsoft.ApplicationBlocks.Common.Crypto {
	
    /// <include file='doc\DPAPIProvider.uex' path='docs/doc[@for="DPAPIProvider"]/*'/>
    /// <devdoc>
    /// A data protection provider based on DPAPI.
    /// </devdoc>
    public sealed class DPAPIProvider : ICryptographicProvider {
        #region Attributes
        // the store to use
        private DPAPIStore          dataProtectionStore;
        // the validation algorithm
        private KeyedHashAlgorithm  validationAlgorithm;		
        #endregion

        #region Constants
        private const int CRYPTPROTECT_UI_FORBIDDEN = 0x1;
        private const int CRYPTPROTECT_LOCAL_MACHINE = 0x4;		
        #endregion

        #region P/Invoke structures
        [StructLayout(LayoutKind.Sequential, CharSet = CharSet.Unicode)]
            internal struct DATA_BLOB {
            public int cipherBlobData;
            public IntPtr plainBlobData;
        }

        [StructLayout(LayoutKind.Sequential, CharSet = CharSet.Unicode)]
            internal struct CRYPTPROTECT_PROMPTSTRUCT {
            public int cipherBlobSize;
            public int dwPromptFlags;
            public IntPtr hwndApp;
            public String szPrompt;
        }

        #endregion

        #region P/Invoke methods
        [DllImport("Crypt32.dll", SetLastError=true, CharSet = CharSet.Auto)]
        private static extern bool CryptProtectData(
            ref DATA_BLOB pDataIn, 
            String szDataDescr, 
            ref DATA_BLOB pOptionalEntropy,
            IntPtr pvReserved, 
            ref CRYPTPROTECT_PROMPTSTRUCT 
            pPromptStruct, 
            int dwFlag, 
            ref DATA_BLOB pDataOut);

        [DllImport("Crypt32.dll", SetLastError=true, CharSet = CharSet.Auto)]
        private static extern bool CryptUnprotectData(
            ref DATA_BLOB pDataIn, 
            String szDataDescr, 
            ref DATA_BLOB pOptionalEntropy, 
            IntPtr pvReserved, 
            ref CRYPTPROTECT_PROMPTSTRUCT 
            pPromptStruct, 
            int dwFlag, 
            ref DATA_BLOB pDataOut);
		
        [DllImport("kernel32.dll", CharSet = CharSet.Auto)]
        private unsafe static extern int FormatMessage(int dwFlag, 
            ref IntPtr lpSource, 
            int dwMessageId,
            int dwLanguageId, 
            ref String lpBuffer, 
            int nSize,
            IntPtr *Arguments);
        #endregion
		
        #region Constructors
        /// <include file='doc\DPAPIProvider.uex' path='docs/doc[@for="DPAPIProvider.DPAPIProvider"]/*'/>
        public DPAPIProvider() : this(DPAPIStore.Machine) {
        }

        /// <include file='doc\DPAPIProvider.uex' path='docs/doc[@for="DPAPIProvider.DPAPIProvider1"]/*'/>
        public DPAPIProvider(DPAPIStore store) {
            dataProtectionStore = store;
        }		
        #endregion
		
        #region IProvider Members
        /// <include file='doc\DPAPIProvider.uex' path='docs/doc[@for="DPAPIProvider.Initialize"]/*'/>
        /// <devdoc>
        ///	Initializes the data protection provider.
        /// </devdoc>                
        public void Initialize(XPathNavigator parameters) {
            if(Object.Equals(parameters, null)) {
                throw new ArgumentNullException("parameters");
            }

            // Variable Declarations
            String validationKey = null ;
            String validation = null;

            validationKey = parameters.GetAttribute(ResourceTable.GetString(ResourceTable.ValidationKey), String.Empty);
            Debug.Assert(validationKey != null);
            
            if((Object.Equals(validationKey, null)) || (validationKey.Trim().Length == 0)) {
                // Throw exception if the validation key length is zero				
                throw new ConfigurationException(ResourceTable.GetString(ResourceTable.InvalidBaseString));
            }

            validation = parameters.GetAttribute(ResourceTable.GetString(ResourceTable.Validation), String.Empty);
            Debug.Assert(validation != null);

            validationAlgorithm = CreateAlgorithmFromName(validation);
            if(Object.Equals(validationAlgorithm, null) ) {
                throw new ConfigurationException(ResourceTable.GetString(ResourceTable.InvalidAlg));
            }			
            validationAlgorithm.Key = Convert.FromBase64String(validationKey);            
        }
        #endregion
		
        #region ICrytpocraphicProvider Memebers
        /// <include file='doc\DPAPIProvider.uex' path='docs/doc[@for="DPAPIProvider.Encrypt"]/*'/>
        /// <devdoc>
        ///	Encrypts raw bytes of data.
        /// </devdoc>		       
        public Byte[] Encrypt (Byte[] plainText) {		            
            if( Object.Equals(plainText, null) ) {
                throw new ArgumentNullException("plainText");
            }
            return Encrypt(plainText, null);
        }       

        /// <include file='doc\DPAPIProvider.uex' path='docs/doc[@for="DPAPIProvider.Decrypt"]/*'/>
        /// <devdoc>
        ///	Decrypts raw bytes of data.
        /// </devdoc>		       
        public Byte[] Decrypt (Byte[] cipherText) {
            if( Object.Equals(cipherText, null)) {
                throw new ArgumentNullException("cipherText");
            }
            return Decrypt( cipherText, null );
        }

        /// <include file='doc\DPAPIProvider.uex' path='docs/doc[@for="DPAPIProvider.ComputeHash"]/*'/>
        /// <devdoc>
        ///	Computes a hash for data validation.
        /// </devdoc>        
        public Byte[] ComputeHash (Byte[] plainText) {
            if(Object.Equals(plainText, null)) {
                throw new ArgumentNullException("plainText");
            }

            return validationAlgorithm.ComputeHash(plainText);
        }
		
        /// <devdoc>
        ///	Create the algorithm for validation of data.
        /// </devdoc>        
        private KeyedHashAlgorithm CreateAlgorithmFromName(String name) {
            switch (name.ToLower(System.Globalization.CultureInfo.CurrentCulture)) {
                case "sha1":
                    return new HMACSHA1();
                case "3des":
                    return new MACTripleDES(); 
                default:
                    return new HMACSHA1();
            }			
        }
        #endregion
        

        #region Methods
        /// <include file='doc\DPAPIProvider.uex' path='docs/doc[@for="DPAPIProvider.Encrypt1"]/*'/>
        /// <devdoc>
        ///	Encrypts raw bytes of data with optional entropy.
        /// </devdoc>		       
        public Byte[] Encrypt (Byte[] plainText, Byte[] optionalEntropy ) {
            if( Object.Equals(plainText, null) ) {
                throw new ArgumentNullException("plainText");
            }

            DATA_BLOB plainTextBlob = new DATA_BLOB();
            DATA_BLOB cipherTextBlob = new DATA_BLOB();
            DATA_BLOB entropyBlob = new DATA_BLOB();

            try {				
                bool isValue;
                CRYPTPROTECT_PROMPTSTRUCT prompt = new CRYPTPROTECT_PROMPTSTRUCT();
                InitPromptstruct(ref prompt);
                int dwFlag;
                int bytesSize;
                byte[] cipherText;
				
				
                bytesSize = plainText.Length;
                plainTextBlob.plainBlobData = Marshal.AllocHGlobal(bytesSize);
                if( Object.Equals(IntPtr.Zero, plainTextBlob.plainBlobData ) ) {
                    throw new CryptographicException(ResourceTable.GetString(ResourceTable.PlainTextBuffErr));
                }

                plainTextBlob.cipherBlobData = bytesSize;
                try {
                    // Copy to unmanaged area 
                    Marshal.Copy( plainText,
                        0, 
                        plainTextBlob.plainBlobData,
                        bytesSize );
                } catch(ArgumentOutOfRangeException encryptException) {
                    throw new CryptographicException(ResourceTable.GetString(ResourceTable.MarshalDataErr), encryptException);
                }
                if (!Object.Equals(DPAPIStore.Machine,dataProtectionStore)) {
                    // Using the machine store, should be providing entropy
                    dwFlag = CRYPTPROTECT_LOCAL_MACHINE |
                        CRYPTPROTECT_UI_FORBIDDEN;

                    // Check to see if the entropy is null
                    if (Object.Equals(optionalEntropy, null) ) {
                        // Allocate something
                        optionalEntropy = new byte[0];
                    }
					
                    bytesSize = optionalEntropy.Length;
                    entropyBlob.plainBlobData = Marshal.AllocHGlobal(optionalEntropy.Length);

                    if (!Object.Equals(IntPtr.Zero,
                        entropyBlob.plainBlobData) ) {
                        throw new CryptographicException(ResourceTable.GetString(ResourceTable.EntropyBuffErr));
                    }
                    try {
                        // Copy to unmanaged area
                        Marshal.Copy(optionalEntropy,
                            0, 
                            entropyBlob.plainBlobData, 
                            bytesSize );
                        entropyBlob.cipherBlobData = bytesSize;
                    } catch(ArgumentOutOfRangeException encryptException ) {
                        throw new CryptographicException(ResourceTable.GetString(ResourceTable.MarshalEntropyErr), encryptException );
                    }
                } else {
                    // Using the user store
                    dwFlag = CRYPTPROTECT_UI_FORBIDDEN;
                }

                isValue = CryptProtectData( ref plainTextBlob, 
                    String.Empty, 
                    ref entropyBlob, 
                    IntPtr.Zero, 
                    ref prompt, 
                    dwFlag, 
                    ref cipherTextBlob);                
				

                if (Object.Equals(isValue, false)) {
                    throw new CryptographicException(String.Concat(ResourceTable.GetString(ResourceTable.EncryptionFailed), GetErrorMessage(Marshal.GetLastWin32Error())));
                }

                cipherText = new byte[ cipherTextBlob.cipherBlobData ];
                try {
                    // Copy back to managed area
                    Marshal.Copy(cipherTextBlob.plainBlobData,
                        cipherText,
                        0, 
                        cipherTextBlob.cipherBlobData );
                } catch(ArgumentOutOfRangeException encryptException ) {
                    throw new CryptographicException(encryptException.Message, encryptException );
                }
                return cipherText;
            } finally {
                if (!Object.Equals(IntPtr.Zero,
                    cipherTextBlob.plainBlobData)) {
                    Marshal.FreeHGlobal(cipherTextBlob.plainBlobData);
                }
                if (!Object.Equals(IntPtr.Zero,
                    entropyBlob.plainBlobData)) {
                    Marshal.FreeHGlobal(entropyBlob.plainBlobData);
                }
                if (!Object.Equals(IntPtr.Zero,
                    plainTextBlob.plainBlobData)) {
                    Marshal.FreeHGlobal(plainTextBlob.plainBlobData);
                }
            }
        }	
        
		/// <include file='doc\DPAPIProvider.uex' path='docs/doc[@for="DPAPIProvider.Decrypt1"]/*'/>
        /// <devdoc>
        ///	Decrypts raw bytes of data.
        /// </devdoc>		       
        public Byte[] Decrypt(Byte[] cipherText, Byte[] optionalEntropy) {
            if( Object.Equals(cipherText, null)) {
                throw new ArgumentNullException("cipherText");
            }

            DATA_BLOB plainTextBlob = new DATA_BLOB();
            DATA_BLOB cipherBlob = new DATA_BLOB();
            DATA_BLOB entropyBlob = new DATA_BLOB();	

            try {
                bool isValue;
                CRYPTPROTECT_PROMPTSTRUCT prompt = new CRYPTPROTECT_PROMPTSTRUCT();
                InitPromptstruct(ref prompt);
                int cipherTextSize;
                int dwFlag;
                int bytesSize;
                byte[] plainText;				
                try {
                    cipherTextSize = cipherText.Length;
                    cipherBlob.plainBlobData = Marshal.AllocHGlobal(cipherTextSize);

                    if (Object.Equals(IntPtr.Zero, cipherBlob.plainBlobData)) {
                        throw new CryptographicException(ResourceTable.GetString(ResourceTable.CipherTextErr));
                    }
                    cipherBlob.cipherBlobData = cipherTextSize;
                    Marshal.Copy(cipherText,
                        0, 
                        cipherBlob.plainBlobData, 
                        cipherBlob.cipherBlobData );
                }
                catch(ArgumentOutOfRangeException decryptException) {
                    throw new CryptographicException(ResourceTable.GetString(ResourceTable.MarshalDataErr), decryptException);
                }
                //				DATA_BLOB entropyBlob = new DATA_BLOB();				

                if (Object.Equals(DPAPIStore.Machine, dataProtectionStore)) {
                    // Using the machine store, should be providing entropy
                    dwFlag = CRYPTPROTECT_LOCAL_MACHINE | 
                        CRYPTPROTECT_UI_FORBIDDEN;

                    // Check to see if the entropy is null
                    if(optionalEntropy == null) {
                        // Allocate something
                        optionalEntropy = new byte[0];
                    }
                    bytesSize = optionalEntropy.Length;
                    entropyBlob.plainBlobData = Marshal.AllocHGlobal(
                        bytesSize);
                    if( Object.Equals(IntPtr.Zero, 
                        entropyBlob.plainBlobData) ) {
                        throw new CryptographicException(ResourceTable.GetString(ResourceTable.EntropyBuffErr));								
                    }
                    entropyBlob.cipherBlobData = bytesSize;

                    try {						
                        // Copy to unmanaged area
                        Marshal.Copy(optionalEntropy, 
                            0, 
                            entropyBlob.plainBlobData,
                            bytesSize);
                    }
                    catch(ArgumentOutOfRangeException decryptException ) {
                        throw new CryptographicException(ResourceTable.GetString(ResourceTable.MarshalEntropyErr), decryptException);
                    }
                }
                else {
                    // Using the user store
                    dwFlag = CRYPTPROTECT_UI_FORBIDDEN;
                }
                isValue = CryptUnprotectData(ref cipherBlob,
                    null,
                    ref entropyBlob, 
                    IntPtr.Zero,
                    ref prompt,
                    dwFlag, 
                    ref plainTextBlob);

                if (Object.Equals(isValue, false)) {
                    throw new CryptographicException(String.Concat(ResourceTable.GetString(ResourceTable.DecryptionFailed), GetErrorMessage(Marshal.GetLastWin32Error())));
                }

                plainText = new byte[plainTextBlob.cipherBlobData];
                // Copy to managed area
                try {
                    Marshal.Copy( plainTextBlob.plainBlobData,
                        plainText,
                        0, 
                        plainTextBlob.cipherBlobData );
                } catch(ArgumentOutOfRangeException encryptException ) {
                    throw new CryptographicException(encryptException.Message, encryptException );
                }

                return plainText;
            }			
            finally {
                if(! Object.Equals(IntPtr.Zero,
                    cipherBlob.plainBlobData)) {
                    Marshal.FreeHGlobal(cipherBlob.plainBlobData);
                }
                if(! Object.Equals(IntPtr.Zero,
                    entropyBlob.plainBlobData)) {
                    Marshal.FreeHGlobal(entropyBlob.plainBlobData);
                }
                if(!Object.Equals(IntPtr.Zero,
                    plainTextBlob.plainBlobData)) {
                    Marshal.FreeHGlobal(plainTextBlob.plainBlobData);
                }
            }            
        }	

        /// <devdoc>
        ///	Initialize the value for the Promptstruct.
        /// </devdoc>		
        private void InitPromptstruct(ref CRYPTPROTECT_PROMPTSTRUCT promptStruct ) {
            promptStruct.cipherBlobSize = Marshal.SizeOf(typeof(CRYPTPROTECT_PROMPTSTRUCT));
            promptStruct.dwPromptFlags = 0;
            promptStruct.hwndApp = IntPtr.Zero;
            promptStruct.szPrompt = null;			
        }

        /// <devdoc>
        ///	Gets the error message.
        /// </devdoc>		
        private unsafe static String GetErrorMessage ( int errorCode ) {
            int     FORMAT_MESSAGE_ALLOCATE_BUFFER = 0x00000100;
            int     FORMAT_MESSAGE_IGNORE_INSERTS = 0x00000200;
            int     FORMAT_MESSAGE_FROM_SYSTEM  = 0x00001000;
            int     messageSize = 255;
            String  lpMsgBuf = "";
            int     dwFlag = FORMAT_MESSAGE_ALLOCATE_BUFFER | FORMAT_MESSAGE_FROM_SYSTEM | FORMAT_MESSAGE_IGNORE_INSERTS;
            IntPtr  ptrlpSource = new IntPtr();
            IntPtr  prtArguments = new IntPtr();
            		
            int isValue = FormatMessage( dwFlag,
                ref ptrlpSource, 
                errorCode,
                0, 
                ref lpMsgBuf,
                messageSize, 
                &prtArguments );

            if( isValue != 0) {
                throw new CryptographicException(String.Concat(ResourceTable.GetString(ResourceTable.FormatFailed), errorCode));
            }
            return lpMsgBuf;			
        }		
        #endregion
    }
}