//=============================================================================
// Copyright  2004 Microsoft Corporation
// All rights reserved.
// THIS CODE AND INFORMATION IS PROVIDED "AS IS" WITHOUT WARRANTY
// OF ANY KIND, EITHER EXPRESSED OR IMPLIED, INCLUDING BUT NOT
// LIMITED TO THE IMPLIED WARRANTIES OF MERCHANTABILITY AND/OR
// FITNESS FOR A PARTICULAR PURPOSE.
//=============================================================================
using System;
using System.IO;
using System.IO.IsolatedStorage;
using System.Configuration;
using System.Runtime.Serialization.Formatters.Binary;
using Microsoft.ApplicationBlocks.Cache;
using Microsoft.ApplicationBlocks.ExceptionManagement;
using System.Xml;
using System.Xml.XPath;
using System.Globalization;
using System.Runtime.CompilerServices;
using System.Collections;

namespace Microsoft.ApplicationBlocks.SmartClient.Offline
{
	/// <summary>
	/// Storage Provider for Cache Manager that stores data in Isolated Storage. The Cache Manager ensures
	/// that the key and data are not null before they are passed to any method of this class.
	/// </summary>
	public class IsolatedStorageCacheStorage : ICacheStorage, ICacheMetadataWithCallback
	{
		private IsolatedStorageFile storageSpace;
		private IsolatedStorageElement.Signed isSigned = IsolatedStorageElement.Signed.IsNotSigned;
		private IsolatedStorageElement.Encrypted isEncrypted = IsolatedStorageElement.Encrypted.IsNotEncrypted;
		private const string DirectoryRootName = ".Cache";        
		private const string DirectoryName = DirectoryRootName + "/";
        private const string MetaDataDirectoryRootName = DirectoryName + "MetaData";
        private const string MetaDataDirectoryName = MetaDataDirectoryRootName + "/";

		/// <summary>
		/// Getter property to return the number of elements in the cache
		/// </summary>
		/// <value>Number of elements in the cache</value>
		public int Size
		{
			[MethodImpl(MethodImplOptions.Synchronized)]
			get 
			{ 
				string key = CreateQualifiedKeyName("*");
				object [] elementsInIsolatedStorage = storageSpace.GetFileNames(key);
				return elementsInIsolatedStorage.Length;
			}
		}

        /// <summary>
        /// Getter property to return the number of elements in the metadata cache
        /// </summary>
        /// <value>Number of elements in the metadata cache</value>
        public int MetadataSize {
            [MethodImpl(MethodImplOptions.Synchronized)]
            get { 
                string key = CreateMetadataQualifiedKeyName("*");
                object [] elementsInIsolatedStorage = storageSpace.GetFileNames(key);
                return elementsInIsolatedStorage.Length;
            }
        }

		/// <summary>
		/// Returns the current size in bytes of the isolated storage area
		/// </summary>
		/// <value>Size in bytes of the isolated storage area</value>
		public ulong CurrentSize { get { return storageSpace.CurrentSize; }}

		/// <summary>
		/// Property to get and set the signed property of elements in the cache. The setter for this property exists for testing
		/// purposes and should not be called from client code
		/// </summary>
		/// <value>Stores whether the element in cache is signed or not</value>
		public bool IsSigned 
		{ 
			get { return isSigned == IsolatedStorageElement.Signed.IsSigned; }
			set { isSigned = value ? IsolatedStorageElement.Signed.IsSigned : IsolatedStorageElement.Signed.IsNotSigned; }
		}

		/// <summary>
		/// Property to get and set the encrypted property of elements in the cache. The setter for this property exists for
		/// testing purposes and should not be called from client code
		/// </summary>
		/// <value>Stores whether the element in cache is encrypted or not</value>
		public bool IsEncrypted
		{
			get { return isEncrypted == IsolatedStorageElement.Encrypted.IsEncrypted; }
			set { isEncrypted = value ? IsolatedStorageElement.Encrypted.IsEncrypted : IsolatedStorageElement.Encrypted.IsNotEncrypted; }
		}

		/// <summary>
		/// Constructor for instance. Initializes isolated storage based on user, domain, and application
		/// </summary>
		public IsolatedStorageCacheStorage()
		{
			storageSpace = IsolatedStorageFile.GetUserStoreForDomain();
			if(DoesCacheItemDirectoryExist() == false)
			{
				storageSpace.CreateDirectory(DirectoryRootName);
			}
            if(DoesMetadataDirectoryExist() == false) {
                storageSpace.CreateDirectory(MetaDataDirectoryRootName);
            }
		}

		/// <summary>
		/// Method used to initialize the provider with information from configuration file. Called from cache block during initialization
		/// </summary>
		/// <param name="configSection">Section in the config file used for initialization</param>
		public void Init(XmlNode configSection)
		{
			// Variable Declarations
			string validatedString;
			string encryptedString;

			const string CONST_TRUE = "true";
			const string CONST_FALSE = "false";

			try
			{
				if(Object.Equals(configSection, null))
				{
					throw(new ArgumentNullException("configSection",
						CacheResources.ResourceManager[
						"RES_ExceptionNullConfigSection"]));
				}
			
				XmlAttribute validatedAttribute = configSection.Attributes[CacheResources.ResourceManager["RES_ConfigValidated"]];
				if( ! Object.Equals( configSection.Attributes[
					CacheResources.ResourceManager[ 
					"RES_ConfigValidated" ] ], null ) ) 
				{
					validatedString = configSection.Attributes[CacheResources.
						ResourceManager["RES_ConfigValidated"]].Value ;
				}
				else
				{
					throw(new ConfigurationException(CacheResources.
						ResourceManager["RES_ExceptionEmptyNodeValidated"]));
				}

				if(Object.Equals(validatedString, null))
				{
					throw(new ConfigurationException(CacheResources.
						ResourceManager["RES_ExceptionNullValidated"]));
				}
				else if (validatedString.Length == 0)
				{
					throw(new ConfigurationException(CacheResources.
						ResourceManager["RES_ExceptionEmptyValidated"]));
				}
				else if(!(Object.Equals(validatedString.ToLower(CultureInfo.CurrentCulture), CONST_TRUE)
					|| Object.Equals(validatedString.ToLower(CultureInfo.CurrentCulture), CONST_FALSE)))
				{
					throw(new ConfigurationException(CacheResources.
						ResourceManager["RES_ExceptionInvalidValidated"]));
				}
				else
				{
					IsSigned = bool.Parse(validatedString.ToLower(
						CultureInfo.CurrentCulture));
				}				
				
				if( ! Object.Equals(configSection.Attributes[
					CacheResources.ResourceManager[
					"RES_ConfigEncrypted" ] ], null) ) 
				{
					encryptedString = configSection.Attributes[CacheResources.
						ResourceManager["RES_ConfigEncrypted"]].Value;
				}
				else
				{
					throw(new ConfigurationException(CacheResources.
						ResourceManager["RES_ExceptionEmptyNodeEncrypted"]));
				}

				if(Object.Equals(encryptedString, null))
				{
					throw(new ConfigurationException(CacheResources.
						ResourceManager["RES_ExceptionNullEncrypted"]));
				}
				else if(encryptedString.Length == 0)
				{
					throw(new ConfigurationException(CacheResources.
						ResourceManager["RES_ExceptionEmptyEncrypted"]));
				}
				else if(!(Object.Equals(encryptedString.ToLower(CultureInfo.CurrentCulture),CONST_TRUE)
					|| Object.Equals(encryptedString.ToLower(CultureInfo.CurrentCulture), CONST_FALSE)))
				{
					throw(new ConfigurationException(CacheResources.
						ResourceManager["RES_ExceptionInvalidEncrypted"]));
				}
				else
				{
					IsEncrypted = bool.Parse(encryptedString.ToLower(
						CultureInfo.CurrentCulture));
				}
			}
			catch( Exception genException )
			{
				ExceptionManager.Publish( genException );
				throw;
			}
		}

		/// <summary>
		/// Removes an element from the cache
		/// </summary>
		/// <param name="key">Key of element to remove</param>
		[MethodImpl(MethodImplOptions.Synchronized)]
		public void Remove(string key)
		{
			key = CreateQualifiedKeyName(key);
			try
			{
				if(FileExists(key)) storageSpace.DeleteFile(key);
			}
			catch(Exception e)
			{
				ExceptionManager.Publish(e);
				throw;
			}
		}

		/// <summary>
		/// Retrieves an element from cache
		/// </summary>
		/// <param name="key">Key of element to retrieve</param>
		/// <returns>Object retrieved from cache</returns>
		[MethodImpl(MethodImplOptions.Synchronized)]
		public object GetData(string key)
		{
			key = CreateQualifiedKeyName(key);
			try
			{
				if(FileExists(key) == false) return null;
				
				using(IsolatedStorageElement storageElement = IsolatedStorageElement.OpenElement(storageSpace, key, isSigned, isEncrypted))
				{
					return storageElement.Read();
				}
			}
			catch(Exception e)
			{
				ExceptionManager.Publish(e);
				throw;
			}
		}

		/// <summary>
		/// Updates an element in cache. This is semantically exactly the same as adding an element
		/// </summary>
		/// <param name="key">Key of element to update</param>
		/// <param name="keyData">Object to be updated in cache</param>
		public void Update(string key, object keyData)
		{
			Add(key, keyData);
		}

		/// <summary>
		/// Adds a new element into the cache
		/// </summary>
		/// <param name="key">Key to identify the new data</param>
		/// <param name="keyData">Object to be stored in cache</param>
		[MethodImpl(MethodImplOptions.Synchronized)]
		public void Add(string key, object keyData)
		{
			key = CreateQualifiedKeyName(key);
			try
			{
				using(IsolatedStorageElement element = IsolatedStorageElement.CreateElement(storageSpace, key, isSigned, isEncrypted))
				{
					element.Write(keyData);
				}
			}
			catch(Exception e)
			{
				ExceptionManager.Publish(e);
				throw;
			}
		}	

		/// <summary>
		/// Removes all entries from the cache
		/// </summary>
		[MethodImpl(MethodImplOptions.Synchronized)]
		public void Flush()
		{
			string key = CreateQualifiedKeyName("*");
			try
			{
				foreach(string fileName in storageSpace.GetFileNames(key))
				{
					Remove(fileName);
				}
			}
			catch(Exception e)
			{
				ExceptionManager.Publish(e);
				throw;
			}
		}	
		
		private bool DoesCacheItemDirectoryExist()
		{
			return storageSpace.GetDirectoryNames(DirectoryRootName).Length > 0;
		}
	
        private bool DoesMetadataDirectoryExist() {
            return storageSpace.GetDirectoryNames(MetaDataDirectoryRootName).Length > 0;
        }
	

		private string CreateQualifiedKeyName(string key)
		{
			return DirectoryName + key;
		}

        private string CreateMetadataQualifiedKeyName(string key) {
            return MetaDataDirectoryName + key;
        }

		/// <summary>
		/// Determines if a file exists in cache.
		/// </summary>
		/// <param name="qualifiedKey">Name to search for</param>
		/// <returns>True if item found, false if not found</returns>
		/// <exception cref="ArgumentException">Key passed to method contains invalid characters</exception>
		private bool FileExists(string qualifiedKey)
		{
			try
			{
				using(IsolatedStorageElement testFile = IsolatedStorageElement.OpenElement(storageSpace, qualifiedKey, isSigned, isEncrypted))
				{
					return true;
				}
			}
			catch(FileNotFoundException)
			{
				return false;
			}
			catch(DirectoryNotFoundException)
			{
				throw new ArgumentException(String.Format(CacheResources.ResourceManager["RES_InvalidCharactersInKeyName"], qualifiedKey), "qualifiedKey");
			}
        }
        #region ICacheMetadata Members
        [MethodImpl(MethodImplOptions.Synchronized)]
        Hashtable ICacheMetadata.GetMetadata() {            
            string key = CreateMetadataQualifiedKeyName("*");
            Hashtable metadata = new Hashtable();
            try {
                foreach(string file in storageSpace.GetFileNames(key)) {
                    string fileName = CreateMetadataQualifiedKeyName(file);
                    if(FileExists(fileName) == false) continue;
				
                    using(IsolatedStorageElement storageElement = IsolatedStorageElement.OpenElement(storageSpace, fileName, IsolatedStorageElement.Signed.IsNotSigned, IsolatedStorageElement.Encrypted.IsNotEncrypted)) {
                        MetadataItem item = storageElement.Read() as MetadataItem;
                        metadata.Add(metadata.Count.ToString(), new DictionaryEntry(file, item.Priority));
                        metadata.Add(metadata.Count.ToString(), new DictionaryEntry(file, item.Expirations));
						metadata.Add(metadata.Count.ToString(), new DictionaryEntry(file, item.Callback));
                    }
                }
            }
            catch(Exception e) {
                ExceptionManager.Publish(e);
                throw;
            }
            return metadata;
        }

        [MethodImpl(MethodImplOptions.Synchronized)]
        void ICacheMetadata.Add(string key, ICacheItemExpiration[] expirations, CacheItemPriority priority) {
//            key = CreateMetadataQualifiedKeyName(key);
//            try {
//                using(IsolatedStorageElement element = IsolatedStorageElement.CreateElement(storageSpace, key, IsolatedStorageElement.Signed.IsNotSigned, IsolatedStorageElement.Encrypted.IsNotEncrypted)) {
//                    element.Write(new MetadataItem(expirations, priority));
//                }
//            }
//            catch(Exception e) {
//                ExceptionManager.Publish(e);
//                throw;
//            }
			((ICacheMetadataWithCallback)this).Add(key, expirations, priority, null);
        }

		[MethodImpl(MethodImplOptions.Synchronized)]
		void ICacheMetadataWithCallback.Add(string key, ICacheItemExpiration[] expirations, CacheItemPriority priority, CacheItemRemovedCallback callback) {
			key = CreateMetadataQualifiedKeyName(key);
			try {
				using(IsolatedStorageElement element = IsolatedStorageElement.CreateElement(storageSpace, key, IsolatedStorageElement.Signed.IsNotSigned, IsolatedStorageElement.Encrypted.IsNotEncrypted)) {
					element.Write(new MetadataItem(expirations, priority, callback));
				}
			}
			catch(Exception e) {
				ExceptionManager.Publish(e);
				throw;
			}
		}

        [MethodImpl(MethodImplOptions.Synchronized)]
        void ICacheMetadata.Remove(string key) {
            key = CreateMetadataQualifiedKeyName(key);
            try {
                if(FileExists(key)) storageSpace.DeleteFile(key);
            }
            catch(Exception e) {
                ExceptionManager.Publish(e);
                throw;
            }
        }

        [MethodImpl(MethodImplOptions.Synchronized)]
        void ICacheMetadata.Flush() {
            string key = CreateMetadataQualifiedKeyName("*");
            try {
                foreach(string fileName in storageSpace.GetFileNames(key)) {
                    ((ICacheMetadata)this).Remove(fileName);
                }
            }
            catch(Exception e) {
                ExceptionManager.Publish(e);
                throw;
            }
        }
        #endregion
    }
}
