//=============================================================================
// Copyright  2004 Microsoft Corporation
// All rights reserved.
// THIS CODE AND INFORMATION IS PROVIDED "AS IS" WITHOUT WARRANTY
// OF ANY KIND, EITHER EXPRESSED OR IMPLIED, INCLUDING BUT NOT
// LIMITED TO THE IMPLIED WARRANTIES OF MERCHANTABILITY AND/OR
// FITNESS FOR A PARTICULAR PURPOSE.
//=============================================================================
using System;
using System.IO;
using System.Runtime.Serialization.Formatters.Binary;
using System.IO.IsolatedStorage;
using Microsoft.ApplicationBlocks.Cache;

namespace Microsoft.ApplicationBlocks.SmartClient.Offline
{
	/// <summary>
	/// This class encapsulates the behavior associated with managing an element in isolated storage. IsolatedStorage acts like a hidden
	/// filesystem, so entries in it are created as files and directories. The process of creating or opening an IsolagedStorageElement
	/// acts on an entry in this filesystem.
	/// </summary>
	public class IsolatedStorageElement : IDisposable
	{
		private IsolatedStorageFileStream elementStream;
		private Signed isSigned = Signed.IsNotSigned;
		private Encrypted isEncrypted = Encrypted.IsNotEncrypted;

		/// <summary>
		/// Describes whether or not the elements in isolated storage are signed
		/// </summary>
		public enum Signed { IsSigned, IsNotSigned };

		/// <summary>
		/// Describes whether or not the elements in isolated storage are encrypted
		/// </summary>
		public enum Encrypted { IsEncrypted, IsNotEncrypted };

		/// <summary>
		/// Getter property for returning the length of the element in isolated storage
		/// </summary>
		public long Length { get { return elementStream.Length; }}

		/// <summary>
		/// Static constructor method. Call this method when you are opening an existing element
		/// </summary>
		/// <param name="storageSpace">The isolated storage area</param>
		/// <param name="key">File name to open</param>
		/// <param name="isSigned">Whether the data is signed</param>
		/// <param name="isEncrypted">Whether the data is encrypted</param>
		/// <returns>New instance of object referring to opened item</returns>
		public static IsolatedStorageElement OpenElement(IsolatedStorageFile storageSpace, string key, Signed isSigned, Encrypted isEncrypted)
		{
			return new IsolatedStorageElement(key, FileMode.Open, FileAccess.Read, FileShare.None, storageSpace, isSigned, isEncrypted);
		}

		/// <summary>
		/// Static constructor method. Call this method when you are creating a new element or overwriting an existing element
		/// </summary>
		/// <param name="storageSpace">The isolated storage area</param>
		/// <param name="key">File name to open</param>
		/// <param name="isSigned">Whether the data is signed</param>
		/// <param name="isEncrypted">Whether the data is encrypted</param>
		/// <returns>New instance of object referring to new or truncated item</returns>
		public static IsolatedStorageElement CreateElement(IsolatedStorageFile storageSpace, string key, Signed isSigned, Encrypted isEncrypted)
		{
			return new IsolatedStorageElement(key, FileMode.Create, FileAccess.Write, FileShare.None, storageSpace, isSigned, isEncrypted);
		}

		private IsolatedStorageElement(string key, FileMode fileMode, FileAccess fileAccess, FileShare fileShare, IsolatedStorageFile storageSpace, Signed isSigned, Encrypted isEncrypted)
		{
			this.elementStream = new IsolatedStorageFileStream(key, fileMode, fileAccess, fileShare, storageSpace);
			this.isSigned = isSigned;
			this.isEncrypted = isEncrypted;
		}

		/// <summary>
		/// Dispose method for this class. The compiler would not allow a call to elementStream.Dispose(), so implemented
		/// as manual calls to flush and close
		/// </summary>
		public void Dispose()
		{
			elementStream.Flush();
			elementStream.Close();
		}

		/// <summary>
		/// Reads an item out of isolated storage. It will sign and decrypt the item as necessary
		/// </summary>
		/// <returns>Object retrieved from isolated storage</returns>
		public object Read()
		{
			byte [] data = new byte[elementStream.Length];
			elementStream.Read(data, 0, data.Length);

			if(isEncrypted == Encrypted.IsEncrypted)
			{
				data = DataProtectionManager.Decrypt(data);
			}

			if(isSigned == Signed.IsSigned)
			{
				data = DataProtectionManager.RemoveMAC(data);
			}

			return ToObject(data);
		}

		/// <summary>
		/// Writes an item to isolated storage. It will sign and encrypt the item as necessary
		/// </summary>
		/// <param name="objectToWrite"></param>
		public void Write(object objectToWrite)
		{
			byte [] dataInBytes = ToBytes(objectToWrite);

			if(isSigned == Signed.IsSigned)
			{
				dataInBytes = DataProtectionManager.AppendMAC(dataInBytes);
			}

			if(isEncrypted == Encrypted.IsEncrypted)
			{
				dataInBytes = DataProtectionManager.Encrypt(dataInBytes);
			}

			elementStream.Write(dataInBytes, 0, dataInBytes.Length);
		}

		/// <summary>
		/// Converts an object to a stream of bytes
		/// </summary>
		/// <param name="objectToSerialize"></param>
		/// <returns></returns>
		private byte[] ToBytes(object objectToSerialize)
		{
			byte [] inMemoryBytes; 
			using(MemoryStream inMemoryData = new MemoryStream())
			{
				new BinaryFormatter().Serialize(inMemoryData, objectToSerialize);
				inMemoryBytes = inMemoryData.ToArray();
			}

			return inMemoryBytes;
		}

		/// <summary>
		/// Converts a stream of bytes into an object
		/// </summary>
		/// <param name="dataBytes"></param>
		/// <returns></returns>
		private object ToObject(byte [] dataBytes)
		{
			using(MemoryStream dataInMemory = new MemoryStream(dataBytes))
			{
				object userData = new BinaryFormatter().Deserialize(dataInMemory);

				return userData;
			}
		}
	}
}
