//=============================================================================
// Copyright  2004 Microsoft Corporation
// All rights reserved.
// THIS CODE AND INFORMATION IS PROVIDED "AS IS" WITHOUT WARRANTY
// OF ANY KIND, EITHER EXPRESSED OR IMPLIED, INCLUDING BUT NOT
// LIMITED TO THE IMPLIED WARRANTIES OF MERCHANTABILITY AND/OR
// FITNESS FOR A PARTICULAR PURPOSE.
//=============================================================================
using System;
using System.Runtime.InteropServices;
using System.Xml.XPath;
using System.Configuration;
using Microsoft.ApplicationBlocks.Common;

namespace Microsoft.ApplicationBlocks.SmartClient.Offline
{
	/// <summary>
	/// This is a simple implementation of a ConnectionDetectionStrategy. It uses the WinINet
	/// function GetInternetConnectionState to detect whether you are actually connected to 
	/// another system anywhere. This strategy is not smart enough to tell if you can talk
	/// to a particular host -- it merely checks to see that you can talk to *any* host. 
	/// This means that there is an ethernet connection to another computer or a network,
	/// or that a modem is connected.               
	/// </summary>
	public class WinINetDetectionStrategy : IConnectionDetectionStrategy, IProvider
	{
		private int pollInterval;

		/// <summary>
		/// Public constant used in test code.
		/// </summary>
		public const int MinimumPollingIntervalInSeconds = 1;

		[DllImport("wininet.dll")]
		private extern static bool InternetGetConnectedState( out int connectionDescription, int reservedValue ) ;

		/// <summary>
		/// Getter property to retrieve the polling interval in seconds
		/// </summary>
		/// <value>Polling interaval in seconds</value>
		public int PollInterval { get { return pollInterval; }}

		/// <summary>
		/// Query method to cause provider to actively detect the connection state
		/// </summary>
		/// <returns>True if provider believes we are connected</returns>
		public bool IsConnected()
		{
			int connectionDescription = 0;
			return InternetGetConnectedState(out connectionDescription, 0);
		}

		/// <summary>
		/// IProvider.Initialize method implementation. Retrieves configuration information from app.config file
		/// </summary>
		/// <param name="configurationNode">XmlNode to parse for configuration information</param>
		public void Initialize(XPathNavigator configurationNode)
		{
			XPathNodeIterator iter = configurationNode.Select("pollingInterval");
			iter.MoveNext();

			pollInterval = GetCurrentPollingInterval(iter.Current.Value);
			if(pollInterval < MinimumPollingIntervalInSeconds)
			{
				throw new ConfigurationException(ProvidersResourceTable.GetString(ProvidersResourceTable.ConnectionManagerPollingIntervalMessage));
			}
		}

		private int GetCurrentPollingInterval(string pollingIntervalString)
		{
			try
			{
				return Convert.ToInt32(pollingIntervalString);
			}
			catch(FormatException e)
			{
				throw new ConfigurationException(ProvidersResourceTable.GetString(ProvidersResourceTable.PollingIntervalShouldBeNumeric), e);
			}
		}
	}
}
