//=============================================================================
// Copyright  2004 Microsoft Corporation
// All rights reserved.
// THIS CODE AND INFORMATION IS PROVIDED "AS IS" WITHOUT WARRANTY
// OF ANY KIND, EITHER EXPRESSED OR IMPLIED, INCLUDING BUT NOT
// LIMITED TO THE IMPLIED WARRANTIES OF MERCHANTABILITY AND/OR
// FITNESS FOR A PARTICULAR PURPOSE.
//=============================================================================
using System;
using System.IO;
using System.IO.IsolatedStorage;

namespace Microsoft.ApplicationBlocks.SmartClient.Offline
{
	/// <summary>
	/// This class encapsulates the behavior associated with managing an element in isolated storage. IsolatedStorage acts like a hidden
	/// filesystem, so entries in it are created as files and directories. The process of creating or opening an IsolagedStorageElement
	/// acts on an entry in this filesystem.
	/// </summary>
	/// <devdoc>The intention was to share this implementation with the IsolatedStorageElement used in the cache, but
	/// there is no administrative way to share code between application blocks
	/// </devdoc>
	public class IsolatedStorageQueueElement : IDisposable
	{
		private IsolatedStorageFileStream elementStream;

		/// <summary>
		/// Getter property for returning the length of the element in isolated storage
		/// </summary>
		public long Length { get { return elementStream.Length; }}

		/// <summary>
		/// Static constructor method. Call this method when you are opening an existing element
		/// </summary>
		/// <param name="storageSpace">The isolated storage area</param>
		/// <param name="key">File name to open</param>
		/// <returns>New instance of object referring to opened item</returns>
		public static IsolatedStorageQueueElement OpenElement(IsolatedStorageFile storageSpace, string key)
		{
			return new IsolatedStorageQueueElement(key, FileMode.Open, FileAccess.ReadWrite, FileShare.None, storageSpace);
		}

		/// <summary>
		/// Static constructor method. Call this method when you are creating a new element or overwriting an existing element
		/// </summary>
		/// <param name="storageSpace">The isolated storage area</param>
		/// <param name="key">File name to open</param>
		/// <returns>New instance of object referring to new or truncated item</returns>
		public static IsolatedStorageQueueElement CreateElement(IsolatedStorageFile storageSpace, string key)
		{
			return new IsolatedStorageQueueElement(key, FileMode.Create, FileAccess.ReadWrite, FileShare.None, storageSpace);
		}

		private IsolatedStorageQueueElement(string key, FileMode fileMode, FileAccess fileAccess, FileShare fileShare, IsolatedStorageFile storageSpace)
		{
			this.elementStream = new IsolatedStorageFileStream(key, fileMode, fileAccess, fileShare, storageSpace);
		}

		/// <summary>
		/// Dispose method for this class. The compiler would not allow a call to elementStream.Dispose(), so implemented
		/// as manual calls to flush and close
		/// </summary>
		public void Dispose()
		{
			elementStream.Flush();
			elementStream.Close();
		}

		/// <summary>
		/// Reads an item out of isolated storage. It will sign and decrypt the item as necessary
		/// </summary>
		/// <returns>Object retrieved from isolated storage</returns>
		public object Read()
		{
			byte [] data = new byte[elementStream.Length];
			elementStream.Read(data, 0, data.Length);

			return StreamingUtils.ToObject(data);
		}

		/// <summary>
		/// Writes an item to isolated storage. It will sign and encrypt the item as necessary
		/// </summary>
		/// <param name="objectToWrite"></param>
		public void Write(object objectToWrite)
		{
			byte [] dataInBytes = StreamingUtils.ToBytes(objectToWrite);
			elementStream.Write(dataInBytes, 0, dataInBytes.Length);
		}
	}
}

