//=============================================================================
// Copyright  2004 Microsoft Corporation
// All rights reserved.
// THIS CODE AND INFORMATION IS PROVIDED "AS IS" WITHOUT WARRANTY
// OF ANY KIND, EITHER EXPRESSED OR IMPLIED, INCLUDING BUT NOT
// LIMITED TO THE IMPLIED WARRANTIES OF MERCHANTABILITY AND/OR
// FITNESS FOR A PARTICULAR PURPOSE.
//=============================================================================
using System;

namespace Microsoft.ApplicationBlocks.SmartClient.Offline
{
	/// <summary>
	/// This class describes the organization of the queue elements stored in IsolatedStorage. It does not attempt to do 
	/// any sanity checking of how it is being called, as its caller is supposed to be doing that for it. It does maintain
	/// the names of the files that are supposed to be used to enqueue the next element and dequeue the next element, and 
	/// know how many elements are current stored in IsolatedStorage.
	/// </summary>
	/// 
	[Serializable()]
	public class IsolatedStorageQueueStorageHeader
	{
		private int nextEnqueueIndex = -1;
		private int nextDequeueIndex = -1;
		private int numberOfElements = 0;
		private int pivotPoint;

		/// <summary>
		/// Getter property to return the number of elements stored in isolated storage
		/// </summary>
		public int Size { get { return numberOfElements; }}

		/// <summary>
		/// Name of the next file to use when enqueuing a message
		/// </summary>
		public string EnqueueFileName { get { return nextEnqueueIndex.ToString(); }}

		/// <summary>
		/// Name of the next file to use when dequeuing a message
		/// </summary>
		public string DequeueFileName {	get { return nextDequeueIndex.ToString(); }}

		/// <summary>
		/// Constructor for instance
		/// </summary>
		/// <param name="pivotPoint">The point at which the file names wrap around back to 0 and start incrementing again</param>
		public IsolatedStorageQueueStorageHeader(int pivotPoint)
		{
			this.pivotPoint = pivotPoint;
		}

		/// <summary>
		/// Call this method before calling EnqueueFileName to add a new message to isolated storage. It does the housekeeping
		/// necessary to allot space for the new item
		/// </summary>
		public void IncrementIndex() 
		{
			nextEnqueueIndex = WrappedIncrement(nextEnqueueIndex);
			numberOfElements++;
		}

		/// <summary>
		/// Call this method before calling DequeueFileName to remove the oldest message from isolated storage. It does the housekeeping
		/// necessary to clean up the space allotted for the old item
		/// </summary>
		public void DecrementIndex()
		{
			numberOfElements--;
			nextDequeueIndex = WrappedIncrement(nextDequeueIndex);
		}

		/// <summary>
		/// Query method to determine if the queue is empty.
		/// </summary>
		/// <returns>True if there are no elements stored in isolated storage</returns>
		public bool IsQueueEmpty() { return Size == 0; }

		/// <summary>
		/// Query method to determine if the queue is full. Full refers to every possible file name being used with a value less
		/// than the pivot point.
		/// </summary>
		/// <returns>True if the queue is full</returns>
		public bool IsQueueFull() { return Size == pivotPoint; }

		/// <summary>
		/// Provides the wrapping functionality to manage the wrapping of the file names at the pivot point.l
		/// </summary>
		/// <param name="initialValue"></param>
		/// <returns></returns>
		private int WrappedIncrement(int initialValue)
		{
			initialValue++;
			if(initialValue == pivotPoint)
			{
				initialValue = 0;
			}

			return initialValue;
		}
	}
}
