//=============================================================================
// Copyright  2004 Microsoft Corporation
// All rights reserved.
// THIS CODE AND INFORMATION IS PROVIDED "AS IS" WITHOUT WARRANTY
// OF ANY KIND, EITHER EXPRESSED OR IMPLIED, INCLUDING BUT NOT
// LIMITED TO THE IMPLIED WARRANTIES OF MERCHANTABILITY AND/OR
// FITNESS FOR A PARTICULAR PURPOSE.
//=============================================================================
using System;
using System.IO.IsolatedStorage;
using System.IO;
using System.Xml.XPath;
using Microsoft.ApplicationBlocks.Common;
using Microsoft.ApplicationBlocks.ExceptionManagement;

namespace Microsoft.ApplicationBlocks.SmartClient.Offline
{
	/// <summary>
	/// This IQueueStorageProvider implementation will store queue messages into IsolatedStorage. The advantage of doing this
	/// is that it provides isolation between different QueueManager users, where the other two providers do not.
	/// </summary>
	public class IsolatedStorageQueueStorageProvider : QueueStorageProvider
	{
		private const string DirectoryRootName = ".Queue";
		private const string DirectoryName = DirectoryRootName + "/";
		private const int MaximumNumberOfMessages = 32000;

		private const string HeaderName = "QueueHeader";
		private IsolatedStorageFile storageSpace;
		private int maximumNumberOfMessages;

		/// <summary>
		/// Getter property to return the number of elements stored in isolated storage
		/// </summary>
		/// <value>Number of elements in the queue</value>
		override public int Size
		{
			get 
			{ 
				IsolatedStorageQueueStorageHeader header = GetQueueStorageHeader();
				return header.Size;
			}
		}

		/// <summary>
		/// Constructor for instance. Responsible for initializing isolated storage before use. This constructor should never
		/// be called by application code, but is provided for testing.
		/// </summary>
		/// <param name="maximumNumberOfMessages">Limit for the maximum number of Queue messages that can be stored</param>
		public IsolatedStorageQueueStorageProvider(int maximumNumberOfMessages)
		{
			storageSpace = IsolatedStorageFile.GetUserStoreForDomain();
			if(storageSpace.GetDirectoryNames(DirectoryRootName).Length == 0)
			{
				storageSpace.CreateDirectory(DirectoryRootName);
			}

			this.maximumNumberOfMessages = maximumNumberOfMessages;
		}

		/// <summary>
		/// Constructor for instance. Called through provider-initialization methods at startup
		/// </summary>
		public IsolatedStorageQueueStorageProvider() : this(MaximumNumberOfMessages)
		{
		}

		/// <summary>
		/// Removes all elements from isolated storage
		/// </summary>
		override public void Flush()
		{
			string key = CreateQualifiedKeyName("*");
			try
			{
				foreach(string fileName in storageSpace.GetFileNames(key))
				{
					storageSpace.DeleteFile(CreateQualifiedKeyName(fileName));
				}
			}
			catch(Exception e)
			{
				ExceptionManager.Publish(e);
				throw;
			}
		}

		/// <summary>
		/// Responsible for storing a byte array into isolated storage at the appropriate location, defined by the EnqueueFileName
		/// </summary>
		/// <param name="objectToStream">Object to store</param>
		override protected void DoEnqueue(byte[] objectToStream) 
		{ 
			IsolatedStorageQueueStorageHeader header = GetQueueStorageHeader();

			if(header.IsQueueFull()) throw new InvalidOperationException("Cannot enqueue message. Queue is full.");

			AddElement(header, objectToStream);
			UpdateHeader(header);
		}

		/// <summary>
		/// Responsible for retrieving a byte array from isolated storage from the appropriate location, defined by the DequeueFileName
		/// </summary>
		/// <returns></returns>
		override protected byte[] DoDequeue() 
		{ 
			IsolatedStorageQueueStorageHeader header = GetQueueStorageHeader();

			if(header.IsQueueEmpty()) throw new InvalidOperationException("Cannot dequeue from empty queue");

			byte[] queuedBytes = RemoveElement(header);
			UpdateHeader(header);

			storageSpace.DeleteFile(CreateQualifiedKeyName(header.DequeueFileName));

			return queuedBytes;
		}

		/// <summary>
		/// No implementation
		/// </summary>
		/// <param name="parameters">Not used</param>
		override protected void DoLocalSetup(XPathNavigator parameters) {}

		/// <summary>
		/// No implementation 
		/// </summary>
		override protected void Initialize() {}

		/// <summary>
		/// Responsible for reading an element from isolated storage and updating the header block
		/// </summary>
		/// <param name="header"></param>
		/// <returns></returns>
		private byte[] RemoveElement(IsolatedStorageQueueStorageHeader header)
		{
			header.DecrementIndex();
			string oldFileName = header.DequeueFileName;

			using(IsolatedStorageQueueElement oldElement = IsolatedStorageQueueElement.OpenElement(storageSpace, CreateQualifiedKeyName(oldFileName)))
			{
				return oldElement.Read() as byte[];
			}
		}

		private string CreateQualifiedKeyName(string key)
		{
			return DirectoryName + key;
		}

		/// <summary>
		/// Determines if a file exists in cache.
		/// </summary>
		/// <param name="qualifiedKey">Name to search for</param>
		/// <returns>True if item found, false if not found</returns>
		/// <exception cref="ArgumentException">key passed to method contains invalid characters</exception>
		private bool FileExists(string qualifiedKey)
		{
			try
			{
				using(IsolatedStorageQueueElement testFile = IsolatedStorageQueueElement.OpenElement(storageSpace, qualifiedKey))
				{
					return true;
				}
			}
			catch(FileNotFoundException)
			{
				return false;
			}
			catch(DirectoryNotFoundException)
			{
				throw new ArgumentException("Invalid characters in key name", qualifiedKey);
			}
		}

		/// <summary>
		/// Finds an existing header block in isolated storage or creates one if needed. If it creates on, it also writes it out, so the 
		/// rest of the code can proceed as if the header is always in isolated storage
		/// </summary>
		/// <returns></returns>
		private IsolatedStorageQueueStorageHeader GetQueueStorageHeader()
		{
			if(FileExists(CreateQualifiedKeyName(HeaderName)))
			{
				using(IsolatedStorageQueueElement headerElement = IsolatedStorageQueueElement.OpenElement(storageSpace, CreateQualifiedKeyName(HeaderName)))
				{
					return headerElement.Read() as IsolatedStorageQueueStorageHeader;
				}
			}
			else
			{
				using(IsolatedStorageQueueElement headerElement = IsolatedStorageQueueElement.CreateElement(storageSpace, CreateQualifiedKeyName(HeaderName)))
				{
					IsolatedStorageQueueStorageHeader header = new IsolatedStorageQueueStorageHeader(maximumNumberOfMessages);
					headerElement.Write(header);
					return header;
				}
			}
		}

		private void AddElement(IsolatedStorageQueueStorageHeader header, object objectToStream)
		{
			header.IncrementIndex();
			string newFileName = header.EnqueueFileName;

			using(IsolatedStorageQueueElement newElement = IsolatedStorageQueueElement.CreateElement(storageSpace, CreateQualifiedKeyName(newFileName)))
			{
				newElement.Write(objectToStream);
			}
		}

		private void UpdateHeader(IsolatedStorageQueueStorageHeader header)
		{
			using(IsolatedStorageQueueElement headerElement = IsolatedStorageQueueElement.CreateElement(storageSpace, CreateQualifiedKeyName(HeaderName)))
			{
				headerElement.Write(header);
			}
		}
	}
}
