//=============================================================================
// Copyright  2004 Microsoft Corporation
// All rights reserved.
// THIS CODE AND INFORMATION IS PROVIDED "AS IS" WITHOUT WARRANTY
// OF ANY KIND, EITHER EXPRESSED OR IMPLIED, INCLUDING BUT NOT
// LIMITED TO THE IMPLIED WARRANTIES OF MERCHANTABILITY AND/OR
// FITNESS FOR A PARTICULAR PURPOSE.
//=============================================================================
using System;
using System.Data;
using System.Data.SqlClient;
using System.IO;
using System.Runtime.Serialization.Formatters.Binary;
using System.Xml;
using System.Xml.XPath;
using System.Configuration;
using Microsoft.ApplicationBlocks.SmartClient.Offline;
using Microsoft.ApplicationBlocks.Data;
using Microsoft.ApplicationBlocks.Common;

namespace Microsoft.ApplicationBlocks.SmartClient.Offline
{
	/// <summary>
	/// Stores queue messages into MSDE. MSDE is a shared resource, so it should not be used if different users are going
	/// to use the application on the same machine.
	/// </summary>
	public class MSDEQueueStorageProvider : QueueStorageProvider
	{
		private SqlConnection connection;
		private SqlParameter applicationName;
		private string applicationNameLabel;
		private string connectionString;

		/// <summary>
		/// Getter property to get the application name as defined in the configuration file
		/// </summary>
		public string ApplicationName { get { return applicationNameLabel; }}

		/// <summary>
		/// Getter property to get the connection string as defined in the configuration file
		/// </summary>
		public string ConnectionString { get { return connectionString; }}

		/// <summary>
		/// Getter property to return the number of elements stored in MSDE
		/// </summary>
		public override int Size
		{
			get 
			{
				object itemCount = SqlHelper.ExecuteScalar( 
					connection,
					CommandType.StoredProcedure,
					"CountByApplication", 
					new SqlParameter[] { applicationName } );
					
				return (int)itemCount;
			}
		}

		/// <summary>
		/// Empty constructor called during configuration at startup
		/// </summary>
		public MSDEQueueStorageProvider() : base()
		{
		}

		/// <summary>
		/// Constructor for instance. Used for testing purposes and should not be called by the application
		/// </summary>
		/// <param name="applicationName">Name to use for application</param>
		/// <param name="connectionString">Connection string to use</param>
		/// <param name="isEncrypted">Whether or not data should be encrypted</param>
		/// <param name="isSigned">Whether or not data should be signed</param>
		public MSDEQueueStorageProvider(string applicationName, string connectionString, bool isEncrypted, bool isSigned) : base(isEncrypted, isSigned)
		{
			this.applicationNameLabel = applicationName;
			this.connectionString = connectionString;
			
			Initialize();
		}

		/// <summary>
		/// Flushes all elements out of MSDE
		/// </summary>
		public override void Flush()
		{
			SqlHelper.ExecuteNonQuery(connection, CommandType.StoredProcedure, "DeleteByApplication",
				new SqlParameter[] { applicationName } );
		}

		/// <summary>
		/// Does any initialization needed for MsdeQueueStorageProvider
		/// </summary>
		override protected void Initialize()
		{
			connection = new SqlConnection(connectionString);
			this.applicationName = new SqlParameter("@application", applicationNameLabel);
		}

		/// <summary>
		/// Does any local setup needed to get data out of configuration XML
		/// </summary>
		/// <param name="parameters">Pointer to node in XML</param>
		protected override void DoLocalSetup(XPathNavigator parameters)
		{
			XPathNodeIterator iter = parameters.Select("applicationName");
			iter.MoveNext();
			applicationNameLabel = iter.Current.Value;
			ValidateConfigurationCustomValue("applicationName", applicationNameLabel, parameters);

			iter = parameters.Select("connectionString");
			iter.MoveNext();
			connectionString = iter.Current.Value;
			ValidateConfigurationCustomValue("connectionString", connectionString, parameters);
		}

		/// <summary>
		/// Responsible for storing array of bytes into MSDE
		/// </summary>
		/// <param name="itemBytes">Bytes to store into MSDE</param>
		override protected void DoEnqueue(byte [] itemBytes)
		{
			SqlParameter[] commandParameters = new SqlParameter[2];
			commandParameters[0] = new SqlParameter("@queueItem", itemBytes);
			commandParameters[1] = applicationName;
			SqlHelper.ExecuteNonQuery(connection, CommandType.StoredProcedure, "Enqueue", commandParameters);
		}

		/// <summary>
		/// Responsible for retrieving array of bytes from MSDE
		/// </summary>
		/// <returns>Byte array retrieved from MSDE</returns>
		override protected byte[] DoDequeue()
		{
			object dbObject = SqlHelper.ExecuteScalar(connection, CommandType.StoredProcedure, "Dequeue",
				new SqlParameter[] { applicationName });
			SqlHelper.ExecuteNonQuery(connection, CommandType.StoredProcedure, "DeleteItemsMarkedAsDeleted", new SqlParameter[0]);

			return dbObject as byte[];
		}
	}
}
