//=============================================================================
// Copyright  2004 Microsoft Corporation
// All rights reserved.
// THIS CODE AND INFORMATION IS PROVIDED "AS IS" WITHOUT WARRANTY
// OF ANY KIND, EITHER EXPRESSED OR IMPLIED, INCLUDING BUT NOT
// LIMITED TO THE IMPLIED WARRANTIES OF MERCHANTABILITY AND/OR
// FITNESS FOR A PARTICULAR PURPOSE.
//=============================================================================
using System;
using System.Messaging;
using System.IO;
using System.Runtime.Serialization.Formatters.Binary;
using System.Xml;
using System.Xml.XPath;
using System.Configuration;
using Microsoft.ApplicationBlocks.Common;

namespace Microsoft.ApplicationBlocks.SmartClient.Offline
{
	/// <summary>
	/// Storage provider for QueueManager that enables queue messages to be stored in MSMQ.
	/// </summary>
	public class MSMQQueueStorageProvider : QueueStorageProvider
	{
		private MessageQueue queue = null;
		private static readonly string PrivateQueuePrefix = @".\Private$\";
		private string queueFullName;

		/// <summary>
		/// Getter property to retrieve the number of elements stored in MSDE
		/// </summary>
		public override int Size
		{
			get
			{
				return queue.GetAllMessages().Length;
			}
		}

		/// <summary>
		/// Getter property to retrieve name of MSMQ queue
		/// </summary>
		public string QueueName { get { return queueFullName; }}

		/// <summary>
		/// Empty constructor called during configuration at startup
		/// </summary>
		public MSMQQueueStorageProvider()
		{
		}

		/// <summary>
		/// Constructor used for testing. Application code should not call this method
		/// </summary>
		/// <param name="queuePath">Base name of MSMQ to use. Does not contain the private$ portion</param>
		public MSMQQueueStorageProvider(string queuePath)
		{
			queueFullName = PrivateQueuePrefix + queuePath;
			Initialize();
		}

		/// <summary>
		/// Responsible for removing all messages from MSMQ
		/// </summary>
		public override void Flush()
		{
			queue.Purge();
		}

		/// <summary>
		/// Does any local setup needed to get data out of configuration XML
		/// </summary>
		/// <param name="parameters">Pointer to node in XML</param>
		override protected void DoLocalSetup(XPathNavigator parameters)
		{
			XPathNodeIterator iter = parameters.Select("queueName");
			iter.MoveNext();
			string queueName = iter.Current.Value;
			queueFullName = PrivateQueuePrefix + queueName;
			ValidateConfigurationCustomValue("queueName", queueName, parameters);
		}

		/// <summary>
		/// Does any initialization needed for MSMQQueueStorageProvider
		/// </summary>
		override protected void Initialize()
		{
			if(MessageQueue.Exists(queueFullName))
			{
				queue = new MessageQueue(queueFullName);
			}
			else
			{
				queue = MessageQueue.Create(queueFullName);
			}

			BinaryMessageFormatter formatter = new BinaryMessageFormatter();
			queue.Formatter = formatter;
		}

		/// <summary>
		/// Responsible for storing array of bytes into MSMQ
		/// </summary>
		/// <param name="objectToStream">Object as a byte array to store into MSMQ</param>
		override protected void DoEnqueue(byte[] objectToStream)
		{
			queue.Send(objectToStream);
		}

		/// <summary>
		/// Responsible for retrieving array of bytes from MSMQ
		/// </summary>
		/// <returns>Byte array retrieved from MSMQ</returns>
		override protected byte[] DoDequeue()
		{
			Message dequeuedMessage = queue.Receive();
			return dequeuedMessage.Body as byte[];
		}
	}
}
