//=============================================================================
// Copyright  2004 Microsoft Corporation
// All rights reserved.
// THIS CODE AND INFORMATION IS PROVIDED "AS IS" WITHOUT WARRANTY
// OF ANY KIND, EITHER EXPRESSED OR IMPLIED, INCLUDING BUT NOT
// LIMITED TO THE IMPLIED WARRANTIES OF MERCHANTABILITY AND/OR
// FITNESS FOR A PARTICULAR PURPOSE.
//=============================================================================
using System;
using System.IO;
using System.Xml;
using System.Xml.XPath;
using System.Configuration;
using System.Runtime.Serialization.Formatters.Binary;
using Microsoft.ApplicationBlocks.Common;
using Microsoft.ApplicationBlocks.Common.Crypto;

namespace Microsoft.ApplicationBlocks.SmartClient.Offline
{
	/// <summary>
	/// Base class for all QueueStorageProvider implementations. This classes uses the template method pattern to 
	/// encapsulate the algorithms for enqueuing and dequeuing data to and from the derived provider, only leaving the 
	/// parts that touch the underlying storage mechanism for the subclasses to implement.
	/// </summary>
	public abstract class QueueStorageProvider : IQueueStorageProvider, IProvider
	{
		private bool isEncrypted;
		private bool isSigned;
		private string cryptoProviderName;
		private ICryptographicProvider cryptoProvider;

		/// <summary>
		/// Getter property to determine if the data is being encrypted
		/// </summary>
		/// <value>Stores whether the data in the queue is encrypted or not</value>
		public bool IsEncrypted { get { return isEncrypted; }}

		/// <summary>
		/// Getter property to determine if the data is being signed
		/// </summary>
		/// <value>Stores whether the data in the queue is signed or not</value>
		public bool IsSigned { get { return isSigned; }}

		/// <summary>
		/// Should return number of elements present in underlying storage mechanism
		/// </summary>
		/// <value>Stores the number of elements in the queue</value>
		public abstract int Size { get; }
		
		/// <summary>
		/// Constructor called by derived classes
		/// </summary>
		protected QueueStorageProvider()
		{
		}

		/// <summary>
		/// Constructor called by derived class constructors for testing purposes
		/// </summary>
		/// <param name="isEncrypted">Whether or not data should be encrypted</param>
		/// <param name="isSigned">Whether or not data should be signed</param>
		protected QueueStorageProvider(bool isEncrypted, bool isSigned)
		{
			this.isEncrypted = isEncrypted;
			this.isSigned = isSigned;
			this.cryptoProviderName = "default";
			
			if(isEncrypted || isSigned) 
			{
				cryptoProvider = CryptographicProviderManager.Current.Create(this.cryptoProviderName);
			}
		}

		/// <summary>
		/// IProvider.Initialize implementation to read all common configuration parameters from Xml
		/// </summary>
		/// <param name="parameters">Xml node to search for configuration information</param>
		public void Initialize(XPathNavigator parameters)
		{
			isEncrypted = GetBooleanAttribute(parameters, "encrypted");
			isSigned = GetBooleanAttribute(parameters, "signed");

			if(isEncrypted || isSigned) 
			{
				cryptoProviderName = parameters.GetAttribute("cryptoProvider", string.Empty);
				ValidateConfigurationParameter("cryptoProvider", cryptoProviderName, parameters);
				cryptoProvider = CryptographicProviderManager.Current.Create(cryptoProviderName);
			}

			DoLocalSetup(parameters);
			Initialize();
		}

		/// <summary>
		/// The algorithm for enqueuing data, independent of underlying storage mechanism
		/// </summary>
		/// <param name="queueMessage">Message to be enqueued</param>
		public void Enqueue(QueueMessage queueMessage)
		{
			byte [] msgBytes = StreamingUtils.ToBytes(queueMessage);

			if(isSigned)
			{
				byte [] signature = cryptoProvider.ComputeHash(msgBytes);
				SignedMessage msg = new SignedMessage(signature, msgBytes);
				msgBytes = StreamingUtils.ToBytes(msg);
			}

			if(isEncrypted)
			{
				msgBytes = cryptoProvider.Encrypt(msgBytes);
			}

			DoEnqueue(msgBytes);
		}

		/// <summary>
		/// The algorithm for dequeuing data, independnet of underlying storage mechanism
		/// </summary>
		/// <returns>Message retrieved from storage mechanism</returns>
		public QueueMessage Dequeue()
		{
			byte [] dequeuedBytes = DoDequeue();

			if(isEncrypted)
			{
				dequeuedBytes = cryptoProvider.Decrypt(dequeuedBytes);
			}

			if(isSigned)
			{
				SignedMessage msg = (SignedMessage)StreamingUtils.ToObject(dequeuedBytes);
				
				dequeuedBytes = msg.MessageData;
				byte [] testSignature = cryptoProvider.ComputeHash(dequeuedBytes);

				if(msg.DoVerify(testSignature) == false)
				{
					throw new Exception(ProvidersResourceTable.GetString(ProvidersResourceTable.QueueMessageFailedVerification));
				}
			}

			return StreamingUtils.ToObject(dequeuedBytes) as QueueMessage;
		}

		/// <summary>
		/// Should remove all elements from underlying storage mechanism
		/// </summary>
		public abstract void Flush();
		
		/// <summary>
		/// Validates that the given attribute exists in the configuration node
		/// </summary>
		/// <param name="parameterName">Name of parameter</param>
		/// <param name="parameterValue">Value of parameter</param>
		/// <param name="parameterNode">Xml node to search for parameter</param>
		protected void ValidateConfigurationParameter(string parameterName, string parameterValue, XPathNavigator parameterNode)
		{
			if(parameterValue.Length == 0)
			{
				XmlNode offendingNode = ((IHasXmlNode)parameterNode).GetNode();
				throw new ConfigurationException(String.Format(ProvidersResourceTable.GetString(ProvidersResourceTable.InvalidOrMissingAttributeInConfigFile), parameterName), offendingNode);
			}
		}		

		/// <summary>
		/// Validates that the given elemetn exists in the configuration node
		/// </summary>
		/// <param name="customElementName">Name of element</param>
		/// <param name="customValue">Value of element</param>
		/// <param name="parameterNode">Xml node to search for element</param>
		protected void ValidateConfigurationCustomValue(string customElementName, string customValue, XPathNavigator parameterNode)
		{
			if(customValue.Length == 0)
			{
				XmlNode offendingNode = ((IHasXmlNode)parameterNode).GetNode();
				throw new ConfigurationException(String.Format(ProvidersResourceTable.GetString(ProvidersResourceTable.InvalidOrMissingElementInConfigFile), customElementName), offendingNode);
			}
		}	
		
		/// <summary>
		/// Should store a byte array to underyling storage mechanism
		/// </summary>
		/// <param name="objectToStream">Byte array to store</param>
		protected abstract void DoEnqueue(byte[] objectToStream);

		/// <summary>
		/// Should retrieve a byte array from underlying storage mechanism
		/// </summary>
		/// <returns>Byte array retrieved from underlying storage mechanism</returns>
		protected abstract byte[] DoDequeue();
		
		/// <summary>
		/// Should read any custom parameters out of configuration information needed by subclass
		/// </summary>
		/// <param name="parameters">Xml node to search for configution information</param>
		protected abstract void DoLocalSetup(XPathNavigator parameters);

		/// <summary>
		/// Should do any custom initialization needed by subclass
		/// </summary>
		protected abstract void Initialize();

		/// <summary>
		/// Retrieves a boolean attribute from configuration information
		/// </summary>
		/// <param name="parameters"></param>
		/// <param name="parameterName"></param>
		/// <returns></returns>
		private bool GetBooleanAttribute(XPathNavigator parameters, string parameterName)
		{
			string parameterStringValue = parameters.GetAttribute(parameterName, string.Empty);
			ValidateConfigurationParameter(parameterName, parameterStringValue, parameters);

			try
			{
				return bool.Parse(parameterStringValue);
			}
			catch(FormatException e)
			{
				throw new ConfigurationException(String.Format(ProvidersResourceTable.GetString(ProvidersResourceTable.InvalidFieldFormatInConfigFile), parameterName), e);
			}
		}
	}
}
