//=============================================================================
// Copyright  2004 Microsoft Corporation
// All rights reserved.
// THIS CODE AND INFORMATION IS PROVIDED "AS IS" WITHOUT WARRANTY
// OF ANY KIND, EITHER EXPRESSED OR IMPLIED, INCLUDING BUT NOT
// LIMITED TO THE IMPLIED WARRANTIES OF MERCHANTABILITY AND/OR
// FITNESS FOR A PARTICULAR PURPOSE.
//=============================================================================
using System;

namespace Microsoft.ApplicationBlocks.SmartClient.Offline
{
	/// <summary>
	/// This class represents a signed message in the QueueManager message queue. It holds the original data, and a
	/// cryptographic hash of that data that can be later verified. 
	/// </summary>
	[Serializable()]
	public struct SignedMessage
	{
		private byte[] signature;
		private byte[] messageData;

		/// <summary>
		/// Returns the signature of the data
		/// </summary>
		public byte[] Signature { get { return signature; }}

		/// <summary>
		/// Returns the message data
		/// </summary>
		public byte[] MessageData { get { return messageData; }}

		/// <summary>
		/// Constructor for instance
		/// </summary>
		/// <param name="signature">Signature of data</param>
		/// <param name="messageData">Message data</param>
		public SignedMessage(byte[] signature, byte[] messageData)
		{
			this.signature = signature;
			this.messageData = messageData;
		}

		/// <summary>
		/// Verifies the signature with the given test signature
		/// </summary>
		/// <param name="testSignature">Signature to compare to</param>
		/// <returns>True if the signatures match</returns>
		public bool DoVerify(byte[] testSignature)
		{
			if(AreBothSignaturesNull(testSignature, signature)) return true;
			if(IsOnlyOneSignatureIsNull(testSignature, signature)) return false;
			if(signature.Length != testSignature.Length) return false;

			for(int i = 0; i < signature.Length; i++)
			{
				if(signature[i] != testSignature[i]) return false;
			}

			return true;
		}

		private bool AreBothSignaturesNull(byte [] firstSignature, byte [] secondSignature)
		{
			return firstSignature == null && secondSignature == null;
		}

		private bool IsOnlyOneSignatureIsNull(byte [] firstSignature, byte [] secondSignature)
		{
			return (firstSignature == null && secondSignature != null) || (secondSignature == null && firstSignature != null);
		}
	}
}
