//=============================================================================
// Copyright  2004 Microsoft Corporation
// All rights reserved.
// THIS CODE AND INFORMATION IS PROVIDED "AS IS" WITHOUT WARRANTY
// OF ANY KIND, EITHER EXPRESSED OR IMPLIED, INCLUDING BUT NOT
// LIMITED TO THE IMPLIED WARRANTIES OF MERCHANTABILITY AND/OR
// FITNESS FOR A PARTICULAR PURPOSE.
//=============================================================================
using System;

namespace Microsoft.ApplicationBlocks.SmartClient.Offline
{
	/// <summary>
	/// The one and only singleton in the system. OfflineBlockBuilder builds all the pieces of the system
	/// and makes available references to whatever is needed
	/// </summary>
	public sealed class OfflineBlockBuilder : IDisposable
	{
		private ReferenceDataCacheBuilder referenceDataCacheBuilder;
		private ConnectionManagerBuilder connectionManagerBuilder;
		private DataLoaderManagerBuilder dataLoaderManagerBuilder;
		private QueueManagerBuilder queueManagerBuilder;
		private ServiceAgentManagerBuilder serviceAgentManagerBuilder;
		private ExecutorBuilder executorBuilder;

		private static OfflineBlockBuilder offlineBlockBuilder;

		/// <summary>
		/// Gets the Singleton BlockBuilder instance.
		/// </summary>
		/// <value>Singleton instance of the Block Builder class</value>
		public static OfflineBlockBuilder Instance
		{
			get
			{
				if(offlineBlockBuilder == null)
				{
					offlineBlockBuilder = new OfflineBlockBuilder();
				}

				return offlineBlockBuilder;
			}
		}

		/// <summary>
		/// Getter to determine if the block is currently created and ready for use, or if it has been disposed.
		/// </summary>
		/// <returns>true if block is not initialized</returns>
		public static bool IsDisposed() { return offlineBlockBuilder == null; }

		/// <summary>
		/// Private constructor for BlockBuilder. Knows how to assemble the block from its pieces
		/// </summary>
		private OfflineBlockBuilder()
		{
			referenceDataCacheBuilder = new ReferenceDataCacheBuilder();
			connectionManagerBuilder = new ConnectionManagerBuilder();
			queueManagerBuilder = new QueueManagerBuilder();
			serviceAgentManagerBuilder = new ServiceAgentManagerBuilder();
				                               
			dataLoaderManagerBuilder = 
				new DataLoaderManagerBuilder(referenceDataCacheBuilder.ReferenceDataCache,
				                             queueManagerBuilder.PayloadConsumer);
			executorBuilder = new ExecutorBuilder(serviceAgentManagerBuilder.ResultConsumer, 
				                                  queueManagerBuilder.QueueMessageProducer);

			connectionManagerBuilder.ConnectionManager.ConnectionStateChangedEvent +=
							new ConnectionStateChangedEventHandler(executorBuilder.Executor.OnlineOfflineCallback);
		}

		/// <summary>
		/// Getter property for DataLoaderManager
		/// </summary>
		/// <value>DataLoaderManager object instance</value>
		public DataLoaderManager DataLoaderManager 
		{
			get { return dataLoaderManagerBuilder.DataLoaderManager; }
		}

		/// <summary>
		/// Getter property for ReferenceDataCache
		/// </summary>
		/// <value>ReferenceDataCache object instance</value>
		public IReferenceDataCache ReferenceDataCache
		{
			get { return referenceDataCacheBuilder.ReferenceDataCache; }
		}

		/// <summary>
		///  Getter property for ServiceAgentManager
		/// </summary>
		/// <value>ServiceAgentManager object instance</value>
		public ServiceAgentManager ServiceAgentManager
		{
			get { return serviceAgentManagerBuilder.ServiceAgentManager; }
		}

		/// <summary>
		/// Getter property for ServiceAgentRegistry
		/// </summary>
		/// <value>ServiceAgentRegistry object instance</value>
		internal ServiceAgentRegistry ServiceAgentRegistry
		{
			get { return serviceAgentManagerBuilder.ServiceAgentRegistry; }
		}

		/// <summary>
		/// Getter method for the FailsafeServiceAgent. This agent is the block-provided ServiceAgent of last resort
		/// when returing error messages and results back to the application layer.
		/// </summary>
		/// <value>FailsafeServiceAgent object instance</value>
		public FailsafeServiceAgent FailsafeServiceAgent
		{
			get { return ServiceAgentRegistry.FailsafeServiceAgent; }
		}

		/// <summary>
		///  Getter property for PayloadConsumer. This is the class in the system that will enqueue message Payloads
		///  that are being sent to a remote service.
		/// </summary>
		/// <value>PayloadConsumer object instance</value>
		public IPayloadConsumer PayloadConsumer
		{
			get { return queueManagerBuilder.PayloadConsumer; }
		}

		/// <summary>
		/// Getter property for the ConnectionManager
		/// </summary>
		/// <value>ConnectionManager object instance</value>
		public ConnectionManager ConnectionManager
		{
			get { return connectionManagerBuilder.ConnectionManager; }
		}

		/// <summary>
		/// Will start the internal operations of this block
		/// </summary>
		public void Start()
		{
			connectionManagerBuilder.ConnectionManager.Start();
		}

		/// <summary>
		/// Will stop the internal operations of this block. This includes shutting down all internal threads. This method may block
		/// while internal threads die and are joined.
		/// </summary>
		public void Stop()
		{
			if(connectionManagerBuilder.ConnectionManager != null) 
			{
				connectionManagerBuilder.ConnectionManager.Stop();
				executorBuilder.Executor.Stop();
			}
		}

		/// <summary>
		/// Properly disposes of all resources in this block. Will also reset the singleton instance back to null.
		/// The next call to GetInstance() will return a newly created instance of BlockBuilder
		/// </summary>
		public void Dispose()
		{
			Stop();

			referenceDataCacheBuilder = null;
			connectionManagerBuilder = null;
			dataLoaderManagerBuilder = null;
			queueManagerBuilder = null;
			serviceAgentManagerBuilder = null;
			executorBuilder = null;

			OfflineBlockBuilder.offlineBlockBuilder = null;
			ServiceAgent.SetServiceAgentRegistry(null);
		}
	}
}
