//=============================================================================
// Copyright  2004 Microsoft Corporation
// All rights reserved.
// THIS CODE AND INFORMATION IS PROVIDED "AS IS" WITHOUT WARRANTY
// OF ANY KIND, EITHER EXPRESSED OR IMPLIED, INCLUDING BUT NOT
// LIMITED TO THE IMPLIED WARRANTIES OF MERCHANTABILITY AND/OR
// FITNESS FOR A PARTICULAR PURPOSE.
//=============================================================================
using System;
using System.Threading;
using System.Runtime.CompilerServices;
using Microsoft.ApplicationBlocks.SmartClient.Offline;

namespace Microsoft.ApplicationBlocks.SmartClient.Offline
{
	/// <summary>
	/// This class is the facade over the ConnectionManager subsystem. It is responsible
	/// for ensuring that the subsystem is configured correctly, based on the configuration
	/// file, managing the thread involved with polling the connection state, and handling
	/// the event notifications as the connection state changes. It is assumed that all methods of 
	/// this class will be called from a single thread, so no attempt was made at making its public
	/// methods thread-safe.
	/// </summary>
	public class ConnectionManager
	{
		private const int ConversionToMilliSeconds = 1000;

		private ConnectionDetector detector;
		private ThreadRunner thread = null;

		private bool forcedOffline = false;
		private int pollInterval;

		/// <summary>
		/// Event raised when state of connection changes.
		/// </summary>
		public event ConnectionStateChangedEventHandler ConnectionStateChangedEvent;

		/// <summary>
		/// Constructor for ConnectionManager.
		/// </summary>
		/// <param name="detectionStrategy">The ConnectionDetectionStrategy chosen from the config file</param>
		/// <param name="pollingIntervalInSeconds">Interval in seconds at which the ConnectionDetectionStrategy will be polled</param>
		public ConnectionManager(IConnectionDetectionStrategy detectionStrategy, int pollingIntervalInSeconds)
		{
			this.detector = new ConnectionDetector(detectionStrategy, this);
			this.pollInterval = pollingIntervalInSeconds * ConversionToMilliSeconds;
			thread = new ThreadRunner(new ThreadStart(WorkerMethod));
		}
	
		/// <summary>
		/// Starts the polling of the ConnectionDetector
		/// </summary>
		[MethodImpl(MethodImplOptions.Synchronized)]
		public void Start()
		{
			if(forcedOffline == false)
			{
				thread.Start();
			}
		}

		/// <summary>
		/// Stops polling of the ConnectionDetector. This method may block momentarily as internal threads die and are joined.
		/// </summary>
		[MethodImpl(MethodImplOptions.Synchronized)]
		public void Stop()
		{
			thread.Stop();
		}

		/// <summary>
		/// Forces the system offline
		/// </summary>
		[MethodImpl(MethodImplOptions.Synchronized)]
		public void GoOffline()
		{
			forcedOffline = true;
			Stop();
			detector.ForceOffline();
		}

		/// <summary>
		/// Forces the system online. 
		/// </summary>
		/// <exception cref="ConnectionUnavailableException">Thrown if ConnectionDetector.CanGoOnline() would return false</exception>
		[MethodImpl(MethodImplOptions.Synchronized)]
		public void GoOnline()
		{
			ThrowExceptionIfWeCannotGoOnline();

			forcedOffline = false;
			detector.ForceOnline();
			Start();
		}

		/// <summary>
		/// Returns true if we are currently polling the ConnectionDetector
		/// </summary>
		/// <returns>Returns true if we are currently polling the ConnectionDetector</returns>
		[MethodImpl(MethodImplOptions.Synchronized)]
		public bool IsPolling()
		{
			return thread.IsRunning;
		}

		/// <summary>
		/// Returns the connected state of the physical connection
		/// </summary>
		/// <value>Stores the enumerated value of the Connection State</value>
		public ConnectionState ConnectionState
		{
			[MethodImpl(MethodImplOptions.Synchronized)]
			get {return this.detector.ConnectedState;}
		}

		internal void RaiseConnectionStateChangedEvent(ConnectionStateChangedEventArgs e)
		{
			EventUtils.BroadcastEvent(ConnectionStateChangedEvent, this, e);
		}

		/// <summary>
		/// Encapulates check for whether we can go online
		/// </summary>
		private void ThrowExceptionIfWeCannotGoOnline()
		{
			if(detector.CanGoOnline() == false)
			{
				throw new ConnectionUnavailableException(ResourceTable.GetString(ResourceTable.ConnectionUnavailableMessage));
			}
		}

		/// <summary>
		/// Method that is invoked on worker thread to actually poll the ConnectionDetector
		/// </summary>
		private void WorkerMethod()
		{	
			detector.DetectConnectionState();
			Thread.Sleep(pollInterval);
		}
	}
}
