//=============================================================================
// Copyright  2004 Microsoft Corporation
// All rights reserved.
// THIS CODE AND INFORMATION IS PROVIDED "AS IS" WITHOUT WARRANTY
// OF ANY KIND, EITHER EXPRESSED OR IMPLIED, INCLUDING BUT NOT
// LIMITED TO THE IMPLIED WARRANTIES OF MERCHANTABILITY AND/OR
// FITNESS FOR A PARTICULAR PURPOSE.
//=============================================================================
using System;
using System.Collections;

namespace Microsoft.ApplicationBlocks.SmartClient.Offline
{
	/// <summary>
	/// The DataLoaderManager is used to request data to be loaded or refreshed into the ReferenceDataCache
	/// </summary>
	public class DataLoaderManager
	{
		private IReferenceDataCache referenceCache;
		private IPayloadConsumer queueManager;

		/// <summary>
		/// constructor for DataLoaderManager
		/// </summary>
		/// <param name="referenceDataCache">The reference data cache</param>
		/// <param name="queueManager">The object in the system responsible for enqueuing message payloads. 
		/// This is usually the QueueManager.</param>
		public DataLoaderManager(IReferenceDataCache referenceDataCache, IPayloadConsumer queueManager)
		{
			referenceCache = referenceDataCache;
			this.queueManager = queueManager;
		}

		/// <summary>
		/// Load data into the reference data cache. Results are returned through the block-provided ReferenceDataUpdaterServiceAgent
		/// </summary>
		/// <param name="dataDefinition">Definition of the reference data to be loaded into the cache</param>
		/// <param name="serviceAgentGuid">GUID of Service Agent to which completion notifications returned</param>
		public void LoadData(ReferenceDataDefinition dataDefinition, Guid serviceAgentGuid)
		{
			ReferenceCacheDataPayload payload = new ReferenceCacheDataPayload(dataDefinition.RefreshMethod, serviceAgentGuid, dataDefinition);
			queueManager.Enqueue(payload);
		}

		/// <summary>
		/// Loads data into the reference data cache. Results are returned through the user-provided ServiceAgentContext
		/// </summary>
		/// <param name="serviceAgentGuid">GUID of Service Agent to which completion notifications returned</param>
		/// <param name="serviceAgentGettingReturnedData">ServiceAgentContext defining which service agent to call to return results back to 
		/// original caller</param>
		/// <param name="dataDefinition">Definition of the reference data to be loaded into the cache</param>
		public void LoadData(Guid serviceAgentGuid, ServiceAgentContext serviceAgentGettingReturnedData, ReferenceDataDefinition dataDefinition)
		{
			ReferenceCacheDataPayload payload = 
				new ReferenceCacheDataPayload(dataDefinition.RefreshMethod, serviceAgentGuid, serviceAgentGettingReturnedData, dataDefinition);
			queueManager.Enqueue(payload);
		}

		/// <summary>
		/// Causes a message to be queued to refresh expired data in the reference data cache
		/// </summary>
		/// <param name="serviceAgentGuid">GUID of Service Agent to which completion notifications returned</param>
		/// <param name="dataDefinitionRetrievedFromReferenceCache">Definition of reference data to be refreshed. This is called
		/// internally by the block.</param>
		public void RefreshData(Guid serviceAgentGuid, ReferenceDataDefinition dataDefinitionRetrievedFromReferenceCache)
		{
			ReferenceCacheDataPayload payload =
				new ReferenceCacheDataPayload(dataDefinitionRetrievedFromReferenceCache.RefreshMethod, 
											  serviceAgentGuid,
											  dataDefinitionRetrievedFromReferenceCache, 
											  true);
			queueManager.Enqueue(payload);
		}
	}
}
