//=============================================================================
// Copyright  2004 Microsoft Corporation
// All rights reserved.
// THIS CODE AND INFORMATION IS PROVIDED "AS IS" WITHOUT WARRANTY
// OF ANY KIND, EITHER EXPRESSED OR IMPLIED, INCLUDING BUT NOT
// LIMITED TO THE IMPLIED WARRANTIES OF MERCHANTABILITY AND/OR
// FITNESS FOR A PARTICULAR PURPOSE.
//=============================================================================
using System;
using System.Threading;

namespace Microsoft.ApplicationBlocks.SmartClient.Offline
{
	/// <summary>
	/// This class implements the threading policy for the SimpleCommandProcessor component of the QueueManager
	/// </summary>
	public class Executor
	{
		private IQueueMessageProducer messageProvider;
		private ICommandProcessor executor;
		private ThreadRunner managedThread;
		
		private static readonly int QueuePollDelayInMilliseconds = 1000;

		/// <summary>
		/// Constructor for class.
		/// </summary>
		/// <param name="messageProvider">Source of QueueMessage objects processed through SimpleCommandProcessor instance. Cannot be null.</param>
		/// <param name="executor">SimpleCommandProcessor instance to be invoked on thread. Cannot be null.</param>
		public Executor(IQueueMessageProducer messageProvider, ICommandProcessor executor)
		{
			this.messageProvider = messageProvider;
			this.executor = executor;
			this.managedThread = new ThreadRunner(new ThreadStart(WorkerThread));
		}

		/// <summary>
		/// Causes the executor to start pulling available messages off the IQueueMessageProducer's message
		/// queue and processing them.
		/// </summary>
		public void Start()
		{
			managedThread.Start();
		}

		/// <summary>
		/// Causes the executor to stop processing incoming QueueMessages. This method may momentarily block while waiting for
		/// the embedded thread to die and be joined
		/// </summary>
		public void Stop()
		{
			managedThread.Stop();
		}

		/// <summary>
		/// Returns true if the SimpleCommandProcessor is currently processing messages
		/// </summary>
		/// <value>Stores the current state of the executor</value>
		public bool IsRunning { get { return managedThread.IsRunning; }}

		/// <summary>
		/// This is the primary worker thread for the SimpleCommandProcessor. It pulls messages off the IQueueMessageProducer's
		/// message queue and hands them to the SimpleCommandProcessor.
		/// </summary>
		private void WorkerThread()
		{
			if(AreMessagesAvailable() == false) 
			{
				Thread.Sleep(QueuePollDelayInMilliseconds);
				return;
			}
            
			QueueMessage message = messageProvider.Dequeue();
			executor.ExecuteMessage(message);
		}

		/// <summary>
		/// Callback method called when the online/offline state of the application changes. It is public to allow for testing. 
		/// This method should not be called by client code.
		/// </summary>
		/// <param name="sender">Raiser of the event</param>
		/// <param name="e">Connection state information</param>
		public void OnlineOfflineCallback(object sender, ConnectionStateChangedEventArgs e)
		{
			if(e.CurrentState == ConnectionState.OffLine) Stop();
			else if(e.CurrentState == ConnectionState.OnLine) Start();
		}

		private bool AreMessagesAvailable()
		{
			return messageProvider.Size != 0;
		}
	}
}
