//=============================================================================
// Copyright  2004 Microsoft Corporation
// All rights reserved.
// THIS CODE AND INFORMATION IS PROVIDED "AS IS" WITHOUT WARRANTY
// OF ANY KIND, EITHER EXPRESSED OR IMPLIED, INCLUDING BUT NOT
// LIMITED TO THE IMPLIED WARRANTIES OF MERCHANTABILITY AND/OR
// FITNESS FOR A PARTICULAR PURPOSE.
//=============================================================================
using System;
using System.Reflection;
using System.Security.Permissions;
using Microsoft.ApplicationBlocks.ExceptionManagement;

namespace Microsoft.ApplicationBlocks.SmartClient.Offline
{
	/// <summary>
	/// This class encapsulates the information necessary to invoke a method through
	/// reflection. It is reasonably important that the class defined by an instance of OnlineProxyContext
	/// be stateless. If it is not stateless, then the instance created through reflection may have to instantiate a 
	/// large part of the system to retrieve the information it needs to operate correctly.
	/// </summary>
	/// 
	[Serializable()]
	public class OnlineProxyContext
	{
		private string assemblyName;
		private string className;
		private string methodName;

		/// <summary>
		/// Constructor for instance
		/// </summary>
		/// <param name="assemblyName">Name of the assembly where the method to execute resides</param>
		/// <param name="className">Fully qualified name of the class where the method to execute resides. This 
		/// name should include the namespace of the class.</param>
		/// <param name="methodName">Name of the method to execute</param>
		public OnlineProxyContext(string assemblyName, string className, string methodName)
		{
			this.assemblyName = assemblyName;
			this.className = className;
			this.methodName = methodName;
		}

		/// <summary>
		/// Invokes the method through reflection defined by this instance. Assuming the calling code has the permissions necessary
		/// to invoke methods through reflection, no exceptions will leak out of this method.
		/// </summary>
		/// <param name="dataToSend">Payload to execute</param>
		/// <exception cref="SecurityException">Caller did not have permission to execute methods through reflection</exception>
		[ReflectionPermission(SecurityAction.Demand, Flags=ReflectionPermissionFlag.MemberAccess)]
		public void InvokeMethodThroughReflection(Payload dataToSend)
		{
			try
			{
				object createdObject = Activator.CreateInstance(assemblyName, className).Unwrap();
				MethodInfo method = createdObject.GetType().GetMethod(methodName, BindingFlags.Instance | BindingFlags.Public | BindingFlags.NonPublic);
				method.Invoke(createdObject, new object[] { dataToSend });
			}
			catch(Exception e)
			{
				InvalidOnlineProxyContextDefinitionException exception = new InvalidOnlineProxyContextDefinitionException(GenerateExceptionMessage(), e);
				ExceptionManager.Publish(exception);

				dataToSend.RecordFailure(exception);
			}
		}

		private string GenerateExceptionMessage()
		{
			if( assemblyName != null && className != null && methodName != null )
			{
				return String.Format(ResourceTable.GetString(ResourceTable.UnableToInvokeMethodThruReflection), assemblyName, className, methodName);
			}
			else
			{
				return ResourceTable.GetString(ResourceTable.UnableToInvokeMethodThruReflectionPlainMessage);
			}
		}
	}
}
