//=============================================================================
// Copyright  2004 Microsoft Corporation
// All rights reserved.
// THIS CODE AND INFORMATION IS PROVIDED "AS IS" WITHOUT WARRANTY
// OF ANY KIND, EITHER EXPRESSED OR IMPLIED, INCLUDING BUT NOT
// LIMITED TO THE IMPLIED WARRANTIES OF MERCHANTABILITY AND/OR
// FITNESS FOR A PARTICULAR PURPOSE.
//=============================================================================
using System;
using System.Runtime.CompilerServices;

namespace Microsoft.ApplicationBlocks.SmartClient.Offline
{
	/// <summary>
	/// This is the facade over the QueueManager subsystem. This subsystem is responsible
	/// for queueing requests to web services while an application is offline. All instance methods of this
	/// class are threadsafe. This class is passive, in that it does not run on its own thread, doing anything
	/// actively. It is used by the application in its thread to store messages, and used by the SimpleCommandProcessor to pull
	/// messages out in its thread.
	/// </summary>
	public class QueueManager : IQueueMessageProducer, IPayloadConsumer
	{
		private IQueueStorageProvider queueStorage;

		/// <summary>
		/// Getter property to return number of messages currently queued
		/// </summary>
		/// <value>Number of messages currently queued</value>
		public int Size 
		{ 
			[MethodImpl(MethodImplOptions.Synchronized)]
			get { return queueStorage.Size;	}
		}

		/// <summary>
		/// Constructor for QueueManager
		/// </summary>
		/// <param name="queueStorage">QueueStorageProvider to use to store QueueMessages</param>
		public QueueManager(IQueueStorageProvider queueStorage)
		{
			this.queueStorage = queueStorage;
		}

		/// <summary>
		/// Enqueues a message onto the QuueStorageProvider
		/// </summary>
		/// <param name="messageToEnqueue">Message to enqueue</param>
		[MethodImpl(MethodImplOptions.Synchronized)]
		public void Enqueue(Payload messageToEnqueue)
		{
				queueStorage.Enqueue(new QueueMessage(messageToEnqueue));
		}

		/// <summary>
		/// Dequeues a message from the QueueMessageProvider
		/// </summary>
		/// <returns>Message that was dequeued</returns>
		[MethodImpl(MethodImplOptions.Synchronized)]
		public QueueMessage Dequeue()
		{
				return queueStorage.Dequeue();
		}

		/// <summary>
		/// Flushes all QueueMessages out of the QueueStorageProvider
		/// </summary>
		[MethodImpl(MethodImplOptions.Synchronized)]
		public void Flush()
		{
				queueStorage.Flush();
		}
	}
}
