//=============================================================================
// Copyright  2004 Microsoft Corporation
// All rights reserved.
// THIS CODE AND INFORMATION IS PROVIDED "AS IS" WITHOUT WARRANTY
// OF ANY KIND, EITHER EXPRESSED OR IMPLIED, INCLUDING BUT NOT
// LIMITED TO THE IMPLIED WARRANTIES OF MERCHANTABILITY AND/OR
// FITNESS FOR A PARTICULAR PURPOSE.
//=============================================================================
using System;
using System.Collections;

namespace Microsoft.ApplicationBlocks.SmartClient.Offline
{
	/// <summary>
	/// This interface defines how the offline block will interact with the portion of
	/// the CacheManager where we will be storing reference data.
	/// </summary>
	public interface IReferenceDataCache
	{
		/// <summary>
		/// Method stores referenceDataValue in reference area of CacheManager while inserting and managing
		/// all metadata associated with the new cached value.
		/// </summary>
		/// <param name="referenceDataDefinition">Metadata about the value being cached</param>
		/// <param name="referenceDataValue">Object to be stored. Must be serializable</param>
		void Store(ReferenceDataDefinition referenceDataDefinition, object referenceDataValue);

		/// <summary>
		/// Returns the object stored in the reference cache at the given key. 
		/// </summary>
		/// <param name="cacheKey">Key used to search for the value in the cache. Should not be null.</param>
		/// <returns>Object stored in cache, or null if not found</returns>
		object Retrieve(string cacheKey);

		/// <summary>
		/// Will update the object stored at the given cache key with the new value. Note that this
		/// method follows the CacheManager semantics of actually replacing the object specified by the 
		/// key with the new value, it does not update the existing object in place. Existing metadata is updated to 
		/// reflect the newly stored value. It is an error to try to update a value that does not already exist in the 
		/// reference cache.
		/// </summary>
		/// <param name="cacheKey">Key used to search for the value in the cache. Should not be null.</param>
		/// <param name="referenceDataValue">Object stored in cache</param>
		/// <exception cref="ReferenceCacheElementNotFoundException">The requested element was not present in the cache to be updated</exception>
		void Update(string cacheKey, object referenceDataValue);

		/// <summary>
		/// Will remove the given element from the cache. If element does not exist, nothing will happen.
		/// </summary>
		/// <param name="cacheKey">Key of element to remove. Should not be null</param>
		void Remove(string cacheKey);

		/// <summary>
		/// Used to determine if the element referenced by the key has been changed from its original value. 
		/// If it has been changed, it will have beeen marked dirty by the ReferenceDataCache, and will remain
		/// dirty until manually cleaned. It is an error to check the dirty status of an element not in the cache.
		/// </summary>
		/// <param name="cacheKey">Key of element to check.</param>
		/// <returns>True if the element has been changed and its value has not been sent to a remote service</returns>
		/// <exception cref="ReferenceCacheElementNotFoundException">The requested element was not present in the cache to be checked</exception>
		bool IsDirty(string cacheKey);

		/// <summary>
		/// Used to mark a cache element as being clean after it has been changed. It is an error to call this method
		/// on an element not already in the cache
		/// </summary>
		/// <param name="cacheKey">Key of element to set as clean</param>
		/// <exception cref="ReferenceCacheElementNotFoundException">The requested element was not present in the cache to be checked</exception>
		void MarkAsClean(string cacheKey);

		/// <summary>
		/// This method will readd an item back into the reference cache after it has
		/// been expired by the Caching Block.
		/// </summary>
		/// <param name="expiredKey">Name of the key to be readded</param>
		/// <returns>The ReferenceDataDefinition of the item readded to the cache</returns>
		ReferenceDataDefinition ItemHasBeenExpired(string expiredKey);
	}
}
