//=============================================================================
// Copyright  2004 Microsoft Corporation
// All rights reserved.
// THIS CODE AND INFORMATION IS PROVIDED "AS IS" WITHOUT WARRANTY
// OF ANY KIND, EITHER EXPRESSED OR IMPLIED, INCLUDING BUT NOT
// LIMITED TO THE IMPLIED WARRANTIES OF MERCHANTABILITY AND/OR
// FITNESS FOR A PARTICULAR PURPOSE.
//=============================================================================
using System;

namespace Microsoft.ApplicationBlocks.SmartClient.Offline
{
	/// <summary>
	/// This class adds the ReferenceDataDefinition to the Payload class. This is needed to allow the information about
	/// the reference data to accompany the download operation throughout the call chain.
	/// </summary>
	/// 
	[Serializable()]
	public class ReferenceCacheDataPayload : Payload
	{
		private ReferenceDataDefinition dataDefinition;
		private bool isRefreshMessage = false;

		/// <summary>
		/// Getter property to return the definition of the reference data
		/// </summary>
		/// <value>Data definition for the reference data</value>
		public ReferenceDataDefinition DataDefinition { get { return dataDefinition; }}

		/// <summary>
		/// Constructor for ReferenceCacheDataPayload.
		/// </summary>
		/// <param name="proxyMethodToInvoke">Object representing the online proxy method to invoke to retrieve the data</param>
		/// <param name="serviceAgentGuid">Guid of service agent to be returned the success or failure of the request</param>
		/// <param name="dataDefinition">The definition of the data to be stored in the ReferenceDataCache</param>
		public ReferenceCacheDataPayload(OnlineProxyContext proxyMethodToInvoke, Guid serviceAgentGuid, ReferenceDataDefinition dataDefinition)
			: this(proxyMethodToInvoke, serviceAgentGuid, new ServiceAgentContext("StoreDataInCache"), dataDefinition)
		{
		}

		/// <summary>
		/// Constructor for ReferenceCacheDataPayload.
		/// </summary>
		/// <param name="proxyMethodToInvoke">Object representing the online proxy method to invoke to retrieve the data</param>
		/// <param name="serviceAgentGuid">Guid of service agent to be returned the success or failure of the request</param>
		/// <param name="dataDefinition">The definition of the data to be stored in the ReferenceDataCache</param>
		/// <param name="isRefreshMessage">True if this represents a refresh operation for data already loaded into the
		/// ReferenceDataCache.</param>
		internal ReferenceCacheDataPayload(OnlineProxyContext proxyMethodToInvoke, Guid serviceAgentGuid, ReferenceDataDefinition dataDefinition, bool isRefreshMessage)
			: this(proxyMethodToInvoke, serviceAgentGuid, dataDefinition)
		{
			this.isRefreshMessage = isRefreshMessage;
		}

		/// <summary>
		/// Constructor for ReferenceCacheDataPayload
		/// </summary>
		/// <param name="proxyMethodToInvoke">Object representing the online proxy method to invoke to retrieve the data</param>
		/// <param name="serviceAgentGuid">Guid of service agent to be returned the success or failure of the request</param>
		/// <param name="serviceAgentToInvoke">ServiceAgent method to invoke to return results</param>
		/// <param name="dataDefinition">The definition of the data to be stored in the ReferenceDataCache</param>
		public ReferenceCacheDataPayload(OnlineProxyContext proxyMethodToInvoke, Guid serviceAgentGuid, ServiceAgentContext serviceAgentToInvoke, ReferenceDataDefinition dataDefinition)
			: base(proxyMethodToInvoke, serviceAgentGuid, serviceAgentToInvoke, null)
		{
			this.dataDefinition = dataDefinition;
		}

		/// <summary>
		/// Query method to see if this is a refresh message
		/// </summary>
		/// <returns>True if this is Payload was created to refresh cache data</returns>
		public bool IsRefreshMessage() { return isRefreshMessage; }
		
		/// <summary>
		/// This method is used to store the data returned during message execution into the Payload
		/// </summary>
		/// <param name="updatedData">Data to be stored into Payload object</param>
		public void UpdateDataToReturn(object updatedData) 
		{
			this.Results = updatedData;
		}
	}
}
