//=============================================================================
// Copyright  2004 Microsoft Corporation
// All rights reserved.
// THIS CODE AND INFORMATION IS PROVIDED "AS IS" WITHOUT WARRANTY
// OF ANY KIND, EITHER EXPRESSED OR IMPLIED, INCLUDING BUT NOT
// LIMITED TO THE IMPLIED WARRANTIES OF MERCHANTABILITY AND/OR
// FITNESS FOR A PARTICULAR PURPOSE.
//=============================================================================
using System;
using System.Security.Cryptography;
using System.Runtime.Serialization.Formatters.Binary;
using System.Collections;
using System.IO;
using System.Runtime.CompilerServices;

using Microsoft.ApplicationBlocks.Cache;
using Microsoft.ApplicationBlocks.Cache.Expirations;

namespace Microsoft.ApplicationBlocks.SmartClient.Offline
{
	/// <summary>
	/// The ReferenceDataCache provides a wrapper over the CacheManager instance. This wrapper provides block
	/// users access to the needed cache functionality, but only through the metadata that the offline block needed
	/// to manage the reference data. It is not intended that offline block users ever access the CacheManager directly 
	/// to retrieve their reference data, as the ReferenceDataCache is responsible for it. 
	/// </summary>
	public class ReferenceDataCache : IReferenceDataCache
	{
		private CacheManager cache = CacheManager.GetCacheManager();
		private object lockObject = new object();

		/// <summary>
		/// This method will store the given data into the ReferenceDataCache, along with the metadata associated with it.
		/// </summary>
		/// <param name="referenceDataDef">Metadata associated with the data to be stored</param>
		/// <param name="newDataValue">The data to be stored</param>
		/// <exception cref="ArgumentException">Cache key is not valid</exception>
		[MethodImpl(MethodImplOptions.Synchronized)]
		public void Store(ReferenceDataDefinition referenceDataDef, object newDataValue)
		{
			EnsureCacheKeyValid(referenceDataDef);
			InsertDataIntoCache(referenceDataDef, newDataValue);
		}

		/// <summary>
		/// Updates the given cacheKey with the new reference data. This will overwrite the existing data, but does not affect
		/// the metadata. If the data is dirty, this method will mark it as clean.
		/// </summary>
		/// <param name="cacheKey">Key of element to update</param>
		/// <param name="newReferenceDataValue">New data value</param>
		/// <exception cref="ArgumentException">Cache key is not valid</exception>
		[MethodImpl(MethodImplOptions.Synchronized)]
		public void Update(string cacheKey, object newReferenceDataValue)
		{
			EnsureObjectExistsInCache(cacheKey);

			ReferenceDataDefinition dataDef = GetReferenceDataDefinition(cacheKey);
			InsertDataIntoCache(dataDef, newReferenceDataValue);
		}

		/// <summary>
		/// Retrieves data from the reference cache.
		/// </summary>
		/// <param name="cacheKey">Key of element to return</param>
		/// <returns>Data retrieved from cache, or null if not found</returns>
		[MethodImpl(MethodImplOptions.Synchronized)]
		public object Retrieve(string cacheKey)
		{
			ReferenceDataDefinition dataDef = GetReferenceDataDefinition(cacheKey);
			if(dataDef == null) return null;

			return cache.GetData(dataDef.UserDataCacheKey);
		}

		/// <summary>
		/// Removes an element from the cache. Removes metadata as well. If element is not found, method has no effect.
		/// </summary>
		/// <param name="cacheKey">Key of element to be removed</param>
		[MethodImpl(MethodImplOptions.Synchronized)]
		public void Remove(string cacheKey)
		{
			ReferenceDataDefinition dataDef = GetReferenceDataDefinition(cacheKey);
			if(dataDef == null) return;

			RemoveDataFromCache(dataDef);
		}

		/// <summary>
		/// Query method to determine if the element in the cache is marked dirty. Elements that are dirty have been updated
		/// on the local machine, but have not had their new values sent to a remote service. Once the values have been successfully
		/// sent, the element is marked clean.
		/// </summary>
		/// <param name="cacheKey">Key of item to check if dirty</param>
		/// <returns>True if the element is marked dirty</returns>
		/// <exception cref="ArgumentException">Cache key is not valid</exception>
		[MethodImpl(MethodImplOptions.Synchronized)]
		public bool IsDirty(string cacheKey)
		{
			EnsureObjectExistsInCache(cacheKey);

			ReferenceDataDefinition dataDef = GetReferenceDataDefinition(cacheKey);
			return dataDef.Dirty;
		}

		/// <summary>
		/// Marks an element as clean. Elements that are marked as dirty more than once must be marked as clean an equal number of times 
		/// before IsDirty will return false.
		/// </summary>
		/// <param name="cacheKey">Key of item to mark as clean</param>
		/// <exception cref="ArgumentException">Cache key is not valid</exception>
		[MethodImpl(MethodImplOptions.Synchronized)]
		public void MarkAsClean(string cacheKey)
		{
			EnsureObjectExistsInCache(cacheKey);

			ReferenceDataDefinition dataDef = GetReferenceDataDefinition(cacheKey);
			dataDef.MarkAsClean();

			cache.Add(dataDef.CacheKey, dataDef);
		}

		/// <summary>
		/// Marks the metadata associated with an element in the cache as having been expired. This starts the
		/// process of refreshing the element in the reference cache.
		/// </summary>
		/// <param name="expiredKey">Key of element that has been expired</param>
		/// <returns>The ReferenceDataDefinition of the element that has been expired if there is further 
		/// processing required to refresh that element, or null if the cache element is currently marked dirty or already expired</returns>
		[MethodImpl(MethodImplOptions.Synchronized)]
		public ReferenceDataDefinition ItemHasBeenExpired(string expiredKey)
		{
			ReferenceDataDefinition expiredDataDef = GetReferenceDataDefinition(expiredKey);
			AddRootElementToCache(expiredDataDef);

			if(HasItemMetadataAlreadyBeenMarkedAsExpired(expiredDataDef)) return null;
			if(IsItemDirty(expiredDataDef)) return null;

			expiredDataDef.Expired = true;
			cache.Add(expiredDataDef.CacheKey, expiredDataDef);

			return expiredDataDef;
		}

		private bool IsItemDirty(ReferenceDataDefinition expiredItem)
		{
			return expiredItem.Dirty == true;
		}

		private bool HasItemMetadataAlreadyBeenMarkedAsExpired(ReferenceDataDefinition expiredItem)
		{
			return expiredItem.Expired == true;
		}

		virtual protected CacheItemRemovedCallback GetExpirationDelegate()
		{
			return new CacheItemRemovedCallback(ReferenceDataRefreshController.CacheItemExpiredCallback);
		}

		private void EnsureCacheKeyValid(ReferenceDataDefinition dataDef)
		{
			if(dataDef.IsCacheKeyValid() == false) 
				throw new ArgumentException(String.Format(ResourceTable.GetString(ResourceTable.InvalidKeyToAddDataToReferenceCache), dataDef.OriginalKey));
		}

		private string GenerateMetadataKey(string key)
		{
			return "_" + key; 
		}

		private void EnsureObjectExistsInCache(string cacheKey)
		{
			if(DoesObjectExistInCache(cacheKey) == false)
			{
				throw new ReferenceCacheElementNotFoundException(cacheKey);
			}
		}

		private bool DoesObjectExistInCache(string cacheKey)
		{
			return GetReferenceDataDefinition(cacheKey) != null;
		}

		/// <summary>
		/// Inserts data into the cache as defined by the dataDef. This method attempts to be as exception safe as possible
		/// by catching any exception thrown during the adding of any part of this element and removing all traces of 
		/// the element and its metadata from the reference cache.
		/// </summary>
		/// <param name="dataDef"></param>
		/// <param name="valueToCache"></param>
		private void InsertDataIntoCache(ReferenceDataDefinition dataDef, object valueToCache)
		{
			UpdateDirtyCleanStateOfCachedData(dataDef, valueToCache);

			try
			{
				AddRootElementToCache(dataDef);
				cache.Add(dataDef.CacheKey, dataDef);
				cache.Add(dataDef.UserDataCacheKey, valueToCache);
			}
			catch(Exception errorOccuredInsideCacheBlock)
			{
				cache.Remove(dataDef.UserDataCacheKey);
				cache.Remove(dataDef.CacheKey);
				cache.Remove(dataDef.OriginalKey);

				throw errorOccuredInsideCacheBlock;
			}
		}

		/// <summary>
		/// Called to add the root element back into the cache after the caching block removed it when it expired.
		/// </summary>
		/// <param name="dataDef"></param>
		private void AddRootElementToCache(ReferenceDataDefinition dataDef)
		{
			if(dataDef.IsItemRefreshable())
			{
				ICacheItemExpiration [] expirations = CreateAbsoluteTimeExpiration(dataDef.RefreshInterval);
				cache.Add(dataDef.OriginalKey, dataDef.CacheKey, expirations, CacheItemPriority.Normal, GetExpirationDelegate());
			}
			else
			{
				cache.Add(dataDef.OriginalKey, dataDef.CacheKey);
			}
		}

		/// <summary>
		/// Creates the item expiration data used to control expirations in the caching block
		/// </summary>
		/// <param name="refreshInterval"></param>
		/// <returns></returns>
		private ICacheItemExpiration [] CreateAbsoluteTimeExpiration(int refreshInterval)
		{
			DateTime refreshDelay = DateTime.Now.AddSeconds(refreshInterval);
			AbsoluteTime refreshTime = new AbsoluteTime(refreshDelay);
			ICacheItemExpiration[] expirations = new ICacheItemExpiration[1];
			expirations[0] = refreshTime;

			return expirations;
		}

		private void RemoveDataFromCache(ReferenceDataDefinition dataDef)
		{
			cache.Remove(dataDef.UserDataCacheKey);
			cache.Remove(dataDef.CacheKey);
			cache.Remove(dataDef.OriginalKey);
		}

		private void UpdateDirtyCleanStateOfCachedData(ReferenceDataDefinition referenceDataDef, object newDataValue)
		{
			if(IsObjectAlreadyInCache(referenceDataDef.OriginalKey))
			{
				if(HasDataChanged(cache.GetData(referenceDataDef.UserDataCacheKey), newDataValue))
				{
					referenceDataDef.MarkAsDirty();
				}
			}
		}
		
		private bool IsObjectAlreadyInCache(string cacheKey)
		{
			return GetReferenceDataDefinition(cacheKey) != null;
		}

		private bool HasDataChanged(object oldValue, object newValue)
		{
			byte [] oldBytes = HashObject(oldValue);
			byte [] newBytes = HashObject(newValue);

			return AreHashedObjectsDifferent(oldBytes, newBytes);
		}

		private byte [] HashObject(object valueToHash)
		{
			SHA1 hashingAlgorithm = new SHA1CryptoServiceProvider();
			BinaryFormatter formatter = new BinaryFormatter();
			using(MemoryStream stream = new MemoryStream())
			{
				formatter.Serialize(stream, valueToHash);
				return hashingAlgorithm.ComputeHash(stream.ToArray());
			}
		}

		private bool AreHashedObjectsDifferent(byte [] oldValues, byte [] newValues)
		{
			if(oldValues.Length != newValues.Length) return true;

			for(int i = 0; i < oldValues.Length; i++)
			{
				if(oldValues[i] != newValues[i]) return true;
			}

			return false;
		}

		private ReferenceDataDefinition GetReferenceDataDefinition(string cacheKey)
		{
			string metadataKey = GenerateMetadataKey(cacheKey);
			return cache.GetData(metadataKey) as ReferenceDataDefinition;
		}
	}
}
