//=============================================================================
// Copyright  2004 Microsoft Corporation
// All rights reserved.
// THIS CODE AND INFORMATION IS PROVIDED "AS IS" WITHOUT WARRANTY
// OF ANY KIND, EITHER EXPRESSED OR IMPLIED, INCLUDING BUT NOT
// LIMITED TO THE IMPLIED WARRANTIES OF MERCHANTABILITY AND/OR
// FITNESS FOR A PARTICULAR PURPOSE.
//=============================================================================
using System;

namespace Microsoft.ApplicationBlocks.SmartClient.Offline
{
	/// <summary>
	/// ReferenceDataDefinition is the metadata associated with items in the ReferenceCache. 
	/// It controls how the information is refreshed, when it expires, and maintains dirty versus
	/// clean state of the data.
	/// </summary>
	/// 
	[Serializable()]
	public class ReferenceDataDefinition
	{
		private static readonly string CacheKeyPrefix = "_";
		private static readonly string UserKeyPrefix = "__";
		public static readonly int NoRefreshNeededEver = -1;

		private string cacheKey;
		private OnlineProxyContext refreshMethod;
		private int numberOfChanges;
		private int ageInSecondsFromCreationUntilRefreshNeeded;
		private bool isExpired;
		private object associatedRequestData;

		// Names should be CacheKey, MetadataCacheKey, and UserDataCacheKey

		/// <summary>
		/// Getter property for the cache key as given by the user. This is the key that is visible throughout the user code.
		/// What is stored at this location in the ReferenceDataCache is a pointer to the CacheKey
		/// </summary>
		/// <value>Original cache key given by the user</value>
		public string OriginalKey { get { return cacheKey; }}

		/// <summary>
		/// Getter property for the cache key of the metadata ReferenceDataDefinition
		/// </summary>
		/// <value>Cache key of the metadata</value>
		public string CacheKey { get {return CacheKeyPrefix + cacheKey; }}

		/// <summary>
		/// Getter property for the cache key at which the user data is actually stored.
		/// </summary>
		internal string UserDataCacheKey { get { return UserKeyPrefix + cacheKey; }}

		/// <summary>
		/// Getter property to retrieve the OnlineProxyContext used to retreive the data
		/// </summary>
		/// <value>Method in the Online proxy agent to do the refresh</value>
		public OnlineProxyContext RefreshMethod { get { return refreshMethod; }}

		/// <summary>
		/// Getter property used to retrieve the interval at which the data should be refreshed in the cache. if the 
		/// data will not ever be refreshed, this will return ReferenceDataDefinition.NoRefreshNeededEver
		/// </summary>
		/// <value>Refresh interval for the cached data</value>
		public int RefreshInterval { get { return ageInSecondsFromCreationUntilRefreshNeeded; }}

		/// <summary>
		/// Getter property for the arbitrary request data associated with this piece of data. This is a good place to store
		/// authorization credentials, qualifiers used by the online proxy to further discriminate the data retrieved, or any
		/// other data that would be useful when executing the online proxy.
		/// </summary>
		/// <value>Custom data associated with the request</value>
		public object AssociatedRequestData { get { return associatedRequestData; }}

		/// <summary>
		/// Getter property to retrieve the dirty state of the data
		/// </summary>
		/// <value>To store whether the reference data has been touched or not</value>
		public bool Dirty { get { return numberOfChanges != 0; }}

		/// <summary>
		/// Property used to expose the expired field of the data. 
		/// </summary>
		/// <value>Stores the expiry of the reference data</value>
		public bool Expired
		{
			get { return isExpired;	}
			set { isExpired = value; }
		}

		/// <summary>
		/// Constructor for ReferenceDataDefinition. Use this constructor if the element in the cache never needs to be refreshed
		/// </summary>
		/// <param name="cacheKey">Key of element in cache</param>
		/// <param name="refreshMethod">Object encapsulating online proxy to use to refresh this data</param>
		public ReferenceDataDefinition(string cacheKey, OnlineProxyContext refreshMethod)
			: this(cacheKey, NoRefreshNeededEver, refreshMethod, null)
		{
		}

		/// <summary>
		/// Constructor for ReferenceDataDefinition. Use this constructor if the element in the cache never needs to be refreshed
		/// </summary>
		/// <param name="cacheKey">Key of element in cache</param>
		/// <param name="refreshMethod">Object encapsulating online proxy to use to refresh this data</param>
		/// <param name="associatedRequestData">object used to pass arbitrary data used by the online proxy to retrieve the data</param>
		public ReferenceDataDefinition(string cacheKey, OnlineProxyContext refreshMethod, object associatedRequestData)
			: this(cacheKey, NoRefreshNeededEver, refreshMethod, associatedRequestData)
		{
		}

		/// <summary>
		/// Constructor for ReferenceDataDefinition. Use this constructor if refresh interval needs to be specified for this element.
		/// </summary>
		/// <param name="cacheKey">Key of element in cache</param>
		/// <param name="delayInSecondsUntilRefresh">Refresh delay in seconds</param>
		/// <param name="refreshMethod">object encapsulating online proxy to use to refresh this data</param>
		public ReferenceDataDefinition(string cacheKey, int delayInSecondsUntilRefresh, OnlineProxyContext refreshMethod)
			: this(cacheKey, delayInSecondsUntilRefresh, refreshMethod, null)
		{
		}

		/// <summary>
		/// Constructor for ReferenceDataDefinition. Use this constructor if refresh interval needs to be specified for this element.
		/// </summary>
		/// <param name="cacheKey">Key of element in cache</param>
		/// <param name="delayInSecondsUntilRefresh">Refresh delay in seconds</param>
		/// <param name="refreshMethod">object encapsulating online proxy to use to refresh this data</param>
		/// <param name="associatedRequestData">object used to pass arbitrary data used by the online proxy to retrieve the data</param>
		public ReferenceDataDefinition(string cacheKey, int delayInSecondsUntilRefresh, OnlineProxyContext refreshMethod, object associatedRequestData)
		{
			this.cacheKey = cacheKey;
			this.refreshMethod = refreshMethod;
			this.ageInSecondsFromCreationUntilRefreshNeeded = delayInSecondsUntilRefresh;
			this.associatedRequestData = associatedRequestData;
			this.numberOfChanges = 0;
		}

		/// <summary>
		/// Query method used to determine if a refresh is needed for this data. A refresh is possible when the data is considered
		/// both expired and clean. Dirty data is never refreshed.
		/// </summary>
		/// <returns>True if the data is ready to be refreshed</returns>
		public bool IsRefreshNeeded() { return Expired && !Dirty; }

		/// <summary>
		/// Query method to determine if the cache key is valid. Valid keys are not null and not empty
		/// </summary>
		/// <returns>True if the key is valid</returns>
		public bool IsCacheKeyValid() { return OriginalKey != null && OriginalKey.Length != 0; }

		/// <summary>
		/// Query method to determine if the item ever needs to be refreshed.
		/// </summary>
		/// <returns>True if the cache item was created with a refresh interval</returns>
		public bool IsItemRefreshable() { return ageInSecondsFromCreationUntilRefreshNeeded != NoRefreshNeededEver; }

		/// <summary>
		/// Marks the item in the cache as clean. Note that the item will still be considered dirty until MarkAsClean has been called
		/// once for each call to MarkAsDirty
		/// </summary>
		public void MarkAsClean()
		{
			if(numberOfChanges > 0) numberOfChanges--;
			isExpired = false;
		}

		/// <summary>
		/// Marks the item in the cache has having been changed.
		/// </summary>
		public void MarkAsDirty()
		{
			numberOfChanges++;
		}
	}
}
