//=============================================================================
// Copyright  2004 Microsoft Corporation
// All rights reserved.
// THIS CODE AND INFORMATION IS PROVIDED "AS IS" WITHOUT WARRANTY
// OF ANY KIND, EITHER EXPRESSED OR IMPLIED, INCLUDING BUT NOT
// LIMITED TO THE IMPLIED WARRANTIES OF MERCHANTABILITY AND/OR
// FITNESS FOR A PARTICULAR PURPOSE.
//=============================================================================
using System;
using Microsoft.ApplicationBlocks.Cache;

namespace Microsoft.ApplicationBlocks.SmartClient.Offline
{
	/// <summary>
	/// This class acts as the controller for the process of refreshing an
	/// expired element in the ReferenceDataCache. It is responsibile for knowing
	/// the algorithm to request that data be readded to the cache, and that a refresh
	/// message be queued through the DataLoaderManager. It is intentionally kept separate from the
	/// ReferenceDataCache to avoid dependency issues between that class and the DataLoaderManager.
	/// </summary>
	public class ReferenceDataRefreshController
	{
		private DataLoaderManager dlm;
		private IReferenceDataCache referenceCache;

		/// <summary>
		/// Constructor for ReferenceDataRefreshController. This method is left public to make it available for unit testing,
		/// but should not be invoked by client code.
		/// </summary>
		/// <param name="dlm">The DataLoaderManager</param>
		/// <param name="referenceCache">The ReferenceDataCache</param>
		public ReferenceDataRefreshController(DataLoaderManager dlm, IReferenceDataCache referenceCache)
		{
			this.dlm = dlm;
			this.referenceCache = referenceCache;
		}

		/// <summary>
		/// The algorithm for how an item is refreshed in the cache. This method is left public to make it available for unit tests.
		/// It should not be called by client code.
		/// </summary>
		/// <param name="expiredItemInCache">Key of element expired in cache</param>
		public void Refresh(string expiredItemInCache)
		{
			ReferenceDataDefinition expiredItem = referenceCache.ItemHasBeenExpired(expiredItemInCache);
			if(DoesItemNeedRefreshing(expiredItem))
			{
				Guid noAgentEverCalledBackForCacheRefreshGuid = Guid.Empty;
				dlm.RefreshData(noAgentEverCalledBackForCacheRefreshGuid, expiredItem);
			}
		}

		/// <summary>
		/// The entry point of the refresh algorithm. This is called to start the entire refresh process. The signature of this class
		/// is determined by the requirements placed on remove callback methods by the caching block
		/// </summary>
		/// <param name="expiredItemKey">Key of item expired in the cache</param>
		/// <param name="removedCause">Specification of why the item was removed from the cache</param>
		public static void CacheItemExpiredCallback(string expiredItemKey, CacheItemRemoveCause removedCause)
		{
			if(removedCause == CacheItemRemoveCause.Expired)
			{
				DataLoaderManager dlm = OfflineBlockBuilder.Instance.DataLoaderManager;
				IReferenceDataCache referenceCache = OfflineBlockBuilder.Instance.ReferenceDataCache;

				ReferenceDataRefreshController refreshController = new ReferenceDataRefreshController(dlm, referenceCache);
				refreshController.Refresh(expiredItemKey);
			}
		}

		private bool DoesItemNeedRefreshing(ReferenceDataDefinition item)
		{
			return item != null;
		}
	}
}
