//=============================================================================
// Copyright  2004 Microsoft Corporation
// All rights reserved.
// THIS CODE AND INFORMATION IS PROVIDED "AS IS" WITHOUT WARRANTY
// OF ANY KIND, EITHER EXPRESSED OR IMPLIED, INCLUDING BUT NOT
// LIMITED TO THE IMPLIED WARRANTIES OF MERCHANTABILITY AND/OR
// FITNESS FOR A PARTICULAR PURPOSE.
//=============================================================================
using System;
using System.Collections.Specialized;
using System.Collections;

namespace Microsoft.ApplicationBlocks.SmartClient.Offline
{
	/// <summary>
	/// The ServiceAgentRegistry acts as a repository for currently active Service Agents in
	/// an offline enabled application. The agents are added to the registry, and looked up by guid. Each 
	/// guid represents a single, instantiated Service Agent.
	/// </summary>
	public class ServiceAgentRegistry
	{
		private IDictionary registeredAgents = new HybridDictionary();
		private FailsafeServiceAgent failsafeAgent = new FailsafeServiceAgent();

		/// <summary>
		/// Getter property that returns the single FailsafeServiceAgent provided by the block
		/// </summary>
		/// <value>FailsafeServiceAgent object instance</value>
		public FailsafeServiceAgent FailsafeServiceAgent { get { return failsafeAgent; }}

		/// <summary>
		/// Returns the service agent identified by the given guid.
		/// </summary>
		/// <param name="serviceAgentGuid">Guid used as key for search. This should not be null.</param>
		/// <returns>ServiceAgent instance found by Guid, or FailsafeServiceAgent instance if no agent found</returns>
		public ServiceAgent Find(Guid serviceAgentGuid)
		{
			lock(registeredAgents.SyncRoot) {
				ServiceAgent agentToReturn = registeredAgents[serviceAgentGuid] as ServiceAgent;
				return agentToReturn != null ? agentToReturn : failsafeAgent;
			}
		}

		/// <summary>
		/// Register a service agent with the registry. This agent will be maintained in the registry until 
		/// it is explicitly removed. An exception is thrown if a duplicate agent is added to the registry.
		/// </summary>
		/// <param name="agentToRegister">ServiceAgent to register. This should not be null</param>
		public void Register(ServiceAgent agentToRegister)
		{
			if(agentToRegister.Guid.Equals(Guid.Empty)) 
			{
				throw new ArgumentException(ResourceTable.GetString(ResourceTable.UnableToRegisterServiceAgentWithEmptyGuid));
			}

			lock(registeredAgents.SyncRoot) {
				registeredAgents.Add(agentToRegister.Guid, agentToRegister);
			}
		}

		/// <summary>
		/// Removes a service agent from the registry.
		/// </summary>
		/// <param name="agentToRemove">Guid used as key for search. This should not be null.</param>
		public void Remove(Guid agentToRemove)
		{
			lock(registeredAgents.SyncRoot) {
				registeredAgents.Remove(agentToRemove);
			}
		}
	}
}
