//=============================================================================
// Copyright  2004 Microsoft Corporation
// All rights reserved.
// THIS CODE AND INFORMATION IS PROVIDED "AS IS" WITHOUT WARRANTY
// OF ANY KIND, EITHER EXPRESSED OR IMPLIED, INCLUDING BUT NOT
// LIMITED TO THE IMPLIED WARRANTIES OF MERCHANTABILITY AND/OR
// FITNESS FOR A PARTICULAR PURPOSE.
//=============================================================================
using System;
using System.Threading;
using System.Reflection;
using System.Security.Permissions;

using Microsoft.ApplicationBlocks.ExceptionManagement;

namespace Microsoft.ApplicationBlocks.SmartClient.Offline
{
	/// <summary>
	/// This class is responsible for returning the data to the original service agent on
	/// a thread pool thread.
	/// </summary>
	public class ThreadPoolInvoker
	{
		private ServiceAgent serviceAgent;
		private Payload payload;

		/// <summary>
		/// Constructor for ThreadPoolInvoker.
		/// </summary>
		/// <param name="serviceAgent">ServiceAgent to which to return results. Will never be null.</param>
		/// <param name="payload">Payload through which results will be returned. Will never be null</param>
		public ThreadPoolInvoker(ServiceAgent serviceAgent, Payload payload)
		{
			this.serviceAgent = serviceAgent;
			this.payload = payload;
		}

		/// <summary>
		/// Returns any results on thread pool thread
		/// </summary>
		public void ReturnResults()
		{
			if(UserSpecifiedServiceAgentFound())
			{
				WaitCallback methodToInvoke = new WaitCallback(ReturnResultsOnOtherThread);
				ThreadPool.QueueUserWorkItem(methodToInvoke);
			}
		}

		/// <summary>
		/// Returns any failures on thread pool thread
		/// </summary>
		public void ReturnFailure()
		{
			WaitCallback methodToInvoke = new WaitCallback(ReportFailureOnOtherThread);
			ThreadPool.QueueUserWorkItem(methodToInvoke);
		}

		/// <summary>
		/// Behavior to invoke on thread pool thread.
		/// </summary>
		/// <param name="notUsedStateObject">Unused state object. State comes from object itself</param>
		private void ReturnResultsOnOtherThread(object notUsedStateObject)
		{
			try
			{
				InvokeReturnResultsOnOtherThread();
			}
			catch(Exception e)
			{
				PublishException(e);

				payload.RecordFailure(e);
				ReportFailureOnOtherThread(notUsedStateObject);
			}
		}

		/// <summary>
		/// This was split off from the ReturnResultsOnOtherThread method to allow the SecurityException possibly
		/// thrown by the below code to be caught be the try/catch above. With it all in one method, the exception was
		/// thrown before you even entered the method, and the try/catch did not catch it. This meant that the exception 
		/// was lost, as was the data. This was a Bad Thing.
		/// </summary>
		[ReflectionPermission(SecurityAction.Demand, Flags=ReflectionPermissionFlag.MemberAccess)]
		private void InvokeReturnResultsOnOtherThread()
		{
			string methodToInvoke = payload.ResultCallbackTarget.MethodToInvoke;
			MethodInfo method = serviceAgent.GetType().GetMethod(methodToInvoke, 
				BindingFlags.Instance | BindingFlags.Public | BindingFlags.NonPublic);

			method.Invoke(serviceAgent, new object[] { payload });
		}

		/// <summary>
		///  This method is needed because ExceptionManagement requires reflection permission in order to publish
		///  an exception, and one reason why an exception might have been thrown is that we don't have reflection
		///  permission. By putting it into its own method, I can easily catch and ignore the exception thrown from
		///  ExceptionManagement
		/// </summary>
		/// <param name="e"></param>
		private void PublishException(Exception e)
		{
			try
			{
				ExceptionManager.Publish(e);
			}
			catch(Exception)
			{
			}
		}

		/// <summary>
		/// Behavior to invoke on thread pool thread.
		/// </summary>
		/// <param name="notUsedStateObject"></param>
		private void ReportFailureOnOtherThread(object notUsedStateObject)
		{
			try
			{
				serviceAgent.ReportError(payload);
			}
			catch(Exception e)
			{
				ExceptionManager.Publish(e);
			}
		}

		/// <summary>
		/// Ensures that the ServiceAgent being used is the one specified by the user, not a FailsafeAgent. The
		/// FailsafeServiceAgent should never be used on the nominal return path, as it doesn't have the method expected
		/// by the client code to be called.
		/// </summary>
		/// <returns>True if the ServiceAgent used to instantiate this class is a user-defined ServiceAgent</returns>
		private bool UserSpecifiedServiceAgentFound()
		{
			return (serviceAgent as FailsafeServiceAgent) == null;
		}
	}
}
