//=============================================================================
// Copyright  2004 Microsoft Corporation
// All rights reserved.
// THIS CODE AND INFORMATION IS PROVIDED "AS IS" WITHOUT WARRANTY
// OF ANY KIND, EITHER EXPRESSED OR IMPLIED, INCLUDING BUT NOT
// LIMITED TO THE IMPLIED WARRANTIES OF MERCHANTABILITY AND/OR
// FITNESS FOR A PARTICULAR PURPOSE.
//=============================================================================
using System;
using System.Threading;
using System.Runtime.CompilerServices;
using Microsoft.ApplicationBlocks.ExceptionManagement;

namespace Microsoft.ApplicationBlocks.SmartClient.Offline
{
	/// <summary>
	/// This class handles the mechanics of starting, stopping, and managing non-thread pool threads. All 
	/// methods of this class are thread safe.
	/// </summary>
	public class ThreadRunner
	{
		private WorkerThread workerThread;
		private ThreadStart delegateToRun;
		private object lockObject = new object();

		/// <summary>
		/// Constructor for ThreadRunner. 
		/// </summary>
		/// <param name="delegateToRun">ThreadStart delegate for method to be run in other thread</param>
		public ThreadRunner(ThreadStart delegateToRun)
		{
			if(delegateToRun == null) throw new ArgumentNullException("delegateToRun", "Delegate to thread to be started cannot be null.");
			this.delegateToRun = delegateToRun;
		}

		/// <summary>
		/// Causes a thread to be started with the delegateToRun as the method invoked on that thread.
		/// This method will not start a second thread until the first thread has been stopped through calling
		/// the Stop() method
		/// </summary>
		[MethodImpl(MethodImplOptions.Synchronized)]
		public void Start()
		{
			if(workerThread == null)
			{
				workerThread = CreateWorkerThread(DelegateToRun);
				workerThread.Start();
			}
		}

		/// <summary>
		/// Stops a thread dead in its tracks. Thread will complete when it polls value of isRunning again.
		/// </summary>
		[MethodImpl(MethodImplOptions.Synchronized)]
		public void Stop()
		{
			if(workerThread != null) 
			{
				workerThread.Stop();
				workerThread = null;
			}
		}

		/// <summary>
		/// Property to return current state of embedded thread.
		/// </summary>
		public bool IsRunning
		{  
			[MethodImpl(MethodImplOptions.Synchronized)]
			get
			{
				if(workerThread == null) return false;
				return workerThread.IsRunning;
			}
		}

		/// <summary>
		/// Factory method to create the worker thread. Subclasses may override this to change the type of worker thread
		/// to create
		/// </summary>
		/// <param name="delegateToRun">Delegate referring to the client behavior to invoke on worker thread</param>
		/// <returns></returns>
		protected virtual WorkerThread CreateWorkerThread(ThreadStart delegateToRun)
		{
			return new WorkerThread(delegateToRun);
		}

		/// <summary>
		/// Getter property to retrieve the delegate to run
		/// </summary>
		protected ThreadStart DelegateToRun { get { return delegateToRun; }}
	}
}
