//=============================================================================
// Copyright  2004 Microsoft Corporation
// All rights reserved.
// THIS CODE AND INFORMATION IS PROVIDED "AS IS" WITHOUT WARRANTY
// OF ANY KIND, EITHER EXPRESSED OR IMPLIED, INCLUDING BUT NOT
// LIMITED TO THE IMPLIED WARRANTIES OF MERCHANTABILITY AND/OR
// FITNESS FOR A PARTICULAR PURPOSE.
//=============================================================================
using System;
using System.Threading;

using Microsoft.ApplicationBlocks.ExceptionManagement;

namespace Microsoft.ApplicationBlocks.SmartClient.Offline
{
	/// <summary>
	/// WorkerThread is used by ThreadRunner to manage the housekeeping information belonging to
	/// a particular thread. There is one WorkerThread instance created for each thread to be spawned, 
	/// and each one contains its own instance of housekeeping information.
	/// </summary>
	public class WorkerThread
	{
		private ThreadStart delegateToRun;
		private Thread workingThread;
		private bool	isRunning = false;
		private const int FiveSecondJoinDelay = 5000;

		/// <summary>
		/// Getter property to return running state of thread. Returns true if thread is running
		/// </summary>
		public bool IsRunning 
		{
			get { return isRunning; }
		}

		/// <summary>
		/// Getter property to return the delete to run on this thread.
		/// </summary>
		protected ThreadStart DelegateToRun { get { return delegateToRun; }}

		/// <summary>
		/// Constructor for instance.
		/// </summary>
		/// <param name="delegateToRun">Delegate referring to client behavior to invoke on this thread</param>
		public WorkerThread(ThreadStart delegateToRun)
		{
			this.delegateToRun = delegateToRun;
		}

		/// <summary>
		/// Starts execution of this thread.
		/// </summary>
		public void Start()
		{
			isRunning = true;

			workingThread = new Thread(new ThreadStart(WorkerMethod));
			workingThread.Start();
		}

		/// <summary>
		/// Stops execution of this thread. This method may block momentarily while waiting for the thread to die and be joined.
		/// </summary>
		public void Stop()
		{
			isRunning = false;
			workingThread.Join(FiveSecondJoinDelay);
			workingThread = null;
		}

		/// <summary>
		/// This method is the worker method that will be spawned on the thread managed by this class. It is
		/// the responsibility of the caller to provide the delegateToRun, which will be run inside a loop until
		/// IsRunning is set to false through calling Stop(). Callers can also provide their own implementation for
		/// WorkerMethod, changing the looping logic to be whatever they need for their particular situation.
		/// </summary>
		virtual protected void WorkerMethod()
		{
			while(isRunning)
			{
				try
				{
					DelegateToRun();
				}
				catch(Exception e)
				{
					ExceptionManager.Publish(e);
				}
			}
		}
	}
}
