//=============================================================================
// Copyright  2004 Microsoft Corporation
// All rights reserved.
// THIS CODE AND INFORMATION IS PROVIDED "AS IS" WITHOUT WARRANTY
// OF ANY KIND, EITHER EXPRESSED OR IMPLIED, INCLUDING BUT NOT
// LIMITED TO THE IMPLIED WARRANTIES OF MERCHANTABILITY AND/OR
// FITNESS FOR A PARTICULAR PURPOSE.
//=============================================================================
using System;

namespace Microsoft.ApplicationBlocks.SmartClient.Offline
{
	/// <summary>
	/// This class represents the base class for all client messages that are going
	/// to be queued by the QueueManager. The several properties in this class represent 
	/// the application-level housekeeping information needed by the system to send the
	/// message, store the returned data into the cache through the CacheManager,
	/// or return the data back to the application.
	/// </summary>
	/// 
	[Serializable()]
	public class Payload
	{
		private OnlineProxyContext methodToExecute;
		private ServiceAgentContext resultCallbackTarget;
		private Guid payloadGuid = Guid.NewGuid();
		private Guid agentGuid = Guid.Empty;
		private Exception failureReason = null;
		private bool commandFailed = false;

		private object requestData;
		private object results;
		
		/// <summary>
		/// Getter property to retrieve the OnlineProxyContext describing the method to execute
		/// </summary>
		/// <value>Method context to execute</value>
		public OnlineProxyContext MethodToExecute { get { return methodToExecute; }}

		/// <summary>
		/// Getter property to retrieve the ServiceAgentContext describing the ServiceAgent method to execute to return
		/// the successful or failure notification
		/// </summary>
		/// <value>ServiceAgent instance of the caller</value>
		public ServiceAgentContext ResultCallbackTarget { get { return resultCallbackTarget; }}

		/// <summary>
		/// Getter property to retrieve the Guid associated with this Payload
		/// </summary>
		/// <value>Guid associated with the Payload</value>
		public Guid PayloadGuid { get { return payloadGuid; }}

		/// <summary>
		/// Getter property to retrieve the Guid associated with the ServiceAgent that initiated this execution
		/// </summary>
		/// <value>Guid associated with the ServiceAgent</value>
		public Guid ServiceAgentGuid { get { return agentGuid; }}

		/// <summary>
		/// Getter property to get the data accompanying this execution
		/// </summary>
		/// <value>Data passed for execution</value>
		public object RequestData {	get { return this.requestData; }}

		/// <summary>
		/// Properties to get and set the results associated with this execution
		/// </summary>
		/// <value>Results associated with this execution</value>
		public object Results
		{
			get { return this.results; }
			set { results = value; }
		}

		/// <summary>
		/// Properties to get and set the exception associated with any failure detected during execution
		/// </summary>
		/// <value>Exception containing the failure reason</value>
		public Exception FailureReason
		{
			get { return failureReason; }
			set { failureReason = value; }
		}

		/// <summary>
		/// Getter property to inform as to whether this execution succeeded or failed
		/// </summary>
		/// <value>Holds the success or failure status of the execution</value>
		public bool Success  { get { return commandFailed == false; }}

		/// <summary>
		/// Constructor for Payload base class.
		/// </summary>
		/// <param name="methodToExecute">Definition of method to call to send data to remote server</param>
		/// <param name="serviceAgentGuid">Guid of ServiceAgent that should be associated with this Payload</param>
		/// <param name="resultCallbackTarget">Definition of method to return results back to original caller</param>
		/// <param name="dataToSend">Data to send to the remote server</param>
		public Payload(OnlineProxyContext methodToExecute, Guid serviceAgentGuid, ServiceAgentContext resultCallbackTarget, object dataToSend)
		{
			this.methodToExecute = methodToExecute;
			this.resultCallbackTarget = resultCallbackTarget;
			this.requestData = dataToSend;
			this.agentGuid = serviceAgentGuid;
		}

		/// <summary>
		/// Client code should call this method in an online proxy when a failure is detected. This method sets the state of the 
		/// Payload to failure and records the exception that was thrown during the failure.
		/// </summary>
		/// <param name="e">Exception that occured during execution</param>
		public void RecordFailure(Exception e)
		{
			commandFailed = true;
			failureReason = e;
		}

		/// <summary>
		/// Client code should call this before attempting to reuse and requeue this payload for a retry after a failure. It
		/// clears out any embedded failures and exceptions, allowing the Payload to be reused for a retry of a failed command.
		/// </summary>
		public void ClearResults()
		{
			failureReason = null;
			commandFailed = false;
			results = null;
		}
	}
}
